/***********************************************************************
 *
 * Copyright (C) 2004 - 2005 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * @Author: Markus Pietrek
 * @References: [1] NS9750 Hardware Reference/December 2003
 * @TODO: Improve like sa1100_high_gpio_handler, handle all interrupts
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
/***********************************************************************
 *  @History:
 *     2005/10/24 : fixed bug #423
 *     2005/09/23 : fixed bug #409
 *     2004/11/26 : USB Host added
 *     2004/08/17 : removed addr_
 *     2004/06/09 : added DMA interrupts
 ***********************************************************************/

#include <linux/init.h>		/* __init */
#include <linux/cpufreq.h>	/* for mach/irq */
#include <linux/spinlock.h>	/* spinlock_irqsave */

#include <asm/irq.h>		/* NR_IRQS */
#include <asm/io.h>		/* writel */
#include <asm/mach/irq.h>	/* set_irq_flags */

#include <asm/arch/ns9750_sys.h> /* NS_SYS* */
#include <asm/arch/ns9750_bbus.h> /* NS_BBUS_MODULE_BASE */
#include <asm/arch/ns9750_bbus_bridge.h> /* NS_BBUS_BRIDGE */
#include <asm/arch/ns9750_usb.h> /* NS_USB_MODULE_BASE */
#include <asm/arch/a9m9xx0dev.h> /* FPGA_IER */


static struct resource resources[] = {
	{
		.name	= "irq_ahb",
		.start	= (NS_SYS_MODULE_BASE + NS_SYS_INT_VEC_ADR_BASE),
		.end		= (NS_SYS_MODULE_BASE + NS_SYS_INT_STAT_RAW + 3)
	}, {
		.name	= "irq_bbus_bridge",
		.start	= (NS_BBUS_BRIDGE_MODULE_BASE + NS_BBUS_BRIDGE_ISTAT),
		.end		= (NS_BBUS_BRIDGE_MODULE_BASE + NS_BBUS_BRIDGE_IEN + 3)
	}, {
		.name	= "irq_bbus",
		.start	= (NS_BBUS_MODULE_BASE + NS_BBUS_DMA_ISTAT),
		.end		= (NS_BBUS_MODULE_BASE + NS_BBUS_DMA_IEN + 3)
	}, {
		.name	= "irq_usb",
		.start	= (NS_USB_MODULE_BASE + NS_USB_IEN),
		.end		= (NS_USB_MODULE_BASE + NS_USB_ISTAT + 3)
	}, {
		.name	= "irq_fpga",
		.start	= (FPGA_IER),
		.end	= (FPGA_ISR)
	}, {
		.name = NULL
	}
};

static inline void ns_disable_ahb_irq( unsigned int irq );
static inline void ns_enable_ahb_irq( unsigned int irq );
static inline void ns_ack_ahb_irq( unsigned int irq );

static inline void ns_disable_bbus_irq( unsigned int irq );
static inline void ns_enable_bbus_irq( unsigned int irq );
static inline void ns_ack_bbus_irq( unsigned int irq );

static inline void ns_disable_bbus_dma_irq( unsigned int irq );
static inline void ns_enable_bbus_dma_irq( unsigned int irq );
static inline void ns_ack_bbus_dma_irq( unsigned int irq );

static inline void ns_disable_usb_irq( unsigned int irq );
static inline void ns_enable_usb_irq( unsigned int irq );
static inline void ns_ack_usb_irq( unsigned int irq );

static inline void ns_disable_fpga_irq( unsigned int irq );
static inline void ns_enable_fpga_irq( unsigned int irq );
static inline void ns_ack_fpga_irq( unsigned int irq );

static void ns_fpga_irq_demux_handler( unsigned int irq,
				       struct irqdesc *desc,
				       struct pt_regs *regs );


static struct irqchip ns_ahb_irq_chip = {
	.ack	= ns_ack_ahb_irq,
	.mask	= ns_disable_ahb_irq,
	.unmask = ns_enable_ahb_irq
};

static struct irqchip ns_bbus_irq_chip = {
	.ack	= ns_ack_bbus_irq,
	.mask	= ns_disable_bbus_irq,
	.unmask = ns_enable_bbus_irq
};

static struct irqchip ns_bbus_dma_irq_chip = {
	.ack	= ns_ack_bbus_dma_irq,
	.mask	= ns_disable_bbus_dma_irq,
	.unmask = ns_enable_bbus_dma_irq
};

static struct irqchip ns_usb_irq_chip = {
	.ack	= ns_ack_usb_irq,
	.mask	= ns_disable_usb_irq,
	.unmask = ns_enable_usb_irq
};

static struct irqchip ns_fpga_irq_chip = {
	.ack	= ns_ack_fpga_irq,
	.mask	= ns_disable_fpga_irq,
	.unmask = ns_enable_fpga_irq
};

spinlock_t ns_irq_lock;

void __init ns_init_irq( void )
{
	int i = 0;

	while( resources[ i ].name != NULL ) {
		request_resource( &iomem_resource, &resources[ i++ ] );
	}

	spin_lock_init( &ns_irq_lock );

	/* disable all IRQs */
	for( i = 0; i < 8; i++ )
		writel( 0, NS_SYS_INT_CFG_BASE + i*4 );

	for( i = 0; i < 32; i++ )
		writel( i, NS_SYS_INT_VEC_ADR( i ) );

	/* configure linux interrupt handling */
	for( i = 0; i < NR_IRQ_BBUS_START; i++ ) {
		set_irq_chip( i, &ns_ahb_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}
	for( i = NR_IRQ_BBUS_START; i < NR_IRQ_BBUS_DMA_START; i++ ) {
		set_irq_chip( i, &ns_bbus_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}
	for( i = NR_IRQ_BBUS_DMA_START; i < NR_IRQ_USB_START; i++ ) {
		set_irq_chip( i, &ns_bbus_dma_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}
/*	for( i = NR_IRQ_USB_START; i < NR_IRQS; i++ ) {
		set_irq_chip( i, &ns_usb_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}
*/
	for( i = NR_IRQ_USB_START; i < NR_IRQ_FPGA_START; i++ ) {
		set_irq_chip( i, &ns_usb_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}
	for( i = NR_IRQ_FPGA_START; i < NR_IRQS; i++ ) {
		set_irq_chip( i, &ns_fpga_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}

	set_irq_chained_handler( IRQ_EXT_2, ns_fpga_irq_demux_handler );

	/* disable all BBUS Interrupts all but DMA controller but propagate
	 * later enabled ones to sys module */
	writel( NS_BBUS_BRIDGE_IEN_GLBL |
		NS_BBUS_BRIDGE_IEN_USB  |
		NS_BBUS_BRIDGE_IEN_BBDMA,
		NS_BBUS_BRIDGE_IEN );
	/* we need no handler for it. IRQ_BBUS_AGGREGATE will trigger the BBUS
	 * handlers immediately in entry-armv.S */
	ns_enable_ahb_irq( IRQ_BBUS_AGGREGATE );
	ns_enable_bbus_irq( IRQ_BBUS_DMA_AGG );
}

static inline void ns_disable_ahb_irq( unsigned int irq )
{
	unsigned int reg;
	unsigned int byte;
	unsigned long flags;

	reg = irq / 4;
	byte = 3-(irq % 4);	/* INT_CFG is big endian, buggy chip! */

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_SYS_INT_CFG_BASE + reg*4,
	      & ~(NS_SYS_INT_CFG_IE << (byte*8) ) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_ahb_irq( unsigned int irq )
{
	unsigned int reg;
	unsigned int byte;
	unsigned int int_cfg;
	unsigned long flags;

	reg = irq / 4;
	byte = 3-(irq % 4);	/* INT_CFG is big endian, buggy chip! */

	/* @TODO: maybe we should set up some priority (timer first, then
	 * ethernet? */
	int_cfg = NS_SYS_INT_CFG_IE | irq;

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_SYS_INT_CFG_BASE + reg*4, |int_cfg << (byte*8) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_ack_ahb_irq( /*@unused@*/ unsigned int irq )
{
 	ns_disable_ahb_irq( irq );
	writel( 0, NS_SYS_ISRADDR ); /* [1] p. 125 */
}

static inline void ns_disable_bbus_irq( unsigned int irq )
{
	unsigned long flags;

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_BBUS_BRIDGE_IEN,
	      & ~(1 << (irq - NR_IRQ_BBUS_START)));
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_bbus_irq( unsigned int irq )
{
	unsigned long flags;

	spin_lock_irqsave( &ns_irq_lock, flags );
	/* enable BBUS IRQ */
	rmwl( NS_BBUS_BRIDGE_IEN,
	      | 1 << (irq - NR_IRQ_BBUS_START) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
	ns_enable_ahb_irq( IRQ_BBUS_AGGREGATE );  
}

static inline void ns_ack_bbus_irq( /*@unused@*/ unsigned int irq )
{
	ns_disable_bbus_irq( irq );
	ns_disable_ahb_irq( IRQ_BBUS_AGGREGATE );
	writel( 0, NS_SYS_ISRADDR ); /* [1] p. 125 */
}

static inline void ns_disable_bbus_dma_irq( unsigned int irq )
{
	unsigned long flags;

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_BBUS_DMA_IEN,
	      & ~(1 << (irq - NR_IRQ_BBUS_DMA_START) ) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_bbus_dma_irq( unsigned int irq )
{
	unsigned long flags;

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_BBUS_DMA_IEN,
	      | 1 << (irq - NR_IRQ_BBUS_DMA_START) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_ack_bbus_dma_irq( /*@unused@*/ unsigned int irq )
{
	ns_disable_bbus_dma_irq( irq );
	writel( 0, NS_SYS_ISRADDR ); /* [1] p. 125 */
}

static inline void ns_disable_usb_irq( unsigned int irq )
{
	unsigned long flags;
	unsigned int mask = 0;

	spin_lock_irqsave( &ns_irq_lock, flags );
	/* don't allow selection of individual DMA interrupts. */
	if( irq == IRQ_USB_DMA )
		mask = NS_USB_INT_DMA_ALL;
	else if( irq == IRQ_USB_FIFO )
		mask = NS_USB_INT_FIFO_ALL;
	else if( irq == IRQ_USB_HOST )
		mask = NS_USB_INT_HOST;

	writel( (readl( NS_USB_IEN ) & ~mask) |	NS_USB_INT_GBL_EN,
		NS_USB_IEN );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_usb_irq( unsigned int irq )
{
	unsigned long flags;
	unsigned int mask = 0;

	spin_lock_irqsave( &ns_irq_lock, flags );
	/* don't allow selection of individual DMA interrupts. */
	if( irq == IRQ_USB_DMA )
		mask = NS_USB_INT_DMA_ALL;
	else if( irq == IRQ_USB_FIFO )
		mask = NS_USB_INT_FIFO_ALL;
	else if( irq == IRQ_USB_HOST )
		mask = NS_USB_INT_HOST;

	rmwl( NS_USB_IEN, | (mask | NS_USB_INT_GBL_EN) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_ack_usb_irq( /*@unused@*/ unsigned int irq )
{
	ns_disable_usb_irq( irq );
	writel( 0, NS_SYS_ISRADDR ); /* [1] p. 125 */
}

static inline void ns_disable_fpga_irq( unsigned int irq )
{
	unsigned char temp;
	unsigned long flags;

	if (irq == IRQ_FPGA_TOUCH)
		/*
		 * If the interrupt number is the touch screen interrupt...
		 * do not disable that interrupt, this will be managed directly
		 * into the driver
		 */
		return;

	spin_lock_irqsave( &ns_irq_lock, flags );
	temp = readb( FPGA_IER );
	temp &= ~( 0x01 << ( irq - NR_IRQ_FPGA_START ) );
	writeb( temp, FPGA_IER);

	// If all interrupts disabled, disable IRQ_EXT_2
	// @TODO
	// if spurious irq this doesn't work
	//if( !temp )
	//	ns_disable_ahb_irq( IRQ_EXT_2 );

	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_fpga_irq( unsigned int irq )
{
	unsigned char temp;
	unsigned long flags;

	if (irq != IRQ_FPGA_TOUCH) {
		spin_lock_irqsave( &ns_irq_lock, flags );

		temp = readb( FPGA_IER );
		temp |= 0x01 << ( irq - NR_IRQ_FPGA_START );
		writeb( temp, FPGA_IER);
		spin_unlock_irqrestore( &ns_irq_lock, flags );
	}

	ns_enable_ahb_irq( IRQ_EXT_2 );
}

static inline void ns_ack_fpga_irq( unsigned int irq )
{
	ns_ack_ahb_irq( IRQ_EXT_2 );
}

static void ns_fpga_irq_demux_handler( unsigned int irq,
				       struct irqdesc *desc,
				       struct pt_regs *regs )
{
	unsigned char stat;
	unsigned int irqno;
	int i;

	stat = readb( FPGA_ISR ) & 0xff;

	for( i = 0; i < 8 && stat != 0; i++ ) {
		if( stat & 0x01 ) {
			irqno = NR_IRQ_FPGA_START + i;
			desc = irq_desc + irqno;

			desc->handle( irqno, desc, regs );
		}
		stat >>= 1;
	}
}

