/************************************************************************
 *
 * Copyright (C) 2005 by Sistemas Embebidos S.A.
 * A Digi International Company.
 * All rights reserved.
 *
 * $Id: ns9xx0_spi.c,v 1.3 2005-09-12 10:21:07 jdietsch Exp $
 * @Author: Pedro Perez de Heredia
 * @Descr: SPI communication functions for the NS9750/NS9360 NetSilicon
 *         uControllers.
 * @References: [1] NS9360 Hardware Reference Manual Version 90000675_A
 *                  March 2005.
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 ************************************************************************
 * @History:
 *	01/06/2005	PPH: Initial Version.
 *	12/09/2005	JD: removed inline
 *
 ************************************************************************
 * @TODO, @TOIMPROVE
 *
 ************************************************************************/

#include <linux/config.h>

#include <asm/arch/ns9750_ser.h>	// NS9750 serial registers
#include <asm/arch/ns9750_bbus.h>	// set_gpio_cfg
#include <asm/arch/hardware.h>		// BBUS_CLK_FREQ
#include <asm/arch/irqs.h>		// IRQ_BBUS_S1RXE
#include <asm/delay.h>			// udelay

#include <asm/arch/ns9xx0_spi.h>	// SPI_MASTER...

#define	NS_SPI		"ns9xx0_spi"


/***********************************************************************
 * @Function: ns9xx0_spi_dump_regs
 * @Return: nothing
 * @Descr: Dumps the register contents.
 ***********************************************************************/
void ns9xx0_spi_dump_regs( ns_spi_dev *dev )
{
	printk( "Control A:    0x%08x\n", readl( dev->offset + NS_SER_CTRL_A ) );
	printk( "Control B:    0x%08x\n", readl( dev->offset + NS_SER_CTRL_B ) );
	printk( "Bitrate:      0x%08x\n", readl( dev->offset + NS_SER_BITRATE ) );
	printk( "Rec Char T:   0x%08x\n", readl( dev->offset + NS_SER_RX_CHAR_TIMER ) );
	printk( "Rec Buf T:    0x%08x\n", readl( dev->offset + NS_SER_RX_BUF_TIMER ) );
	printk( "Rec Match:    0x%08x\n", readl( dev->offset + NS_SER_RX_MATCH ) );
	printk( "Rec Match M:  0x%08x\n", readl( dev->offset + NS_SER_RX_MATCH_MASK ) );
	printk( "Flow Ctrl:    0x%08x\n", readl( dev->offset + NS_SER_FLOW_CTRL ) );
	printk( "Flow Ctrl F:  0x%08x\n", readl( dev->offset + NS_SER_FLOW_CTRL_FORCE ) );
}


/***********************************************************************
 * @Function: ns9xx0_spi_init_port
 * @Return: nothing
 * @Descr: Initializes the device with the default settings.
 ***********************************************************************/
void __init ns9xx0_spi_init_port( ns_spi_dev *dev, int port )
{
	dev->port = port;
	dev->master = SPI_MASTER;
	dev->mode = SPI_MODE_0;
	dev->bitrate = NS_SPI_DEFAULT_BITRATE;
	dev->bit_order = SPI_MSBF;
	dev->gpios = ns_spi_gpios[port];
	dev->irq_rx = ( ( dev->port < 2 ) ? 0 : dev->port * 2 ) +
	              ( ( dev->port & 1 ) ? 0 : 2 ) + IRQ_BBUS_DMA_1;
	dev->irq_tx = dev->irq_rx + 1;
	dev->dma_rx = 1 + ( ( dev->port < 2 ) ? ( dev->port == 0 ) ? 2 : 0  : ( dev->port * 2 ) );
	dev->dma_tx = dev->dma_rx + 1;
	dev->base_addr = addr_ser_channel( NS_SER_CTRL_A, port );
	dev->offset = ( ( dev->port < 2 ) ? 0 : NS_SER_CHAN_OFFSET ) +
	              ( ( dev->port & 1 ) ? 0 : 0x40 );
}
EXPORT_SYMBOL( ns9xx0_spi_init_port );

/***********************************************************************
 * @Function: ns9xx0_spi_set_bitorder
 * @Return: 0 on success, -1 on error
 * @Descr: Configures the bitorder.
 ***********************************************************************/
int ns9xx0_spi_set_bitorder( ns_spi_dev *dev, spi_bitord_t order )
{
	unsigned int ctrl_b;

	if( order != SPI_MSBF && order != SPI_LSBF )
		return -1;

	ctrl_b = readl( dev->offset + NS_SER_CTRL_B ) & ~NS_SER_CTRL_B_BITORDR;
	if( order == SPI_MSBF )
		ctrl_b |= NS_SER_CTRL_B_BITORDR;
	writel( ctrl_b, dev->offset + NS_SER_CTRL_B );
	dev->bit_order = order;
	return 0;
}
EXPORT_SYMBOL( ns9xx0_spi_set_bitorder );

/***********************************************************************
 * @Function: ns9xx0_spi_set_master_slave
 * @Return: 0 on success, -1 on error
 * @Descr: Configures the SPI master/slave mode.
 ***********************************************************************/
int ns9xx0_spi_set_master_slave( ns_spi_dev *dev, spi_master_t master )
{
	unsigned int ctrl_b = 0;

	switch( master ) {
		case SPI_SLAVE:  ctrl_b |= NS_SER_CTRL_B_MODE_SPI_S; break;
		case SPI_MASTER: ctrl_b |= NS_SER_CTRL_B_MODE_SPI_M; break;
		default:	return -1;
	}

	writel( ( readl(dev->offset + NS_SER_CTRL_B) & ~NS_SER_CTRL_B_MODE_MA ) |
	        ctrl_b, dev->offset + NS_SER_CTRL_B );

	dev->master = master;
	return 0;
}
EXPORT_SYMBOL( ns9xx0_spi_set_master_slave );

/***********************************************************************
 * @Function: ns9xx0_spi_set_mode
 * @Return: 0 on success, -1 on error
 * @Descr: Configures the SPI mode.
 ***********************************************************************/
int ns9xx0_spi_set_mode( ns_spi_dev *dev, spi_mode_t mode )
{
	unsigned int mode_mask = 0;

	switch( mode ) {
		case SPI_MODE_0:
			mode_mask = NS_SER_BITRATE_SPCPOL_LOW;	break;
		case SPI_MODE_1:
			mode_mask = NS_SER_BITRATE_TXCINV | NS_SER_BITRATE_RXCINV; break;
		case SPI_MODE_2:
			mode_mask = NS_SER_BITRATE_SPCPOL_LOW |
		                    NS_SER_BITRATE_TXCINV | NS_SER_BITRATE_RXCINV; break;
		case SPI_MODE_3:
			mode_mask = 0;	break;
		default:	return -1;
	}

	writel( ( readl(dev->offset + NS_SER_BITRATE) & ~NS_SER_BITRATE_MOD_MA ) |
	        mode_mask, dev->offset + NS_SER_BITRATE );

	dev->mode = mode;
	return 0;
}
EXPORT_SYMBOL( ns9xx0_spi_set_mode );

/***********************************************************************
 * @Function: ns9xx0_spi_set_bitrate
 * @Return: 0 on success, -1 on error
 * @Descr: Configures the SPI bitrate. It configures the clock mux to
 *         use the BBUS clock.
 ***********************************************************************/
int ns9xx0_spi_set_bitrate( ns_spi_dev *dev, int bitrate )
{
	unsigned int N_value;
	unsigned int bitrate_reg;

	if( (bitrate < 0) || (bitrate > NS_SPI_MAX_BITRATE) )
		return -1;

	N_value = BBUS_CLK_FREQ / ( bitrate * 2 ) - 1;

	if( (N_value & NS_SER_BITRATE_N_MA) != N_value ) {
		 printk( KERN_WARNING NS_SPI
		         ": Unable to configure baudrate: %d\n", bitrate );
		return -2;
	}

	bitrate_reg = NS_SER_BITRATE_EBIT | NS_SER_BITRATE_CLKMUX_BCLK | NS_SER_BITRATE_TMODE;

	writel( ( readl(dev->offset + NS_SER_BITRATE) & ~NS_SER_BITRATE_N_MA ) |
	        N_value | bitrate_reg, dev->offset + NS_SER_BITRATE );

	dev->bitrate = bitrate;
	return 0;
}
EXPORT_SYMBOL( ns9xx0_spi_set_bitrate );


/***********************************************************************
 * @Function: ns9xx0_spi_port_setup
 * @Return:
 * @Descr:
 ***********************************************************************/
int ns9xx0_spi_port_setup( ns_spi_dev *dev )
{
	// Reset by writing 0 to control Reg A
	writel( 0x00000000,  dev->offset + NS_SER_CTRL_A );

	if( ns9xx0_spi_set_bitorder( dev, dev->bit_order ) )
		return -1;

	if( ns9xx0_spi_set_master_slave( dev, dev->master ) )
		return -1;

	if( ns9xx0_spi_set_mode( dev, dev->mode ) )
		return -1;

	if( ns9xx0_spi_set_bitrate( dev, dev->bitrate ) )
		return -1;

	writel( 0x0, dev->offset + NS_SER_RX_BUF_TIMER );
	writel( 0x0, dev->offset + NS_SER_RX_CHAR_TIMER );
	writel( 0x0, dev->offset + NS_SER_RX_MATCH );
	writel( 0x0, dev->offset + NS_SER_RX_MATCH_MASK );

	// Enable this channel
	writel( NS_SER_CTRL_A_CE |
	        NS_SER_CTRL_A_WLS_8,
	        dev->offset + NS_SER_CTRL_A );

	return 0;
}
EXPORT_SYMBOL( ns9xx0_spi_port_setup );

/***********************************************************************
 * @Function: ns9xx0_spi_request_port
 * @Return: 0 on success, error code otherwise.
 * @Descr: Allocates resources for the specified port: gpios and mem
 *         area.
 ***********************************************************************/
int ns9xx0_spi_request_port( ns_spi_dev *dev )
{
	int ret = 0, i;

	// request and configure GPIOs
	for( i = 0; i < 4; i++ ) {
		if( request_gpio( dev->gpios[i], 1, NS_SPI) == NULL ) {
			printk( KERN_ERR NS_SPI
				":GPIO %i reserved\n", dev->gpios[i] );
			goto error;
		}

		set_gpio_cfg( dev->gpios[i], NS_GPIO_CFG_FUNC_0 );
	}

	if( request_mem_region( dev->base_addr,
	                        addr_ser_channel( NS_SER_CTRL_A, 1 ) -
	                        addr_ser_channel( NS_SER_CTRL_A, 0 ),
	                        NS_SPI ) == NULL ) {
		printk( KERN_ERR NS_SPI ":request_mem_region failed\n" );
		ret = -EBUSY;
		goto error;
	}

	return ret;

error:
	/* i is +1 of the last reserved gpio */
	for( i--; i > 0; i-- )
		release_gpio( dev->gpios[ i ], 1 );

	return ret;
}
EXPORT_SYMBOL( ns9xx0_spi_request_port );


/***********************************************************************
 * @Function: ns9xx0_spi_release_port
 * @Return: nothing
 * @Descr: Releases the allocated resources: gpios and mem area.
 ***********************************************************************/
void ns9xx0_spi_release_port( ns_spi_dev *dev )
{
	int i;

	/* release gpios and the requested mem area */
	for( i = 0; i < 4; i++ ) {
		release_gpio( dev->gpios[i], 1 );
		set_gpio_cfg( dev->gpios[i], NS_GPIO_CFG_INPUT | NS_GPIO_CFG_FUNC_GPIO );
	}
	release_mem_region( dev->base_addr,
	                    addr_ser_channel( NS_SER_CTRL_A, 1 ) -
	                    addr_ser_channel( NS_SER_CTRL_A, 0 ) );

}
EXPORT_SYMBOL( ns9xx0_spi_release_port );

/***********************************************************************
 * @Function: ns9xx0_spi_stop_tx_dma
 * @Return: nothing
 * @Descr: Stops the SPI dma transmitions.
 ***********************************************************************/
void ns9xx0_spi_stop_tx_dma( ns_spi_dev *dev )
{
	rmwl( dev->offset + NS_SER_CTRL_A, & ~NS_SER_CTRL_A_ETXDMA );
}
EXPORT_SYMBOL( ns9xx0_spi_stop_tx_dma );

/***********************************************************************
 * @Function: ns9xx0_spi_start_tx_dma
 * @Return: nothing
 * @Descr: Starts (enables) the SPI dma transmitions.
 ***********************************************************************/
void ns9xx0_spi_start_tx_dma( ns_spi_dev *dev )
{
	rmwl( dev->offset + NS_SER_CTRL_A, | NS_SER_CTRL_A_ETXDMA );
}
EXPORT_SYMBOL( ns9xx0_spi_start_tx_dma );

/***********************************************************************
 * @Function: ns9xx0_spi_stop_rx_dma
 * @Return: nothing
 * @Descr: Stops the SPI dma reception.
 ***********************************************************************/
void ns9xx0_spi_stop_rx_dma( ns_spi_dev *dev )
{
	rmwl( dev->offset + NS_SER_CTRL_A, & ~NS_SER_CTRL_A_ERXDMA );
}
EXPORT_SYMBOL( ns9xx0_spi_stop_rx_dma );

/***********************************************************************
 * @Function: ns9xx0_spi_start_rx_dma
 * @Return: nothing
 * @Descr: Starts (enables) the SPI dma reception.
 ***********************************************************************/
inline void ns9xx0_spi_start_rx_dma( ns_spi_dev *dev )
{
	rmwl( dev->offset + NS_SER_CTRL_A, | NS_SER_CTRL_A_ERXDMA );
}
EXPORT_SYMBOL( ns9xx0_spi_start_rx_dma );

/***********************************************************************
 * @Function: ns9xx0_spi_tx_byte
 * @Return: Nothing.
 * @Descr: Writes a byte to the serial fifo register.
 ***********************************************************************/
void ns9xx0_spi_tx_byte( ns_spi_dev *dev, unsigned char byte )
{
	// Wait until spi channel ready
	while( !(readl( dev->offset + NS_SER_STAT_A ) & NS_SER_STAT_A_TRDY ) )
		/* do nothing, wait for characters in FIFO being sent */
		;

	writeb( byte, dev->offset + NS_SER_FIFO );
}
EXPORT_SYMBOL( ns9xx0_spi_tx_byte );

/***********************************************************************
 * @Function: ns9xx0_spi_tx_word
 * @Return: Nothing
 * @Descr: Writes a word to the serial fifo register.
 ***********************************************************************/
void ns9xx0_spi_tx_word( ns_spi_dev *dev, unsigned short int word )
{
	// Wait until spi channel ready
	while( !(readl( dev->offset + NS_SER_STAT_A ) & NS_SER_STAT_A_TRDY ) )
		/* do nothing, wait for characters in FIFO being sent */
		;

	writew( word, dev->offset + NS_SER_FIFO );
}
EXPORT_SYMBOL( ns9xx0_spi_tx_word );

/***********************************************************************
 * @Function: ns9xx0_spi_tx_dword
 * @Return: Nothing
 * @Descr: Writes a double word to the serial fifo register.
 ***********************************************************************/
void ns9xx0_spi_tx_dword( ns_spi_dev *dev, unsigned int dword )
{
	// Wait until spi channel ready
	while( !(readl( dev->offset + NS_SER_STAT_A ) & NS_SER_STAT_A_TRDY ) )
		/* do nothing, wait for characters in FIFO being sent */
		;

	writel( dword, dev->offset + NS_SER_FIFO );
}
EXPORT_SYMBOL( ns9xx0_spi_tx_dword );

/***********************************************************************
 * @Function: ns9xx0_spi_rx_byte
 * @Return: The byte read.
 * @Descr: Reads a byte from the serial fifo register.
 ***********************************************************************/
unsigned char ns9xx0_spi_rx_byte( ns_spi_dev *dev )
{
	unsigned int stat_reg;
	unsigned int max_loops = 10000000;

	while( max_loops-- ) {
		stat_reg = readl( dev->offset + NS_SER_STAT_A );

		if( stat_reg & NS_SER_STAT_A_RBC )
			writel( NS_SER_STAT_A_RBC, dev->offset + NS_SER_STAT_A );

		if( stat_reg & NS_SER_STAT_A_RRDY )
			return (unsigned char)readl( dev->offset + NS_SER_FIFO );
	}
	return 0;
}
EXPORT_SYMBOL( ns9xx0_spi_rx_byte );

/***********************************************************************
 * @Function: ns9xx0_spi_rx_word
 * @Return: The word read.
 * @Descr: Reads a word from the serial fifo register.
 ***********************************************************************/
unsigned short ns9xx0_spi_rx_word( ns_spi_dev *dev )
{
	unsigned int stat_reg;
	unsigned int max_loops = 10000000;

	while( max_loops-- ) {
		stat_reg = readl( dev->offset + NS_SER_STAT_A );

		if( stat_reg & NS_SER_STAT_A_RBC )
			writel( NS_SER_STAT_A_RBC, dev->offset + NS_SER_STAT_A );

		if( stat_reg & NS_SER_STAT_A_RRDY )
			return (unsigned short)readl( dev->offset + NS_SER_FIFO );
	}
	return 0;
}
EXPORT_SYMBOL( ns9xx0_spi_rx_word );

/***********************************************************************
 * @Function: ns9xx0_spi_rx_dword
 * @Return: The double word read.
 * @Descr: Reads a double word from the serial fifo register.
 ***********************************************************************/
unsigned int ns9xx0_spi_rx_dword( ns_spi_dev *dev )
{
	unsigned int stat_reg;
	unsigned int max_loops = 10000000;

	while( max_loops-- ) {
		stat_reg = readl( dev->offset + NS_SER_STAT_A );

		if( stat_reg & NS_SER_STAT_A_RBC )
			writel( NS_SER_STAT_A_RBC, dev->offset + NS_SER_STAT_A );

		if( stat_reg & NS_SER_STAT_A_RRDY )
			return readl( dev->offset + NS_SER_FIFO );
	}
	return 0;
}
EXPORT_SYMBOL( ns9xx0_spi_rx_dword );

/***********************************************************************
 * @Function: ns_spi_rx_byte
 * @Return:
 * @Descr:
 ***********************************************************************/
unsigned int ns_spi_rx_byte( ns_spi_dev *dev )
{
	unsigned int stat_reg;
	unsigned int max_loops = 1000000;

	while( max_loops-- ) {
		stat_reg = readl( dev->offset + NS_SER_STAT_A );

		if( stat_reg & NS_SER_STAT_A_RBC )
			writel( NS_SER_STAT_A_RBC, dev->offset + NS_SER_STAT_A );

		if( stat_reg & NS_SER_STAT_A_RRDY )
			return readl( dev->offset + NS_SER_FIFO );
	}
	return 0;
}
EXPORT_SYMBOL( ns_spi_rx_byte );

/***********************************************************************
 * @Function: ns_spi_tx_byte_rx
 * @Return:
 * @Descr:
 ***********************************************************************/
unsigned int ns_spi_tx_byte_rx( ns_spi_dev *dev, unsigned char byte )
{
	//unsigned char recived;
	unsigned int recived;

	// Wait until spi channel ready
	while( !(readl( dev->offset + NS_SER_STAT_A ) & NS_SER_STAT_A_TRDY ) )
		/* do nothing, wait for characters in FIFO being sent */
		;

	writeb( byte, dev->offset + NS_SER_FIFO );
	recived = ns_spi_rx_byte( dev );

	return recived;
}
EXPORT_SYMBOL( ns_spi_tx_byte_rx );

/***********************************************************************
 * @Function: ns_spi_tx_dword_rx
 * @Return:
 * @Descr:
 ***********************************************************************/
unsigned int ns_spi_tx_dword_rx( ns_spi_dev *dev, unsigned int dword )
{
	unsigned int recived;

	// Wait until spi channel ready
	while( !(readl( dev->offset + NS_SER_STAT_A ) & NS_SER_STAT_A_TRDY ) )
		/* do nothing, wait for characters in FIFO being sent */
		;

	writel( dword, dev->offset + NS_SER_FIFO );
	recived = ns_spi_rx_byte( dev );

	return recived;
}
EXPORT_SYMBOL( ns_spi_tx_dword_rx );

