/***********************************************************************
 *
 *  Copyright 2007 by Digi International
 *  All rights reserved.
 *
 *  $Id: i2c_eeprom.c,v 1.3 2007-01-25 14:27:37 tkuhlman Exp $
 *  Author: Pedro Perez de Heredia
 *  Descr:  Eeprom driver. Derived from eeprom.c by Frodo Looijaard
 *          and Philip Edelbrock. Also based in the 24C64.c driver
 *          written by Markus Pietrek.
 *  
 ***********************************************************************
 *  @History:
 *     2005/01/14 : Initial version.
 *     2007/01/25 : fixed: last byte not read (tk)
 ***********************************************************************/

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/init.h>
#include <asm/uaccess.h>		// copy_to_user and friends...
#include <linux/config.h>
#include <asm/delay.h>
#include <linux/init.h>
#include <linux/cdev.h>

#ifdef CONFIG_DEVFS_FS
# include <linux/devfs_fs_kernel.h>
#endif

#include "i2c_eeprom.h"			// specific data for this driver


// Some constant values
#define	I2C_EEPROM_VERSION	"V1.1"
#define	I2C_EEPROM_DRV_NAME	"i2c_eeprom"
#define I2C_EEPROM_ADDR		0x50	// I2C device address



// Function prototypes
static int i2c_eeprom_attach_adapter( struct i2c_adapter *adapter );
static int i2c_eeprom_detect( struct i2c_adapter *adapter, int address,
			 unsigned short flags, int kind );
static int i2c_eeprom_detach_client( struct i2c_client *client );
static ssize_t i2c_eeprom_read_page( struct i2c_client* client, char* buf,
                                loff_t off, size_t count, int page_size );
static ssize_t i2c_eeprom_write_page( struct i2c_client* client, char* buf, 
                             loff_t off, size_t count, int page_size );
static ssize_t i2c_eeprom_read_bytes( struct i2c_client* client, char* buf,
                               loff_t off, size_t count );
static ssize_t i2c_eeprom_write_bytes( struct i2c_client* client, char* buf,
                                       loff_t off, size_t count );
static loff_t i2c_eeprom_lseek( struct file * file, loff_t offset, int orig );
static ssize_t i2c_eeprom_write( struct file * file, const char * buf, size_t count,
                                 loff_t * ppos );
static ssize_t i2c_eeprom_read( struct file * file, char * buf, size_t count,
                                loff_t * ppos );
static int i2c_eeprom_open( struct inode * inode, struct file * file );
static int i2c_eeprom_release( struct inode * inode, struct file * file );


// Client data 
struct i2c_eeprom_data {
	
	struct semaphore	update_lock;
	i2c_eeprom_t 		eeprom;
};

// Driver data 
static struct i2c_driver i2c_eeprom_driver = {
	.owner		= THIS_MODULE,
	.name		= "i2c_eeprom",
	.id		= I2C_DRIVERID_EEPROM,
	.flags		= I2C_DF_NOTIFY,
	.attach_adapter	= i2c_eeprom_attach_adapter,
	.detach_client	= i2c_eeprom_detach_client,
};

// Driver file functions
static struct file_operations i2c_eeprom_fops = {
	llseek:		i2c_eeprom_lseek,
	read:		i2c_eeprom_read,
	write:		i2c_eeprom_write,
	open:		i2c_eeprom_open,
	release:	i2c_eeprom_release,
};

// Device structure
struct eep_dev {
	int			addr;
	int			type;
	loff_t			current_addr;
	struct i2c_client 	*i2c_eeprom_client;
	struct i2c_eeprom_data	data;
	struct cdev		cdev;     // Char device structure
};

static int i2c_eeprom_major = I2C_EEPROM_MAJOR;
static int i2c_eeprom_minor = 0;
static int i2c_eep_nr_devs = I2C_EEPROM_MAX_DEVS;
static int addr[I2C_EEPROM_MAX_DEVS] = {I2C_EEPROM_ADDR, 0, 0, 0};
static int type[I2C_EEPROM_MAX_DEVS] = {EEPROM_24C64, -1, -1, -1}; // 24C64, none, none, none
static struct eep_dev *eep_devices;

#define I2C_EEPROM_MAX_SUPPORTED	(sizeof( i2c_eeprom_supported ) / sizeof( i2c_eeprom_t ))



/***********************************************************************
 * @Function: i2c_eeprom_attach_adapter
 * @Return:   
 * @Descr:    
 ***********************************************************************/
static int i2c_eeprom_attach_adapter( struct i2c_adapter *adapter )
{
	int ret = 0;
	int i;
#ifdef CONFIG_DEVFS_FS
	int dev_counter = 0;
#endif	
	for( i = 0; i < i2c_eep_nr_devs; i++ ) {
	    if( eep_devices[i].type >= 0 ) {
	        ret = i2c_eeprom_detect( adapter, eep_devices[i].addr, i, eep_devices[i].type );
	        if( ret ) {
	            printk( KERN_INFO I2C_EEPROM_DRV_NAME ": dev(%d) Error i2c_eeprom_detect %d\n", i, ret );
	            break;
	        } else {
#ifdef CONFIG_DEVFS_FS
	            if( dev_counter == 0 )
	                devfs_mk_dir( "eeprom" );
	            devfs_mk_cdev( MKDEV( i2c_eeprom_major, dev_counter),
	                           S_IFCHR | S_IRUSR | S_IWUSR, "eeprom/%d", dev_counter );
	            dev_counter++;
#endif
	        }
	    }
	}
	
	return ret;
}


/***********************************************************************
 * @Function: i2c_eeprom_detect
 * @Return:   
 * @Descr:    
 ***********************************************************************/
static int i2c_eeprom_detect( struct i2c_adapter *adapter, int address,
                              unsigned short flags, int kind)
{
	int loop_cnt;
	struct i2c_client *client;
	struct i2c_eeprom_data *data;
	int err = 0;
	
	if( ( err = i2c_check_functionality( adapter, I2C_FUNC_SMBUS_BYTE_DATA ) ) == 0 )
	    return 1;
	    
	// Create the client structure
	if( ! ( client = kmalloc( sizeof( struct i2c_client ), GFP_KERNEL ))) 
	    return -ENOMEM;
        
	memset( client, 0x00, sizeof( struct i2c_client ) );

	data = &eep_devices[flags].data;
	client->addr = address;
	client->adapter = adapter;
	client->driver = &i2c_eeprom_driver;
	client->flags = I2C_CLIENT_ALLOW_USE | I2C_CLIENT_ALLOW_MULTIPLE_USE;

	// Determine the chip type
	for( loop_cnt = 0; 
	     loop_cnt < I2C_EEPROM_MAX_SUPPORTED;
	     loop_cnt++ ) 
	{
	    if( kind == i2c_eeprom_supported[loop_cnt].eep_id ) {
	        printk( "EEPROM type: %s, size: %d KB, page size: %d B\n", 
	                i2c_eeprom_supported[loop_cnt].eep_name, 
	                i2c_eeprom_supported[loop_cnt].eep_size / 1024, 
	                i2c_eeprom_supported[loop_cnt].eep_page_size );
	
	        memcpy( &data->eeprom, &i2c_eeprom_supported[loop_cnt], sizeof( i2c_eeprom_t ) );
	        break;
	    }
	}
	
	// Fill in the remaining client fields and put it into the global list 
	strncpy( client->name, "EEPROM chip", DEVICE_NAME_SIZE );
	init_MUTEX( &data->update_lock );

	// Notify to the I2C layer a new client has arrived
	if( ( err = i2c_attach_client( client ) ) ) {
	    kfree( client );
	    return err;
	}
	
	eep_devices[flags].i2c_eeprom_client = client;
	
	return 0;
}


/***********************************************************************
 * @Function: i2c_eeprom_detach_client
 * @Return:   
 * @Descr:    
 ***********************************************************************/
static int i2c_eeprom_detach_client( struct i2c_client *client )
{
	int err;
    
	if( ( err = i2c_detach_client( client ) ) ) {
		printk( KERN_ERR I2C_EEPROM_DRV_NAME 
		        ": Client deregistration failed, client not detached\n" );
		return err;
	}

	kfree( client );

	return 0;
}


/***********************************************************************
 * @Function: i2c_eeprom_read
 * @Return:   The number of bytes read or -EIO for I2C failure.
 * @Descr:    If possible, uses page accesses to read from eeprom. 
 *            Otherwise performs random accesses.
 ***********************************************************************/
static ssize_t i2c_eeprom_read_page( struct i2c_client* client, char* buf,
                                loff_t off, size_t count, int page_size )
{
	ssize_t overall = 0;
	ssize_t trans = 0;
	ssize_t trans_in;
	unsigned int mask = page_size - 1;

	
	while( count > overall ) {
	  
	    trans_in = (off & mask) ? 1 : ((count > page_size) ? page_size : count); 
	    
	    trans = i2c_eeprom_read_bytes( client, buf, off, trans_in );
	    if( trans != trans_in )
	        return -EIO;

	    overall += trans;
	    buf += trans;
	    off += trans;
	}

	return overall;
}

/***********************************************************************
 * @Function: i2c_eeprom_write
 * @Return:   The number of bytes written or -EIO for I2C failure.
 * @Descr:    If possible, uses page accesses to write to eeprom.
 *            Otherwise performs random accesses.
 ***********************************************************************/
static ssize_t i2c_eeprom_write_page( struct i2c_client* client, char* buf, 
                                      loff_t off, size_t count, int page_size )
{
	ssize_t overall = 0;
	ssize_t trans = 0;
	ssize_t trans_in;
	unsigned int mask = page_size - 1;
	
	while( count > overall ) {
	
	    trans_in = (off & mask) ? 1 : ((count > page_size) ? page_size : count); 
	    
	    trans = i2c_eeprom_write_bytes( client, buf, off, trans_in );

	    if( trans != trans_in )
	        return -EIO;

	    overall += trans;
	    buf += trans;
	    off += trans;
	}

	return overall;
}


/***********************************************************************
 * @Function: i2c_eeprom_read_bytes
 * @Return:   The number of bytes read or -EIO on I2C failure.
 * @Descr:    Reads count bytes, from the i2c bus, over the specified
 *            address.
 ***********************************************************************/
static ssize_t i2c_eeprom_read_bytes( struct i2c_client* client, char* buf,
                               loff_t off, size_t count )
{
	struct i2c_msg msg0 = { addr:client->addr,flags:0,len:2, buf:(char*)&off };
	struct i2c_msg msg1 = { addr:client->addr,flags:I2C_M_RD,len:count, buf:buf };
       
	// highest address byte first on wire 
	off = ((off & 0xff) << 8) | (off >> 8);
        if( i2c_transfer( client->adapter, &msg0, 1 ) != 1 )
                return -EIO;
		
	if( i2c_transfer( client->adapter, &msg1, 1 ) != 1 )
                return -EIO;
	
	return count;
}

/***********************************************************************
 * @Function: i2c_eeprom_write_bytes
 * @Return:   The number of Written or -EIO on I2C failure.
 * @Descr:    Writes count bytes, on the i2c bus, over the specified 
 *            address.
 ***********************************************************************/
static ssize_t i2c_eeprom_write_bytes( struct i2c_client* client,
                                       char* buf, loff_t off, size_t count )
{
	char *buffer;	
	struct i2c_msg msg0;
	int retval = count;
	
	// FIXME this doesn't work with current i2c driver
	//struct i2c_msg msg0 = { addr:client->addr,flags:0,len:2, buf:(char*)&ucBuffer[0]};
	//struct i2c_msg msg1 = { addr:client->addr,flags:I2C_M_NOSTART,len: count, buf:buf};
	
	buffer = (char *)kmalloc( count + 2 , GFP_KERNEL );
	if( buffer == NULL )
	{
	    printk( KERN_ERR "i2c_eeprom_write_bytes, malloc\n" );
	    return -ENOMEM;
	} 
	*buffer = (off >> 8) & 0xff;
	*(buffer + 1) = off & 0xff;
	memcpy( (buffer + 2), buf, count );
	
	msg0.addr = client->addr;
	msg0.flags = 0;
	msg0.len = count + 2;
	msg0.buf = buffer;
	
        if( i2c_transfer( client->adapter, &msg0, 1 ) != 1 )
	    retval = -EIO;
	
	kfree( buffer );
	
	return retval;
}

/***********************************************************************
 * @Function: i2c_eeprom_lseek
 * @Return:   The new cursor position or -EINVAL for wrong offset.
 * @Descr:    lseek driver functions.
 ***********************************************************************/
static loff_t i2c_eeprom_lseek( struct file * file, loff_t offset, int orig )
{
	struct eep_dev *dev = (struct eep_dev *)file->private_data;
	struct i2c_eeprom_data *data = &dev->data;
	
	if( offset >= data->eeprom.eep_size )	
	    return -EINVAL;
	else
	    return (dev->current_addr = offset );
}

/***********************************************************************
 * @Function: i2c_eeprom_write
 * @Return:   The number of bytes written or error code.
 * @Descr:    write driver functions.
 ***********************************************************************/
static ssize_t i2c_eeprom_write( struct file * file, const char * buf, size_t count,
                                 loff_t * ppos )
{
	struct eep_dev *dev = (struct eep_dev *)file->private_data;
	struct i2c_client *client = dev->i2c_eeprom_client;
	struct i2c_eeprom_data *data = &dev->data;
	int bytes;
	const char* tmp = buf;

	while( count )
	{
	    if( (dev->current_addr + count) > data->eeprom.eep_size )
	        break;	// don't write beneath length

	    bytes = i2c_eeprom_write_page( client, (char *)buf, dev->current_addr, count, data->eeprom.eep_page_size );
	    if( bytes < 0 ) {
	        printk( KERN_ERR "%s error code %d\n", __FUNCTION__, bytes );
	        break;
	    }
	    dev->current_addr += bytes;
	    count -= bytes;
	    buf += bytes;
	}

	return( buf - tmp );
}

/***********************************************************************
 * @Function: i2c_eeprom_read
 * @Return:   The number of bytes read or error code.
 * @Descr:    read driver functions.
 ***********************************************************************/
static ssize_t i2c_eeprom_read( struct file * file, char * buf, size_t count,
                                loff_t * ppos )
{
	struct eep_dev *dev = (struct eep_dev *)file->private_data;
	struct i2c_client *client = dev->i2c_eeprom_client;
	struct i2c_eeprom_data *data = &dev->data;
	int bytes;
	const char* tmp = buf;

	while( count )
	{
	    if( (dev->current_addr + count) > data->eeprom.eep_size )
	        break;

	    // Read from the eeprom the number of requested bytes...
	    bytes = i2c_eeprom_read_page( client, buf, dev->current_addr, count, data->eeprom.eep_page_size );
	    if( bytes < 0 ) {
	        printk( KERN_ERR "%s error code %d\n", __FUNCTION__, bytes );
	        break;    
	    }
	    dev->current_addr += bytes;
	    count -= bytes;
	    buf += bytes;
	}
	
	return( buf - tmp );
}


/***********************************************************************
 * @Function: i2c_eeprom_open
 * @Return:   0 on success, error code otherwise.
 * @Descr:    open driver functions.
 ***********************************************************************/
static int i2c_eeprom_open( struct inode * inode, struct file * file )
{
	struct eep_dev *dev;

	dev = container_of( inode->i_cdev, struct eep_dev, cdev );
	file->private_data = dev; // to have access from other methods 
	dev->current_addr = 0;
	
	return 0;
}

/***********************************************************************
 * @Function: i2c_eeprom_release
 * @Return:   0 on success, error code otherwise.
 * @Descr:    close driver functions.
 ***********************************************************************/
static int i2c_eeprom_release( struct inode * inode, struct file * file )
{
	return 0;
}

/***********************************************************************
 * @Function: i2c_eeprom_exit
 * @Return:   Nothing.
 * @Descr:    Module unload function. 
 ***********************************************************************/
static void i2c_eeprom_exit(void)
{
	int i;
	dev_t devno = MKDEV( i2c_eeprom_major, i2c_eeprom_minor );

	if( eep_devices ) {
	    for( i = 0; i < i2c_eep_nr_devs; i++ ) {
	        cdev_del( &eep_devices[i].cdev );
	    }
	    
	    kfree( eep_devices );
	}

	unregister_chrdev_region( devno, i2c_eep_nr_devs );
	i2c_del_driver( &i2c_eeprom_driver );

#ifdef CONFIG_DEVFS_FS
	for( i = 0; i < i2c_eep_nr_devs; i++ ) {
	    devfs_remove( "eeprom/%d", i );
	}
#endif

}


/***********************************************************************
 * @Function: i2c_eeprom_init
 * @Return:   0 on success, error code otherwise.
 * @Descr:    Module load function. 
 ***********************************************************************/
static int i2c_eeprom_init( void )
{	
	int ret, i, devno;
	dev_t dev;
	
	
	if( i2c_eeprom_major )
	{
	    dev = MKDEV( i2c_eeprom_major, i2c_eeprom_minor );
	    ret = register_chrdev_region( dev, i2c_eep_nr_devs, I2C_EEPROM_DRV_NAME );
	}
	else
	{
	    ret = alloc_chrdev_region( &dev, i2c_eeprom_minor,
	                               i2c_eep_nr_devs, I2C_EEPROM_DRV_NAME );
	    i2c_eeprom_major = MAJOR( dev );
	}
	
	if( ret < 0 )
	{
	    printk( KERN_WARNING I2C_EEPROM_DRV_NAME ": can't get major %d\n",
	            i2c_eeprom_major );
	    return ret;
	}

        // Allocate memory for the devices
	eep_devices = kmalloc( i2c_eep_nr_devs * sizeof( struct eep_dev ), GFP_KERNEL );
	if( !eep_devices )
	{
	    ret = -ENOMEM;
	    goto fail;
	}
	memset( eep_devices, 0, i2c_eep_nr_devs * sizeof( struct eep_dev ) );

        // Initialize each device
	for( i = 0; i < i2c_eep_nr_devs; i++ ) 
	{
	    eep_devices[i].addr = addr[i];
	    eep_devices[i].type = type[i];
	    
	    devno = MKDEV( i2c_eeprom_major, i2c_eeprom_minor + i );

	    cdev_init( &eep_devices[i].cdev, &i2c_eeprom_fops );
	    
	    eep_devices[i].cdev.owner = THIS_MODULE;
	    eep_devices[i].cdev.ops = &i2c_eeprom_fops;
	    if( cdev_add( &eep_devices[i].cdev, devno, 1 ) )
	        printk( KERN_NOTICE I2C_EEPROM_DRV_NAME ": Error adding %d\n", i );
	}

	if( ( ret = i2c_add_driver( &i2c_eeprom_driver ) ) != 0 )
	{
	    printk( KERN_NOTICE I2C_EEPROM_DRV_NAME ": i2c_add_driver %d\n", ret );
	    goto fail;
	}
	
	// All ok...	
	printk( KERN_INFO I2C_EEPROM_DRV_NAME ": EEPROM I2C driver " I2C_EEPROM_VERSION "\n" );
	return 0;
	
fail: 
	i2c_eeprom_exit();
	return ret;
}



MODULE_AUTHOR( "Pedro Perez de Heredia <pperez@embebidos.com>" );
MODULE_DESCRIPTION( "I2C EEPROM driver" );
MODULE_LICENSE( "GPL" );

module_init(i2c_eeprom_init);
module_exit(i2c_eeprom_exit);

