/*
 * drivers/char/ns9360_rtc.c
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/time.h>
#include <linux/rtc.h>
#include <linux/bcd.h>
#include <linux/interrupt.h>
#include <linux/ioctl.h>

#include <asm/uaccess.h>
#include <asm/rtc.h>
#include <asm/mach/time.h>

#include <asm/arch/regs-rtc.h>
#include <asm/arch/regs-sys.h>
#include <asm/arch/system.h>

#define DRIVER_NAME "ns9360_rtc"

#define ALARM_REGS		0
#define RTC_REGS		1

/* Those are the bits from a classic RTC we want to mimic (see linux/mc146818rtc.h) */
#ifndef RTC_IRQF
# define RTC_IRQF		0x80
#endif
#ifndef RTC_PF
# define RTC_PF			0x40
#endif
#ifndef RTC_AF
# define RTC_AF			0x20
#endif
#ifndef RTC_UF
# define RTC_UF			0x10
#endif

spinlock_t ns9360_rtc_lock;

/*
 * A few helper functions ...
 */

static const unsigned char days_in_mo[] =
    { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

#define is_leap(year) \
	((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))

static const unsigned short int __mon_yday[2][13] =
{
	/* Normal years. */
	{ 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
	/* Leap years. */
	{ 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
};

/* Disable RTC interrupts */
static void ns9360_rtc_irqdisable(u32 irqdis)
{
	RTC_IDIS = irqdis;
	RTC_IDIS = 0;
}

/* Enable alarm interrupt */
static void ns9360_rtc_setaie(int to)
{
	spin_lock_irq(&ns9360_rtc_lock);

	/* Configure the alarm to trigger when hour, min and sec match the programed value
	 * The rtc linux interface doesnt support programming calendar alarms */
	RTC_ALEN = (RTC_ALEN & ~RTC_AEN_MASK) | RTC_AEN_SEC_EN | RTC_AEN_MIN_EN | RTC_AEN_HOUR_EN;

	if (to)
		RTC_IEN = RTC_IEN_ALARM_EN;
	else {
		ns9360_rtc_irqdisable(RTC_IDIS_ALARM_DIS);
	}

	spin_unlock_irq(&ns9360_rtc_lock);
}

/*
 * Returns day since start of the year [0-365]
 *  (from drivers/char/efirtc.c)
 */
static inline int compute_yday(int year, int month, int day)
{
	return __mon_yday[is_leap(year)][month] + day-1;
}

/*
 * Decode time/date into rtc_time structure
 */
static void ns9360_rtc_decodetime(int rtc, struct rtc_time *tm)
{
	u32 time, date;

	if (rtc==RTC_REGS) {
		/* continuous read in case it changes */
		do {
			time = RTC_TIME;
			date = RTC_CAL;
		} while ((time != RTC_TIME) || (date != RTC_CAL));
	} else {
		time = RTC_TAL;
		date = RTC_CALAL;
	}

	tm->tm_hour = BCD2BIN((time & RTC_TIME_HOUR_MASK) >> 24);
	tm->tm_min  = BCD2BIN((time & RTC_TIME_MIN_MASK) >> 16);
	tm->tm_sec  = BCD2BIN((time & RTC_TIME_SEC_MASK) >> 8);

	tm->tm_mon  = (BCD2BIN((date & RTC_CAL_MONTH_MASK) >> 3) - 1);
	tm->tm_mday = BCD2BIN((date & RTC_CAL_MDAY_MASK) >> 8);
	tm->tm_wday = BCD2BIN(date & RTC_CAL_DAY_MASK) - 1;
	tm->tm_year  = BCD2BIN((date & RTC_CAL_CENTURY_MASK) >> 24) * 100;
	tm->tm_year += BCD2BIN((date & RTC_CAL_YEAR_MASK) >> 16);
	tm->tm_yday = compute_yday(tm->tm_year, tm->tm_mon, tm->tm_mday);
	tm->tm_year = tm->tm_year - 1900;

	if (rtc==ALARM_REGS) {
		tm->tm_mon = -1;
		tm->tm_mday = -1;
		tm->tm_wday = -1;
		tm->tm_year = -1;
		tm->tm_yday = -1;
	}

	pr_debug("%s(): %4d-%02d-%02d %02d:%02d:%02d\n", __FUNCTION__,
		1900 + tm->tm_year, tm->tm_mon, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
}

/*
 * Read current time and date in RTC
 */
static int ns9360_rtc_readtime(struct rtc_time *tm)
{
	ns9360_rtc_decodetime(RTC_REGS, tm);
	return 0;
}

/*
 * Set current time and date in RTC
 */
static int ns9360_rtc_settime(struct rtc_time *tm)
{
	u32 reg;

	/* Disable time and calendar */
	RTC_CGR &= ~(RTC_GCR_TIME_MASK | RTC_GCR_CAL_MASK);
	RTC_CGR |= RTC_GCR_TIME_DIS | RTC_GCR_CAL_DIS;

	reg = BIN2BCD(tm->tm_hour) << 24;
	reg |= BIN2BCD(tm->tm_min) << 16;
	reg |= BIN2BCD(tm->tm_sec) << 8;

	RTC_TIME = reg;

	reg = BIN2BCD(tm->tm_wday + 1);
	reg |= BIN2BCD((tm->tm_mon + 1)) << 3;
	reg |= BIN2BCD(tm->tm_mday) << 8;
	reg |= BIN2BCD((tm->tm_year+1900)%100) << 16;
	reg |= BIN2BCD((tm->tm_year+1900)/100) << 24;

	RTC_CAL = reg;

	/* Make sure time and calendar are valid */
	reg = RTC_GSTAT & (RTC_GSTAT_VALID_TIME_CONF|RTC_GSTAT_VALID_CAL_CONF);
	if (reg != (RTC_GSTAT_VALID_TIME_CONF|RTC_GSTAT_VALID_CAL_CONF)) {
		printk( KERN_WARNING DRIVER_NAME ": setting failed! status: 0x%x\n", RTC_GSTAT);
		return -EINVAL;
	}

	/* Enable time and calendar */
	RTC_CGR &= ~(RTC_GCR_TIME_MASK | RTC_GCR_CAL_MASK);
	RTC_CGR |= RTC_GCR_TIME_EN | RTC_GCR_CAL_EN;

	return 0;
}

/*
 * Read alarm time and date in RTC
 */
static int ns9360_rtc_readalarm(struct rtc_wkalrm *alrm)
{
	struct rtc_time *tm = &alrm->time;

	ns9360_rtc_decodetime(ALARM_REGS, tm);
	return 0;
}

/*
 * Set alarm time and date in RTC
 */
static int ns9360_rtc_setalarm(struct rtc_wkalrm *alrm)
{
	struct rtc_time *tm = &alrm->time;
	u32 reg;

	if ((tm->tm_hour >= 24) |
	    (tm->tm_min >= 60) |
	    (tm->tm_sec >= 60))
		return -EINVAL;

	reg = BIN2BCD(tm->tm_hour) << 24;
	reg |= BIN2BCD(tm->tm_min) << 16;
	reg |= BIN2BCD(tm->tm_sec) << 8;

	RTC_TAL = reg;

	/*
	 * The callendar alarm is not programed because 'calendar alarms are not
	 * supported by the rtc common driver code'.
	 */

	/* Make sure time and calendar alarm values are valid */
	reg = RTC_GSTAT & (RTC_GSTAT_VALID_TAL_CONF|RTC_GSTAT_VALID_CALAL_CONF);
	if (reg != (RTC_GSTAT_VALID_TAL_CONF|RTC_GSTAT_VALID_CALAL_CONF)) {
		printk( KERN_WARNING DRIVER_NAME ": alarm setting failed! status: 0x%x\n", RTC_GSTAT);
		return -EINVAL;
	}

	return 0;
}

/*
 * Handle commands from user-space
 */
static int ns9360_rtc_ioctl(unsigned int cmd, unsigned long arg)
{
	int ret = 0;

	switch (cmd) {
		case RTC_AIE_OFF:
		case RTC_AIE_ON:
			ns9360_rtc_setaie((cmd == RTC_AIE_ON) ? 1 : 0);
			break;
		/* The periodic interrupt could be used to program the rollover interrupts */
		case RTC_PIE_OFF:
		case RTC_PIE_ON:
		case RTC_IRQP_READ:	/* read periodic alarm frequency */
		case RTC_IRQP_SET:	/* set periodic alarm frequency */
		case RTC_UIE_ON:
		case RTC_UIE_OFF:
			ret = -EINVAL;
			break;

		default:
			ret = -EINVAL;
			break;
	}

	return ret;
}

/*
 * Provide additional RTC information in /proc/driver/rtc
 */
static int ns9360_rtc_read_proc(char *buf)
{
	u32 ies_reg;
	char *p = buf;

	ies_reg = RTC_ISTAT;
	p += sprintf(p, "alarm_IRQ\t: %s\n", (ies_reg & RTC_ISTAT_ALARM_EN) ? "yes" : "no");
	p += sprintf(p, "rollover_Month\t: %s\n", (ies_reg & RTC_ISTAT_MONTH_EN) ? "yes" : "no");
	p += sprintf(p, "rollover_Date\t: %s\n", (ies_reg & RTC_ISTAT_DATE_EN) ? "yes" : "no");
	p += sprintf(p, "rollover_Hour\t: %s\n", (ies_reg & RTC_ISTAT_HOUR_EN) ? "yes" : "no");
	p += sprintf(p, "rollover_Min\t: %s\n", (ies_reg & RTC_ISTAT_MIN_EN) ? "yes" : "no");
	p += sprintf(p, "rollover_Sec\t: %s\n", (ies_reg & RTC_ISTAT_SEC_EN) ? "yes" : "no");
	p += sprintf(p, "rollover_HSec\t: %s\n", (ies_reg & RTC_ISTAT_HSEC_EN) ? "yes" : "no");

	return p - buf;
}

/*
 * IRQ handler for the RTC
 */
static irqreturn_t ns9360_rtc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 rtc_evn;

	rtc_evn = RTC_EVN & ~RTC_ISTAT;

	if (!rtc_evn)
		return IRQ_NONE;

	if (rtc_evn&RTC_EVN_ALARM)
		rtc_update(1, RTC_AF | RTC_IRQF);

	return IRQ_HANDLED;
}

static int ns9360_rtc_open(void)
{
	return 0;
}

static void ns9360_rtc_release(void)
{
}

static struct rtc_ops rtc_ops = {
	.owner		= THIS_MODULE,
	.open		= ns9360_rtc_open,
	.release	= ns9360_rtc_release,
	.ioctl		= ns9360_rtc_ioctl,
	.read_time	= ns9360_rtc_readtime,
	.set_time	= ns9360_rtc_settime,
	.read_alarm	= ns9360_rtc_readalarm,
	.set_alarm	= ns9360_rtc_setalarm,
	.proc		= ns9360_rtc_read_proc,
};

/*
 * Initialize and install RTC driver
 */
static int ns9360_rtc_probe(struct device *dev)
{
	int ret;

	pr_debug("%s()\n", __FUNCTION__);

	/* The RTC should be already initialized by the bootloader, if not... we will init here */
	if (!(SYS_CCR & SYS_CLK_RTC_ON))
		SYS_CCR |= SYS_CLK_RTC_ON;

	SYS_RTCCCR = ns9xxx_systemclock()/200;

	if (RTC_CGR & RTC_GCR_TIME_DIS) {
		RTC_12_24 = (RTC_12_24 & ~RTC_12_24_MODE_MASK) | RTC_24_HOUR_EN;
		/* TODO check for valid configuration */
	}

	/* Disable all interrupt sources */
	ns9360_rtc_irqdisable(RTC_IDIS_MASK);

	spin_lock_init(&ns9360_rtc_lock);

	ret = request_irq(IRQ_RTC, ns9360_rtc_interrupt, SA_INTERRUPT, DRIVER_NAME, NULL);
	if (ret) {
		printk(KERN_ERR DRIVER_NAME ": IRQ %d already in use %d\n", IRQ_RTC, ret);
		return ret;
	}

	ret = register_rtc(&rtc_ops);
	if (ret) {
		printk(KERN_ERR DRIVER_NAME ": could not register RTC\n");
 		free_irq(IRQ_RTC, NULL);
 		return ret;
 	}

	printk(KERN_INFO "NS9360 Real Time Clock driver\n");

	return 0;
}

/*
 * Disable and remove the RTC driver
 */
static int ns9360_rtc_remove(struct device *dev)
{
	pr_debug("%s()\n", __FUNCTION__);

	/* Disable all interrupt sources */
	ns9360_rtc_irqdisable(RTC_IDIS_MASK);
	free_irq(IRQ_RTC, NULL);
	unregister_rtc(&rtc_ops);

	return 0;
}

#ifdef CONFIG_PM

static int ns9360_rtc_suspend(struct device *dev, u32 state, u32 level)
{
	// TODO
	return 0;
}

static int ns9360_rtc_resume(struct device *dev, u32 level)
{
	// TODO
	return 0;
}
#else
#define ns9360_rtc_suspend NULL
#define ns9360_rtc_resume  NULL
#endif

static struct platform_device ns9360_rtc_device = {
	.name		= DRIVER_NAME,
	.id		= -1,
};

static struct device_driver ns9360_rtc_driver = {
	.name		= DRIVER_NAME,
	.bus		= &platform_bus_type,
	.probe		= ns9360_rtc_probe,
	.remove		= ns9360_rtc_remove,
	.suspend	= ns9360_rtc_suspend,
	.resume		= ns9360_rtc_resume,
};

static int __init ns9360_rtc_init(void)
{
	int ret;

	ret = driver_register(&ns9360_rtc_driver);
	if (ret < 0) {
		printk(KERN_ERR "NS9360 RTC: Failed to register driver\n");
		return ret;
	}

	ret = platform_device_register(&ns9360_rtc_device);
	if (ret < 0) {
		printk(KERN_ERR "NS9360 RTC: Failed to register device\n");
		return ret;
	}

	return ret;
}

static void __exit ns9360_rtc_exit(void)
{
	platform_device_unregister(&ns9360_rtc_device);
	driver_unregister(&ns9360_rtc_driver);
}


module_init(ns9360_rtc_init);
module_exit(ns9360_rtc_exit);

MODULE_AUTHOR("Digi International Inc");
MODULE_DESCRIPTION("RTC driver for NS9360");
MODULE_LICENSE("GPL");

