
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/clk.h>

#include <asm/uaccess.h>
#include <asm/io.h>

#include <asm/arch/regs-sys.h>
#include <asm/arch/system.h>

#define DRIVER_NAME		"ns9xxx_wdt"
#define WD_DEFAULT_TIMEOUT	60

static int nowayout     = WATCHDOG_NOWAYOUT;
static int heartbeat	= WD_DEFAULT_TIMEOUT;
static int soft_noboot	= 0;

module_param(heartbeat,  int, 0);
module_param(nowayout,    int, 0);
module_param(soft_noboot, int, 0);

MODULE_PARM_DESC(heartbeat, "Watchdog heartbeat in seconds. default=" __MODULE_STRING(WD_DEFAULT_TIMEOUT) ")");
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");
MODULE_PARM_DESC(soft_noboot, "Watchdog action, set to 1 to ignore reboots, 0 to reboot (default depends on ONLY_TESTING)");

static DECLARE_MUTEX(open_lock);

struct ns9xxx_wdt {
	struct device	*dev;
	int		reglen;
	int		irq;
	int		timeout;
	void __iomem	*wdt_base;
	unsigned int	wdt_phys;
	unsigned long	ticks_per_sec;
} watchdog;

static struct platform_device *ns9xxx_wdt_dev;

/* watchdog control routines */
static void ns9xxx_wdt_enable(void)
{
	SYS_SWC |= SYS_SWC_EN_ON;
}

// static void ns9xxx_wdt_disable(void)
// {
// 	/* Nothing todo, once the watchdog was enabled, it can't be disabled */
// }

static int ns9xxx_wdt_set_clock(struct ns9xxx_wdt *wd, int timeout)
{
	int j, wdt_clock = ns9xxx_cpuclock()/2;

	if (timeout < 1)
		return -EINVAL;

	for (j=0; j<6; j++, wdt_clock/=2) {
		if (timeout <= (0xffffffff/wdt_clock))
			break;
	}

	if (j == 6)
		return -EINVAL;

	wd->ticks_per_sec = wdt_clock;

	SYS_SWT = wdt_clock * timeout;
	SYS_SWC = (SYS_SWC & ~SYS_SWC_TCS) | j;

	return 0;
}

static int ns9xxx_wdt_keepalive(struct ns9xxx_wdt *wd)
{
	SYS_SWT = wd->ticks_per_sec * wd->timeout;
	return 0;
}

static int ns9xxx_wdt_start(struct ns9xxx_wdt *wd)
{
	int ret;

	if ((ret = ns9xxx_wdt_set_clock(wd, wd->timeout)) < 0)
		return ret;

	if (soft_noboot) {
		SYS_SWC = (SYS_SWC & ~SYS_SWC_IRESP) | SYS_SWC_IRESP_IRQ;
	} else {
		SYS_SWC = (SYS_SWC & ~SYS_SWC_IRESP) | SYS_SWC_IRESP_RST;
	}

	ns9xxx_wdt_enable();

	return 0;
}

/*
 *	/dev/watchdog handling
 */

static int ns9xxx_wdt_open(struct inode *inode, struct file *file)
{
	int ret;
	struct ns9xxx_wdt *wd = platform_get_drvdata(ns9xxx_wdt_dev);

	if(down_trylock(&open_lock))
		return -EBUSY;

	if (nowayout) {
		__module_get(THIS_MODULE);
	}

	file->private_data = wd;

	/* start the timer */
	ret = ns9xxx_wdt_start(wd);
	if (ret < 0)
		return ret;

	return nonseekable_open(inode, file);
}

static int ns9xxx_wdt_release(struct inode *inode, struct file *file)
{
	struct ns9xxx_wdt *wd = file->private_data;

	printk(KERN_CRIT DRIVER_NAME ": Unexpected close, not stopping watchdog!\n");
	ns9xxx_wdt_keepalive(wd);

	up(&open_lock);
	return 0;
}

static ssize_t ns9xxx_wdt_write(struct file *file, const char __user *data,
				size_t len, loff_t *ppos)
{
	struct ns9xxx_wdt *wd = file->private_data;
	size_t i;

	if (len) {
		if (!nowayout) {
			/* In case it was set long ago */
			for (i = 0; i != len; i++) {
				char c;

				if (get_user(c, data + i))
					return -EFAULT;
				/* 'V' is not checked because close is not allowed.
				 * No way to stop the watchdog... */
			}
		}
		ns9xxx_wdt_keepalive(wd);
	}
	return len;
}

#define OPTIONS WDIOF_SETTIMEOUT | WDIOF_KEEPALIVEPING | WDIOF_MAGICCLOSE

static struct watchdog_info ns9xxx_wdt_ident = {
	.options          =     OPTIONS,
	.firmware_version =	0,
	.identity         =	"NS9XXX Watchdog",
};


static int ns9xxx_wdt_ioctl(struct inode *inode, struct file *file,
			    unsigned int cmd, unsigned long arg)
{
	struct ns9xxx_wdt *wd = file->private_data;
	void __user *argp = (void __user *)arg;
	int __user *p = argp;
	int new_hb;

	switch (cmd) {
		default:
			return -ENOIOCTLCMD;

		case WDIOC_GETSUPPORT:
			return copy_to_user(argp, &ns9xxx_wdt_ident,
				sizeof(ns9xxx_wdt_ident)) ? -EFAULT : 0;

		case WDIOC_GETSTATUS:
		case WDIOC_GETBOOTSTATUS:
			return put_user(0, p);

		case WDIOC_KEEPALIVE:
			ns9xxx_wdt_keepalive(wd);
			return 0;

		case WDIOC_SETTIMEOUT:
			if (get_user(new_hb, p))
				return -EFAULT;
			/* TODO check limits... XXX */

			if (ns9xxx_wdt_set_clock(wd, new_hb))
				return -EINVAL;

			wd->timeout = new_hb;
			ns9xxx_wdt_keepalive(wd);
			return put_user(wd->timeout, p);

		case WDIOC_GETTIMEOUT:
			return put_user(wd->timeout, p);
	}
}

/* kernel interface */

static struct file_operations ns9xxx_wdt_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= ns9xxx_wdt_write,
	.ioctl		= ns9xxx_wdt_ioctl,
	.open		= ns9xxx_wdt_open,
	.release	= ns9xxx_wdt_release,
};

static struct miscdevice ns9xxx_wdt_miscdev = {
	.minor		= WATCHDOG_MINOR,
	.name		= "watchdog",
	.fops		= &ns9xxx_wdt_fops,
};

/* interrupt handler code */
static irqreturn_t ns9xxx_wdt_irq(int irqno, void *param,
				  struct pt_regs *regs)
{
	struct ns9xxx_wdt *wd = (struct ns9xxx_wdt *)param;

	printk(KERN_INFO DRIVER_NAME ": Watchdog timer expired!\n");

	ns9xxx_wdt_keepalive(wd);

	SYS_SWC |= SYS_SWC_IC_ON;
	SYS_SWC &= ~SYS_SWC_IC;

	return IRQ_HANDLED;
}


/* device interface */
static int ns9xxx_wdt_probe(struct platform_device *pdev)
{
	struct resource *res;
	struct ns9xxx_wdt *wd;
	int ret;

	wd = kmalloc(sizeof(struct ns9xxx_wdt), GFP_KERNEL);
	if (!wd)
		return -ENOMEM;

	memset(wd, 0, sizeof(struct ns9xxx_wdt));
	wd->dev = &pdev->dev;
	wd->timeout = heartbeat;

	/* Retrieve resource information */
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		printk(KERN_INFO DRIVER_NAME ": failed to get memory region resouce\n");
		ret = -ENOENT;
		goto error;
	}

	wd->reglen = (res->end - res->start) + 1;
	wd->wdt_phys = res->start;

	res = request_mem_region(wd->wdt_phys, wd->reglen, pdev->name);
	if (res == NULL) {
		printk(KERN_INFO DRIVER_NAME ": failed to get memory region\n");
		ret = -ENOENT;
		goto error;
	}

	wd->wdt_base = ioremap(res->start, wd->reglen);
	if (wd->wdt_base == 0) {
		printk(KERN_INFO DRIVER_NAME ": failed to ioremap() region\n");
		ret = -EINVAL;
		goto error1;
	}

	res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (res == NULL) {
		printk(KERN_INFO DRIVER_NAME ": failed to get irq resource\n");
		ret = -ENOENT;
		goto error1;
	}
	wd->irq = res->start;

	ret = request_irq(wd->irq, ns9xxx_wdt_irq, SA_INTERRUPT, pdev->name, wd);
	if (ret != 0) {
		printk(KERN_INFO DRIVER_NAME ":failed to install irq (%d)\n", ret);
		goto error1;
	}

	ns9xxx_wdt_miscdev.dev = &pdev->dev;
	ret = misc_register(&ns9xxx_wdt_miscdev);
	if (ret) {
		printk (KERN_ERR DRIVER_NAME ": cannot register miscdev on minor=%d (%d)\n",
			WATCHDOG_MINOR, ret);
		goto error2;
	}

	platform_set_drvdata(pdev, wd);
	ns9xxx_wdt_dev = pdev;

	printk(KERN_INFO "Watchdog: NS9XXX watchdog initialized\n");

	return 0;

error2:
	free_irq(wd->irq, wd);
error1:
	release_mem_region(wd->wdt_phys, wd->reglen);
error:
	kfree(wd);
	return ret;
}

static int ns9xxx_wdt_remove(struct platform_device *dev)
{
	struct ns9xxx_wdt *wd = platform_get_drvdata(dev);

	release_mem_region(wd->wdt_phys, wd->reglen);
	free_irq(wd->irq, wd);
	iounmap(wd->wdt_base);
	kfree(wd);
	misc_deregister(&ns9xxx_wdt_miscdev);

	return 0;
}

static void ns9xxx_wdt_shutdown(struct platform_device *dev)
{
	/* No way to stop... */
}

#ifdef CONFIG_PM
static int ns9xxx_wdt_suspend(struct platform_device *dev, pm_message_t state)
{
	return 0;
}

static int ns9xxx_wdt_resume(struct platform_device *dev)
{
	return 0;
}

#else
#define ns9xxx_wdt_suspend NULL
#define ns9xxx_wdt_resume  NULL
#endif /* CONFIG_PM */


static struct platform_driver ns9xxx_wdt_driver = {
	.probe		= ns9xxx_wdt_probe,
	.remove		= ns9xxx_wdt_remove,
	.shutdown	= ns9xxx_wdt_shutdown,
	.suspend	= ns9xxx_wdt_suspend,
	.resume		= ns9xxx_wdt_resume,
	.driver		= {
		.owner	= THIS_MODULE,
		.name	= DRIVER_NAME,
	},
};

static int __init ns9xxx_wdt_init(void)
{
	return platform_driver_register(&ns9xxx_wdt_driver);
}

static void __exit ns9xxx_wdt_exit(void)
{
	platform_driver_unregister(&ns9xxx_wdt_driver);
}

module_init(ns9xxx_wdt_init);
module_exit(ns9xxx_wdt_exit);

MODULE_AUTHOR("Digi International Inc");
MODULE_DESCRIPTION("NX9XXX Watchdog Driver");
MODULE_LICENSE("GPL");
