/***********************************************************************
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: nand_a9m9750.c,v 1.6 2006-08-16 13:15:56 bwesterm Exp $
 * @Author: Markus Pietrek
 * @Descr: NAND driver for A9M9750
 * @References: [1] u-boot-1.1.0/include/ns9750_nand.h
 *              [2] spia.c
 *              [3] nand.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
/***********************************************************************
 *  @History:
 *	2005/09/21 : changed static partitions, changed naming to CC9P9xx0
 *	2005/03/07 : changed static partitions
 *     2004/09/07 : fixed commandline parsing
 ***********************************************************************/

#include <linux/init.h>		/* __init */
#include <linux/module.h>	/* module_init */
#include <linux/mtd/mtd.h>	/* add_mtd_device */
#include <linux/mtd/nand.h>	/* NAND_CMD_*  */
#include <linux/mtd/partitions.h> /* parse_mtd_partitions */

#include <asm/io.h>		/* ioremap */
#include <asm/sizes.h>		/* SZ_ */
#include <asm/delay.h>		/* udelay */

static struct mtd_info* a9m9750_mtd = NULL;

#define PHYS_NAND_FLASH	0x50000000 			/* NAND Flash Bank #1 */

#if defined (CONFIG_MACH_CC9C) || defined (CONFIG_MACH_CCW9C)
#define NAND_FLASH_DAT	(0x00000000)
#define NAND_FLASH_ADR	(0x00008000)
#define NAND_FLASH_CMD	(0x00010000)
#else
#define NAND_FLASH_DAT	(0x0000) /* Data Read/Write */
#define NAND_FLASH_ADR	(0x2000) /* Adr Write */
#define NAND_FLASH_CMD	(0x4000) /* Cmd Write */
#endif

#define NAND_IO_ADDR(x)  ((x)->IO_ADDR_W+NAND_FLASH_DAT)
#define NAND_CMD_ADDR(x) ((x)->IO_ADDR_W+NAND_FLASH_CMD)
#define NAND_ADR_ADDR(x) ((x)->IO_ADDR_W+NAND_FLASH_ADR)

#ifdef CONFIG_MTD_PARTITIONS
/*
 * Define static partitions for flash device
 */
static struct mtd_partition partition_info[] = {
        [0] = {
                .name   = "u-boot",
                .size   = 256*1024,
                .offset = 0,
                .mask_flags     = MTD_WRITEABLE,  /* force read-only */
        },
        [1] = {
                .name   = "kernel",
                .size   = (3*1024*1024)-(256*1024),
                .offset = 256*1024,
        },
        [2] = {
                .name   = "rootfs",
                .offset = 3*1024*1024,
                .size   = MTDPART_SIZ_FULL,
        }
};

static const char* probes[] = { "cmdlinepart", NULL };
# define NUM_PARTITIONS 3
#endif /* CONFIG_MTD_PARTITIONS */

/***********************************************************************
 * @Function: a9m9750_nand_hwcontrol
 * @Return: n/a
 * @Descr: does nothing but must be present
 ***********************************************************************/

static void a9m9750_nand_hwcontrol( struct mtd_info* mtd, int cmd )
{
	/* unused. Memory controller does all stuff */
}

/***********************************************************************
 * @Function: a9m9750_nand_command
 * @Return: n/a
 * @Descr: taken from [3] and modified to memory mapped access
 ***********************************************************************/

static void a9m9750_nand_command( struct mtd_info* mtd, unsigned command,
				  int column, int page_addr )
{
	register struct nand_chip *this = mtd->priv;

	if( command != NAND_CMD_SEQIN )
		writeb( command, NAND_CMD_ADDR( this ) );
	else {
		if( mtd->oobblock == 256 && column >= 256 ) {
			column -= 256;
			writeb( NAND_CMD_READOOB, NAND_CMD_ADDR( this ) );
			writeb( NAND_CMD_SEQIN, NAND_CMD_ADDR( this ) );
		} else if( mtd->oobblock == 512 && column >= 256 ) {
			if( column < 512 ) {
				column -= 256;
				writeb( NAND_CMD_READ1,NAND_CMD_ADDR( this ) );
				writeb(NAND_CMD_SEQIN, NAND_CMD_ADDR( this ) );
			} else {
				column -= 512;
				writeb(NAND_CMD_READOOB, NAND_CMD_ADDR( this ) );
				writeb(NAND_CMD_SEQIN, NAND_CMD_ADDR( this ));
			}
		} else {
			writeb( NAND_CMD_READ0, NAND_CMD_ADDR( this ) );
			writeb( NAND_CMD_SEQIN, NAND_CMD_ADDR( this ) );
		}
	}

	if( column != -1 || page_addr != -1 ) {
		/* Serially input address */
		if( column != -1 )
			writeb( column, NAND_ADR_ADDR( this ) );
		if( page_addr != -1 ) {
			writeb( (unsigned char) (page_addr & 0xff),
				NAND_ADR_ADDR( this ) );
			writeb( (unsigned char) ((page_addr >> 8) & 0xff),
				NAND_ADR_ADDR( this ) );
			/* One more address cycle for higher density devices */
			if( mtd->size & 0x0c000000 )
				writeb( (unsigned char) ((page_addr >> 16) &0x0f),
					NAND_ADR_ADDR( this ));
		}
	}

	/*
	 * program and erase have their own busy handlers
	 * status and sequential in needs no delay
	*/
	switch( command ) {
	    case NAND_CMD_PAGEPROG:
	    case NAND_CMD_ERASE1:
	    case NAND_CMD_ERASE2:
	    case NAND_CMD_SEQIN:
	    case NAND_CMD_STATUS:
		return;

	    case NAND_CMD_RESET:
		if( this->dev_ready )
			break;
		writeb( NAND_CMD_STATUS, NAND_IO_ADDR( this ) );
		while( !(readb( this->IO_ADDR_R) & 0x40) );
		return;

	/* This applies to read commands */
	default:
		/*
		 * If we don't have access to the busy pin, we apply the given
		 * command delay
		*/
		if( !this->dev_ready ) {
			udelay( this->chip_delay );
			return;
		}
	}

	/* wait until command is processed */
	while( !this->dev_ready );
}

static int __init a9m9750_nand_init( void )
{
	struct nand_chip* this;
	int mtd_parts_nb = 0;
	struct mtd_partition *mtd_parts = 0;
	const char *part_type = 0;
	void __iomem *nand_virt_base;

	/* Allocate memory for MTD device structure and private data */
	a9m9750_mtd = kmalloc( sizeof(struct mtd_info)+sizeof(struct nand_chip),
			       GFP_KERNEL );
	if( !a9m9750_mtd ) {
		printk( "Unable to allocate CC9P9xx0 MTD device structure.\n" );
		return -ENOMEM;
	}

	nand_virt_base = ioremap( PHYS_NAND_FLASH, SZ_1M );
	if( !nand_virt_base ) {
		printk( "ioremap CC9P9xx0 NAND flash failed\n" );
		kfree( a9m9750_mtd );
		return -EIO;
	}
	printk("NAND Flash memory mapped to virtual %p\n", nand_virt_base);

	/* Get pointer to private data */
	this = (struct nand_chip *) ( &a9m9750_mtd[ 1 ] );

	/* Initialize structures */
	memset((char *) a9m9750_mtd, 0, sizeof( struct mtd_info ) );
	memset((char *) this, 0, sizeof( struct nand_chip ) );

	a9m9750_mtd->name = "CC9P9xx0";

	/* Link the private data with the MTD structure */
	a9m9750_mtd->priv = this;

	/* Set address of NAND IO lines */
	this->IO_ADDR_R = nand_virt_base;
	this->IO_ADDR_W = nand_virt_base;
	/* Set address of hardware control function */
	this->hwcontrol = a9m9750_nand_hwcontrol;
	this->cmdfunc   = a9m9750_nand_command;
	/* 15 us command delay time */
	this->chip_delay = 15;
	this->eccmode = NAND_ECC_SOFT;

	/* Scan to find existence of the device */
	if( nand_scan( a9m9750_mtd, 1 ) ) {
		kfree( a9m9750_mtd );
		return -ENXIO;
	}

	/* Allocate memory for internal data buffer */
	this->data_buf = kmalloc(sizeof(u_char)*(a9m9750_mtd->oobblock+
						 a9m9750_mtd->oobsize),
				 GFP_KERNEL);
	if( !this->data_buf ) {
		printk ( "Unable to allocate NAND data buffer for CC9P9xx0.\n" );
		iounmap(  (void*) nand_virt_base );
		kfree( a9m9750_mtd );
		return -ENOMEM;
	}

#ifdef CONFIG_MTD_CMDLINE_PARTS
	mtd_parts_nb = parse_mtd_partitions( a9m9750_mtd, probes, &mtd_parts, 0);
	if (mtd_parts_nb > 0)
		part_type = "command line";
	else
		mtd_parts_nb = 0;
#endif
	if (mtd_parts_nb == 0)
	{
		mtd_parts = partition_info;
		mtd_parts_nb = NUM_PARTITIONS;
		part_type = "static";
	}

	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions( a9m9750_mtd, mtd_parts, mtd_parts_nb);

	/* Return happy */
	return 0;
}
device_initcall( a9m9750_nand_init );

static void __exit a9m9750_cleanup( void )
{
	struct nand_chip *this = (struct nand_chip *) &a9m9750_mtd[1];

	iounmap( (void*) ((struct nand_chip*)a9m9750_mtd[1].priv)->IO_ADDR_R );

	/* Unregister the device */
	del_mtd_device( a9m9750_mtd );

	/* Free internal data buffer */
	kfree( this->data_buf );

	/* Free the MTD device structure */
	kfree( a9m9750_mtd );
}
module_exit( a9m9750_cleanup );

MODULE_LICENSE( "GPL" );
MODULE_AUTHOR( "Markus Pietrek <mpietrek@fsforth.de" );
MODULE_DESCRIPTION( "Board-specific glue layer for NAND flash on CC9P9xx0 module");
