/*
 * drivers/net/arm/ns9xxx_eth.c
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
/***********************************************************************
 *
 * !Revision: $Revision: 1.10 $
 * !Descr: Ethernet driver for the NetSilicon NS9xxx family like NS9750 and
 *         NS9360
 * !References: [1] NS9360 Hardware Reference, Rev. C, 3/2006
 *                  http://ftp1.digi.com/support/documentation/90000675_C.pdf
 *              [2] http://ftp1.digi.com/support/documentation/Ethernet_RX_Lockup_App_Note_073106.pdf
 *              [3] drivers/net/arm/ns9750_eth.c (Deprecated)
 *              [4] drivers/net/pci_skeleton.c
 *              [5] drivers/net/isa_skeleton.c
 *              [6] drivers/net/au1000_eth.c (linux 2.6.19-rc2)
 * !Author: Markus Pietrek 
 *
 ***********************************************************************/
/*
 * !History:  06/12/05 : reserving physical address
 * !History:  07/03/13 : add cc9p9360 support; fix mii clock divisor setting after a MAC reset
 */

#include <linux/config.h>       /* system configuration */

#include <linux/module.h>
#include <linux/pci.h>          /* DMA_TO_DEVICE */
#include <linux/ioport.h>       /* request_mem_region */
#include <linux/netdevice.h>    /* net_device */
#include <linux/etherdevice.h>  /* eth_type_trans */
#include <linux/delay.h>        /* udelay */
#include <linux/ethtool.h>      /* SPEED_100 */
#include <linux/mii.h>          /* mii_ioctl_data */
#include <linux/phy.h>          /* phy_dev */

#include <asm/io.h>
#include <asm/irq.h>            /* enable_interrupt */

#include <asm/arch/system.h>   /* ns9xxx_ahbclock */

#include <asm/io.h>               /* iowrite32 */
#include <asm/arch-ns9xxx/bbus.h> /* set_gpio_cfg */

#include <asm/arch-ns9xxx/ns9xxx_eth.h>

/* ********** Macros ********** */

#define DRIVER_NAME  "ns9xxx_eth"
#define CLEAR(x)     memset( (x), 0, sizeof( *(x) ) )

#define NS_ETH_BASE_VA io_p2v( NS_ETH_BASE_PA )

#define eth_ioread32( offs)        ioread32( NS_ETH_BASE_VA + (offs))
#define eth_iowrite32( val, offs ) iowrite32( val, NS_ETH_BASE_VA + (offs) )
#define eth_rmw32( offs, op )      eth_iowrite32(eth_ioread32(offs) op,(offs))

/* hardware limits sets from ethernet controller */

#define NS_HW_RX_RINGS          (4) /* [1] p. 326 */
#define NS_ALIGN_IP_OFFSET      (2) /* IP Header is aligned on 16byte border */

#define NS_HW_MAX_ETH_FRAME_LEN (1522)
/* [1] p. 316. For NS an ethernet frame is 1522 Bytes. Possible reason:
   An ethernet frame has 1518 Bytes. But we have aligned the
   IP Header on a 16 Byte boundary instead the default 14 Byte for
   performance reasons. This makes additional 2. Why they another 2 I don't
   know */

/* software limits sets by driver */

#define NS_LINK_TIMEOUT         (5*(HZ)) /* in jiffies */
#define NS_WDOG_TIMEOUT         (5*(HZ)) /* in jiffies */

#define NS_TX_DESC              (64) /* maximum number of frames that can be
                                      * put in hardware queue */
#define NS_RX_DESC_RING         (45) /* descriptors per ring should fit one maxsize PING packet */
#define NS_RX_DESC              (NS_HW_RX_RINGS*NS_RX_DESC_RING) /* overall descriptors */

/* Other configuration stuff */

#define NS_POLYNOMIAL           0x4c11db6L /* taken from [1] p393 */
#define NS_LEFT                 0          /* left rotation */
#define NS_RIGHT                1          /* right rotation */

#define NS_TX_INC_DESC(desc)	\
        do { \
                desc++; \
                if( (desc) == NS_TX_DESC ) \
                        desc = 0; \
        } while( 0 )

#define NS_TX_DESC_PTR(desc) (((tx_buffer_desc_t*) ( NS_ETH_BASE_VA + NS_ETH_TXBD )) + (desc) )

/* Debugging stuff, enable it in xconfig Drivers/net/100 MBit/s */

#define ERROR( format, ... ) printk( KERN_ERR DRIVER_NAME " %s: " format "\n", __func__, ##__VA_ARGS__ )

#ifdef CONFIG_ARM_ETHER_NS9XXX_DEBUG
# define DEBUG_INIT          0x0001
# define DEBUG_MINOR         0x0002
# define DEBUG_RX            0x0004
# define DEBUG_TX            0x0008
# define DEBUG_INT           0x0010
# define DEBUG_POLL          0x0020
# define DEBUG_LINK          0x0040
# define DEBUG_MII           0x0100
# define DEBUG_MII_LOW       0x0200
# define DEBUG_MEM           0x0400
# define DEBUG_TIME          0x0800
# define DEBUG_ERROR         0x4000
# define DEBUG_ERROR_CRIT    0x8000

static unsigned int ns_eth_dbg_level = DEBUG_ERROR_CRIT;

# define DEBUG( flag, format, ... )          \
	do { \
                if( ( ns_eth_dbg_level & (flag) ) == (flag) )           \
                        printk( KERN_INFO DRIVER_NAME "@%s: " format "\n", __func__, ##__VA_ARGS__ ); \
        } while( 0 )

# define DEBUG_FN( flag )                       \
        do {      \
                if( ( ns_eth_dbg_level & ( flag ) ) == ( flag ) )       \
                        printk( KERN_INFO DRIVER_NAME "@%s:line %d\n", __func__, __LINE__ ); \
        } while( 0 )

# define ASSERT( expr, func ) \
	do {                          \
                if( !( expr ) ) {                               \
                        ERROR( "Assertion failed! line %d %s",     \
                                __LINE__,#expr);           \
                        func;                                           \
                }                                                       \
        } while( 0 )

# define DEBUG_MARK( mark ) \
        do {                                    \
                if( ns_eth_dbg_level & ( DEBUG_TIME ) ) \
                        ns_hperf_mark( mark );        \
        } while( 0 )
# define DEBUG_MARK_DUMP() \
        do {                                    \
                if( ns_eth_dbg_level & ( DEBUG_TIME ) ) \
                        ns_hperf_marks_dump();             \
        } while( 0 )

/* for timing collection */
enum {
        START_XMIT = 0,
        START_XMIT_DONE,
        TX_INT,
        TX_INT_DONE,
        RX_INT,
        RX_INT_DONE,
        RX_INT_UNMAP_DMA,
        RX_INT_GIVE_KERNEL,
        RX_INT_GIVE_KERNEL_DONE,
        RX_INT_NEW_SKBUFF,
        RX_INT_NEW_SKBUFF_DONE,
        RX_INT_PROC_RING,
        RX_INT_PROC_RING_DONE,
};

# define NAME(x) #x

static const char* ns_hperf_type_names[] = {
        NAME( START_XMIT ),
        NAME( START_XMIT_DONE ),
        NAME( TX_INT ),
        NAME( TX_INT_DONE ),
        NAME( RX_INT ),
        NAME( RX_INT_DONE ),
        NAME( RX_INT_UNMAP_DMA ),
        NAME( RX_INT_GIVE_KERNEL ),
        NAME( RX_INT_GIVE_KERNEL_DONE ),
        NAME( RX_INT_NEW_SKBUFF ),
        NAME( RX_INT_NEW_SKBUFF_DONE ),
        NAME( RX_INT_PROC_RING ),
        NAME( RX_INT_PROC_RING_DONE ),
};
/* ns_hperf_type_names needs to be define first before including */
# include <asm/arch-ns9xxx/ns9xxx_hperf.h>

#else /* CONFIG_ARM_ETHER_NS9XXX_DEBUG */
# define DEBUG(    flag, format, ... ) do {} while( 0 )
# define DEBUG_FN( flag )              do {} while( 0 )
# define ASSERT(   expr, func )        do {} while( 0 )
# define DEBUG_MARK( mark ) 	       do {} while( 0 )
# define DEBUG_MARK_DUMP() 	       do {} while( 0 )
#endif /* CONFIG_ARM_ETHER_NS9XXX_DEBUG */

/* ********** Type Definitions ********** */

/**
 * struct rx_buffer_desc_t - Rx DMA Buffer Description
 * @pdest: physical destination address (where incoming data is stored)
 *             NS calls it source address in the docu
 * @size:      maximum buffer size when not full, amount valid data when full.
 *             Only lower 11bits are used
 * @unused:    NS calls it destination address in their docu
 * @status:    Status of buffer
 *
 * Buffer descriptors taken from [1] p.315
 */
typedef struct {
        u32  pdest; /* NS calls it source address in documentation */
        u32  len;      /* only lower 11 bits used */
        u32* unused;
        union {
                unsigned int reg;
                struct {
                        unsigned status : 16;
                        unsigned res : 12;
                        unsigned full : 1;
                        unsigned enable : 1;
                        unsigned intr : 1;
                        unsigned wrap : 1;
                } bits;
        } s;
} rx_buffer_desc_t;

/**
 * struct tx_buffer_desc - Tx DMA Buffer Description
 * @psrc: Physical address of frame data.
 * @size:     frame/buffer size
 * @unused:   not used
 * @status:   Status of buffer
 *
 * Buffer descriptors taken from [1] p.316
 * no bitfields, because accesses to tx descriptors must be 32bit wide
 */
typedef struct {
        u32  psrc;
        u32  len;
        u32* unused;
        u32  status;
} tx_buffer_desc_t;

/**
 * struct ns_eth_private - Maintains the driver data for the ethernet chip
 * @skb_tx:       points to socket buffers for outgoing frames
 * @skb_rx:       points to socket buffers for incoming frames
 * @dma_tx:       array of dma/phys addresses of outgoing frame data 
 * @dma_rx:       array of dma/phys addresses for incoming frame data
 * @rx_desc:      points to array of all receiving DMA descriptors
 * @tx_first_unprocessed:  start of the tx descriptor lists that are not being
 *                          used by DMA yet
 * @tx_first_dma: first descriptor that is being used by DMA
 * @tx_first_free: first free tx descriptor
 * @recover_rx_ovl_stall_ws: !TODO
 * @lock:         for locking data structures
 */
typedef struct ns_eth_private {
        struct sk_buff* skb_tx[ NS_TX_DESC ];
        struct sk_buff* skb_rx[ NS_RX_DESC ];
        
        dma_addr_t dma_tx[ NS_TX_DESC ];
        dma_addr_t dma_rx[ NS_RX_DESC ];
        
        /* hw descriptors. The first NS_RX_DESC_RING are for ring desc A, then
         * B etc. No multiarray to keep iterators simpler. */
        rx_buffer_desc_t*  rx_desc; /* is [NS_RX_DESC] */
        dma_addr_t         dma_rx_desc;
        
        int                tx_first_unprocessed;
        int                tx_first_dma;
        int                tx_first_free;
        
        struct work_struct recover_rx_ovl_stall_ws; /* may take some while */
        struct net_device_stats stats;
        
         /* for PHY managemenegt */
        struct phy_device* phy_dev;
        struct mii_bus     mii_bus;
        
        spinlock_t         lock_tx;
        spinlock_t         lock_rx;
} ns_eth_private_t;

/* ********** Local Functions ********** */

static int  ns_eth_init_device( struct net_device* dev );
static int  ns_eth_open( struct net_device* dev );
static int  ns_eth_close( struct net_device* dev );
static int  ns_eth_ioctl( struct net_device* dev, struct ifreq* ifr, int cmd );
static int  ns_eth_start_xmit( struct sk_buff* skb,
                               struct net_device* dev );
static irqreturn_t ns_eth_tx_int( int irq, void* dev_id,struct pt_regs* regs );
static irqreturn_t ns_eth_rx_int( int irq, void* dev_id,struct pt_regs* regs );
static void ns_eth_rx_process_ring( struct net_device* dev, int ring );
static int  ns_eth_create_skbuff( struct net_device* dev,
                                  const int descr );
static struct net_device_stats* ns_eth_get_stats( struct net_device* dev );
static void ns_eth_set_multicast_list( struct net_device* dev );
static int  ns_eth_set_mac_address(struct net_device* dev, void* mac);
static void ns_eth_update_mac_address( struct net_device* dev );
static void ns_eth_irqs_disable( void );
static void ns_eth_irqs_enable( void );
static void ns_eth_enable( struct net_device* dev );
static void ns_eth_update_rx_desc( struct net_device* dev, char from_rx_stall);

static void ns_eth_recover_rx_ovl_stall( void* net_dev );

static void ns_eth_update_egcr( struct net_device* dev );

/* the PHY stuff */
static int  ns_mdio_probe( struct net_device* dev );
static void ns_mdio_adjust_link( struct net_device* dev );
static int  ns_mdio_read( struct mii_bus* bus, int phy_id, int regnum );
static int  ns_mdio_write( struct mii_bus* bus, int phy_id, int regnum,
                           u16 val );
static int ns_mdio_reset( struct mii_bus* bus );
static unsigned int ns_mdio_get_clock_divisor( unsigned int max_mdio_clk );
static unsigned int ns_mdio_poll_busy( void );

/* the multicast stuff */

static void ns_eth_load_mca_hash_table( struct dev_mc_list* mc_list );
static void ns_eth_set_hash_bit( /*@inout@*/ unsigned char* hash_table,
                                 int bit );
static int  ns_eth_calculate_hash_bit( const unsigned char* mac );
static unsigned short ns_eth_rotate( unsigned short value,
                                     int direction, int bits );

/* the module stuff */
static int __init  ns_eth_init_module( void );
static void __exit ns_eth_cleanup_module( void );

/* ********** Local Variables ********** */

static const char* ns_eth_version =
    "Ethernet: " DRIVER_NAME " driver $Revision: 1.10 $\n";

static struct net_device ns_eth_device = {
    init:               ns_eth_init_device,
    open:               ns_eth_open,
    do_ioctl:           ns_eth_ioctl,
    stop:               ns_eth_close,
    hard_start_xmit:    ns_eth_start_xmit,
    get_stats:          ns_eth_get_stats,
    set_multicast_list: ns_eth_set_multicast_list,
    set_mac_address:    ns_eth_set_mac_address,
    
    watchdog_timeo:     NS_WDOG_TIMEOUT,
    base_addr:          NS_ETH_BASE_VA,
    features:           0, /* scatter/gather makes only sense with HW_CHECKSUM which we don't have */
    flags:              ( IFF_MULTICAST | IFF_BROADCAST )
};

static const int ns_eth_used_gpios[] = {
    50 /* MII managemenet */,
    51 /* MII rx data valid */,
    52 /* MII rx error */,
    53 /* MII rx data bit 0 */,
    54 /* MII rx data bit 1 */,
    55 /* MII rx data bit 2 */,
    56 /* MII rx data bit 3 */,
    57 /* MII tx enable */,
    /* 58 not used: MII TX error */
    59 /* MII tx data bit 0 */,
    60 /* MII tx data bit 1 */,
    61 /* MII tx data bit 2 */,
    62 /* MII tx data bit 3 */,
    63 /* MII tx collision */,
    64 /* MII carrier sense */,
    /* 65 not used: MII enet PHY interrupt */
};

/* ********** Local Function Implementations ********** */

/**
 * ns_eth_init_device - initializes the net_device structure
 * @dev: device to initialize
 *
 * @return: 0 on success (always)
 */
static int ns_eth_init_device( struct net_device* dev )
{
        DEBUG_FN( DEBUG_INIT );

        /* we are an an ethernet device */
        ether_setup( dev );

        /* U-Boot is required to have set the MAC Address */
        dev->dev_addr[ 5 ] = (eth_ioread32( NS_ETH_SA1 ) >> 8) & 0xff;
        dev->dev_addr[ 4 ] = (eth_ioread32( NS_ETH_SA1 )) & 0xff;
        dev->dev_addr[ 3 ] = (eth_ioread32( NS_ETH_SA2 ) >> 8) & 0xff;
        dev->dev_addr[ 2 ] = (eth_ioread32( NS_ETH_SA2 )) & 0xff;
        dev->dev_addr[ 1 ] = (eth_ioread32( NS_ETH_SA3 ) >> 8) & 0xff;
        dev->dev_addr[ 0 ] = (eth_ioread32( NS_ETH_SA3 )) & 0xff;

        return 0;
}

/**
 * ns_eth_open - Allocates all resources and opens and resets ethernet device
 * @dev: device to open
 *
 * @return: 0 on ok, 1, if no memory is available
 */
static int ns_eth_open( struct net_device* dev )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        int ret = -EBUSY;
        int i;
        
        DEBUG_FN( DEBUG_INIT );
        
        CLEAR( priv );

        /* reset configuration from U-Boot or previous "ifconfig up" */
        
        eth_iowrite32( 0x0, NS_ETH_EGCR1 );
        eth_iowrite32( 0x0, NS_ETH_EINTREN );
        /* acknowledge any pending interrupt flags from U-Boot */
        eth_iowrite32( eth_ioread32( NS_ETH_EINTR ), NS_ETH_EINTR );

        /* request resources */
        if( NULL == request_mem_region( NS_ETH_BASE_PA,
                                ( NS_ETH_TXBD + 0x400 /*Tx Buffer Descriptors*/ ),
                                DRIVER_NAME ) )
                goto error_region;

        if( request_irq( IRQ_ETHRX, ns_eth_rx_int, 0, DRIVER_NAME " rx", dev ) )
                goto error_irq_rx;

        if( request_irq( IRQ_ETHTX, ns_eth_tx_int, 0, DRIVER_NAME " tx", dev ) )
                goto error_irq_tx;
        /* interrupts are now enabled on the PIC, but ethernet controller has
           them disabled. Disable handling of interrupts until everything is
           set up. */
        ns_eth_irqs_disable();
        
        /* initialize MII structure */

        priv->mii_bus.priv  = dev;
        priv->mii_bus.read  = ns_mdio_read;
        priv->mii_bus.write = ns_mdio_write;
        priv->mii_bus.reset = ns_mdio_reset;
        priv->mii_bus.name  = "ns9xxx_mii";
        priv->mii_bus.id    = 0;
	priv->mii_bus.phy_mask = NS_ETH_PHY_MASK;
        priv->mii_bus.irq   = kmalloc( sizeof( int ) * PHY_MAX_ADDR, GFP_KERNEL );
        for( i = 0; i < PHY_MAX_ADDR; i++ )
                priv->mii_bus.irq[ i ] = PHY_POLL;
        mdiobus_register( &priv->mii_bus );
        
        /* configure kernel structures */
        
        spin_lock_init( &priv->lock_tx );
        spin_lock_init( &priv->lock_rx );

        INIT_WORK( &priv->recover_rx_ovl_stall_ws,
                   ns_eth_recover_rx_ovl_stall,
                   priv );
        
        priv->rx_desc = (rx_buffer_desc_t*)
                dma_alloc_coherent( NULL,
                                    sizeof( rx_buffer_desc_t ) * NS_RX_DESC,
                                    &priv->dma_rx_desc, GFP_DMA );
        if( NULL == priv->rx_desc ) {
                ret = -ENOMEM;
                goto error_nomem_rx;
        }

        /* preallocate any receive skb buffers for DMA transfer */
        for( i = 0; i < NS_RX_DESC; i++ ) {
                if( !ns_eth_create_skbuff( dev, i ) ) {
                        ret = -ENOMEM;
                        goto error;
                }
        }

        /* Reset MAC */
        eth_rmw32( NS_ETH_EGCR1, | NS_ETH_EGCR1_MAC_HRST );
        wmb();
        udelay( 5 );            /* according to [1], p.322 */
        eth_rmw32( NS_ETH_EGCR1, & ~NS_ETH_EGCR1_MAC_HRST );
        wmb();
        
	/* have to reset phy after MAC reset, to set correct clock divisor */
	ns_mdio_reset( &priv->mii_bus );
        if( !ns_mdio_probe( dev ) ) {
                ret = -ENODEV;
                goto error;
        }

        /* reset interrupt status register */
        eth_iowrite32( NS_ETH_EINTR_RX_MA | NS_ETH_EINTR_TX_MA, NS_ETH_EINTR );

        /* all ok, continue with hardware initialization
           from here we have access to ports */

        /* assign hardware address */

        ns_eth_update_mac_address( dev );

        /* Receive broadcasts, so we can see ARP requests */
        eth_rmw32( NS_ETH_SAFR, | NS_ETH_SAFR_BROAD );

        ns_eth_enable( dev );
        ns_eth_irqs_enable();
        
        phy_start( priv->phy_dev );

        /* As of 2.6.17 the phy state machine is triggered only every
           second. In the initialization this may lead to a few seconds until
           auto-negotiation has completed -> affecting nfsroot. Try to do the
           first steps faster. */
        /* !TODO. Has this improved in newer kernels */
#ifdef MODULE
	/*  when loading a module, someone maybe immediately starts to send
            packages. With an uninitialized PHY, this may fail. So we wait a
            little bit. */
        i = 0;
        while( i < 2000 ) {
                priv->phy_dev->phy_timer.function( priv->phy_dev->phy_timer.data );
                if( PHY_RUNNING == priv->phy_dev->state )
                        /* autonegotiation has been completed  */
                        break;
                i++;
                mdelay( 1 );
        }
#else
	/* The kernel still does a little bit of work until the network is
           used.. Don't wait for anything */
        priv->phy_dev->phy_timer.function( priv->phy_dev->phy_timer.data );
        priv->phy_dev->phy_timer.function( priv->phy_dev->phy_timer.data );
#endif /* MODULE */        

        netif_start_queue( dev );

        return 0;

error:
        /* free skb buffers */
        for( i = 0; i < NS_RX_DESC; i++ ) {
                if( NULL != priv->skb_rx[ i ] ) {
                        dev_kfree_skb( priv->skb_rx[ i ] );
                        priv->skb_rx[ i ] = NULL;
                } else
                        /* no more buffers allocated */
                        break;
        }

        dma_free_coherent( NULL, sizeof( rx_buffer_desc_t ) * NS_RX_DESC,
                           priv->rx_desc, priv->dma_rx_desc );
        priv->rx_desc     = NULL;
        priv->dma_rx_desc = 0;

error_nomem_rx:
        free_irq( IRQ_ETHTX, dev );
error_irq_tx:
        free_irq( IRQ_ETHRX, dev );
error_irq_rx:
        release_mem_region( NS_ETH_BASE_PA,
                            NS_ETH_TXBD + 0x400 );
error_region:
        return ret;
}

/**
 * ns_eth_close - Closes the ethernet device and frees all resources
 * @dev: device to close
 *
 * @return: always 0 (success)
 */
static int ns_eth_close( struct net_device* dev )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        int i;

        DECLARE_WAIT_QUEUE_HEAD( sleepy );

        DEBUG_FN( DEBUG_INIT );

        ns_eth_irqs_disable();

        netif_stop_queue( dev ); /* can't transmit any more */

        /* wait until all workqueues have finished */
        wait_event_timeout( sleepy,
                cancel_delayed_work( &priv->recover_rx_ovl_stall_ws ), 1 );

        phy_stop( priv->phy_dev );
        phy_disconnect( priv->phy_dev );
        mdiobus_unregister( &priv->mii_bus );
        
        /* disable hardware */

        eth_rmw32( NS_ETH_MAC1, & ~NS_ETH_MAC1_RXEN );
        eth_rmw32( NS_ETH_EGCR1, & ~(NS_ETH_EGCR1_ERX |
                        NS_ETH_EGCR1_ERXDMA |
                        NS_ETH_EGCR1_ETX |
                        NS_ETH_EGCR1_ETXDMA ) );

        /* release ressources */
        free_irq( IRQ_ETHTX, dev );
        free_irq( IRQ_ETHRX, dev );
        release_mem_region( NS_ETH_BASE_PA,
                            NS_ETH_TXBD + 0x400 );

        /* free DMA buffers */
        for( i = 0; i < NS_RX_DESC; i++ ) {
                dev_kfree_skb( priv->skb_rx[ i ] );
                priv->skb_rx[ i ] = NULL;
        }

        for( i = 0; i < NS_TX_DESC; i++ )
                if( NULL != priv->skb_tx[ i ] ) {
                        /* we won't transmit it any longer */
                        dma_unmap_single( NULL, priv->dma_tx[ i ],
                                          priv->skb_tx[ i ]->len,
                                          DMA_TO_DEVICE );
                        dev_kfree_skb( priv->skb_tx[ i ] );
                        priv->dma_tx[ i ] = 0;
                        priv->skb_tx[ i ] = NULL;
                }

        dma_free_coherent( NULL, sizeof( rx_buffer_desc_t ) * NS_RX_DESC,
                           priv->rx_desc, priv->dma_rx_desc );
        priv->rx_desc     = NULL;
        priv->dma_rx_desc = 0;

        return 0;
}

/**
   ns_eth_ioctl - handles ethtools/miitools requests
   * @dev: device to handle
   * @ifr: interface request
   * @cmd: ioctl command
   */
static int ns_eth_ioctl( struct net_device* dev, struct ifreq* ifr, int cmd )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;

        if( !netif_running( dev ) )
                return -EINVAL;

        return phy_mii_ioctl( priv->phy_dev, if_mii( ifr ), cmd );
}


/**
 * ns_eth_start_xmit - Places a buffer in the DMA descriptor lists for
                       transmission
 * @skb: socket buffer to be transmitted
 * @dev: device where to send to
 *
 * @return: 0 on ok, -EBUSY on failure
 */
static int ns_eth_start_xmit( struct sk_buff* skb, struct net_device* dev )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        unsigned long     flags;
        tx_buffer_desc_t* tx_buffer_desc;
        int               dma_is_running = 
                ( eth_ioread32( NS_ETH_EGCR1 ) & NS_ETH_EGCR1_ETXDMA );

        DEBUG_FN( DEBUG_TX );

        /* we are modifying the descriptors */
        spin_lock_irqsave( &priv->lock_tx, flags );

        DEBUG_MARK( START_XMIT );

        if( dma_is_running && (priv->tx_first_dma == priv->tx_first_free) ) {
                ERROR( "outgoing TX queue full, stopping device" );
                netif_stop_queue( dev );
                goto error;
        }

        /* check that tx_first_free_descr doesn't point to a buffer in use */
        ASSERT( priv->skb_tx[ priv->tx_first_free ] == NULL, );

        /* store buffer information */
        priv->skb_tx[ priv->tx_first_free ] = skb;

        priv->dma_tx[ priv->tx_first_free ] = dma_map_single( NULL, skb->data,
                                                              skb->len,
                                                              DMA_TO_DEVICE );
        /* configure next free buffer descriptor */
        tx_buffer_desc = NS_TX_DESC_PTR( priv->tx_first_free );

        tx_buffer_desc->psrc = priv->dma_tx[priv->tx_first_free];
        tx_buffer_desc->len  = (skb->len >= ETH_ZLEN ? skb->len : ETH_ZLEN);

        tx_buffer_desc->status = 
                NS_ETH_TXBD_FULL |
                NS_ETH_TXBD_LAST;

        if( ( NS_TX_DESC - 1 ) == priv->tx_first_free )
                /* last descriptor needs wraparound */
                tx_buffer_desc->status |= NS_ETH_TXBD_WRAP;

        NS_TX_INC_DESC( priv->tx_first_free );

        if( !dma_is_running ) {
                /* interrupt on complete */
                tx_buffer_desc = NS_TX_DESC_PTR( priv->tx_first_free );
                /* wait on idle */
                tx_buffer_desc->status = 0;

                NS_TX_INC_DESC( priv->tx_first_free );
                ASSERT( priv->skb_tx[ priv->tx_first_free ] == NULL, );

                priv->tx_first_dma         = priv->tx_first_unprocessed;
                priv->tx_first_unprocessed = priv->tx_first_free;

                /* DMA is not yet running, trigger it*/
                eth_iowrite32( priv->tx_first_dma * 4, NS_ETH_TXPTR );
                eth_iowrite32( priv->tx_first_dma * 4, NS_ETH_TXRPTR );
                eth_rmw32( NS_ETH_EGCR1, | NS_ETH_EGCR1_ETXDMA );
                eth_rmw32( NS_ETH_EGCR2, & ~NS_ETH_EGCR2_TCLER );
                eth_rmw32( NS_ETH_EGCR2, | NS_ETH_EGCR2_TCLER );
        }
        
        dev->trans_start = jiffies;

        DEBUG_MARK( START_XMIT_DONE );

        spin_unlock_irqrestore( &priv->lock_tx, flags );

        return 0;

  error:
        spin_unlock_irqrestore( &priv->lock_tx, flags );
        return -EBUSY;
}

/**
 * ns_eth_tx_int - Handles a transmitter interrupt
 * @irq   : irq number
 * @dev_id: net_device
 * @regs  : register set
 */
static irqreturn_t ns_eth_tx_int( int irq, void* dev_id, struct pt_regs* regs )
{
        struct net_device* dev = (struct net_device*) dev_id;
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        unsigned int status;
        int i;
        tx_buffer_desc_t* tx_buffer_desc;
        char transmitted = 0;
        int  first_not_transmitted = -1;
        char kick_dma = 0;
        
        DEBUG_FN( DEBUG_TX | DEBUG_INT );

        spin_lock( &priv->lock_tx );

        DEBUG_MARK( TX_INT );
        /* acknowledge interrupt */
        status = eth_ioread32( NS_ETH_EINTR ) & NS_ETH_EINTR_TX_MA;
        eth_iowrite32( status, NS_ETH_EINTR );

        if( status & ( NS_ETH_EINTR_TXERR |
                       NS_ETH_EINTR_TXBUFNR ) )
                printk( KERN_WARNING "%s: error 0x%x\n", __func__, status);
        
        /* check each frame whether it has been sent */
        i = priv->tx_first_dma;
        tx_buffer_desc = NS_TX_DESC_PTR( i );
        /* frames sent */
        while( tx_buffer_desc->status ) {
                if( NULL != priv->skb_tx[ i ] ) {
                        if( !(tx_buffer_desc->status & NS_ETH_TXBD_FULL ) ) {
                                /* we may have had untransmitted the last round */
                                if( !(tx_buffer_desc->status & NS_ETH_ETSR_TXOK ) ) {
                                        ERROR(  "%s: descr %i empty but not "
                                                "transmitted ok (0x%x)",
                                                __func__, i, tx_buffer_desc->status );
                                }
                                
                                /* buffer sent */
                                dma_unmap_single( NULL, priv->dma_tx[ i ],
                                                  priv->skb_tx[ i ]->len,
                                                  DMA_TO_DEVICE );
                                dev_kfree_skb_irq( priv->skb_tx[ i ] );
                                priv->dma_tx[ i ] = 0;
                                priv->skb_tx[ i ] = NULL;
                                transmitted = 1;
                        } else if( first_not_transmitted == -1 ) {
                              DEBUG( DEBUG_ERROR, "retrying %i (stat=0x%x)\n",
                                i, tx_buffer_desc->status );
                                kick_dma = 1;
                                first_not_transmitted = i;
                                break;
                        }
                }
                NS_TX_INC_DESC( i );
                tx_buffer_desc = NS_TX_DESC_PTR( i );
        }

        if( !kick_dma ) {
                if( !tx_buffer_desc->status ) {
                        if( priv->tx_first_unprocessed != priv->tx_first_free ) {
                                /* there is a new list of frames to be sent */
                                /* skip empty */
                                NS_TX_INC_DESC( i );
                                first_not_transmitted = priv->tx_first_dma = i;
                                kick_dma = 1;

                                tx_buffer_desc = NS_TX_DESC_PTR( priv->tx_first_free );
                                /* mark it for idle */
                                /* we don't have NS_TX_DESC descriptors in
                                   use as we have completed here */
                                tx_buffer_desc->status = 0;
                                NS_TX_INC_DESC( priv->tx_first_free );
                                priv->tx_first_unprocessed = priv->tx_first_free;
                        } else
                                /* we sent all, wind descriptors back to first and
                                   stop DMA */
                                eth_rmw32( NS_ETH_EGCR1, & ~NS_ETH_EGCR1_ETXDMA );
                } else {
                        /* shouldn't happend */
                        ERROR( "Status: 0x%x\n", tx_buffer_desc->status );
                }

                if( netif_queue_stopped( dev ) )
                        netif_wake_queue( dev );
        }

        if( kick_dma ) {
                eth_iowrite32( first_not_transmitted * 4, NS_ETH_TXRPTR );
                eth_rmw32( NS_ETH_EGCR2, & ~NS_ETH_EGCR2_TCLER );
                eth_rmw32( NS_ETH_EGCR2, | NS_ETH_EGCR2_TCLER );
        }
        
        DEBUG_MARK( TX_INT_DONE );

        spin_unlock( &priv->lock_tx );

        return IRQ_HANDLED;
}

/**
 * ns_eth_rx_int - Handles a receiver interrupt
 * @irq: irq number
 * @dev_id: net_device
 * @regs: register set
 */
static irqreturn_t ns_eth_rx_int( int irq, void* dev_id,struct pt_regs* regs )
{
        struct net_device* dev = (struct net_device*) dev_id;
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        unsigned int status;
        int ring;
        int mask;
        
        DEBUG_FN( DEBUG_RX | DEBUG_INT );

        spin_lock( &priv->lock_rx );

        DEBUG_MARK( RX_INT );

        /* acknowledge interrupt */
        status = eth_ioread32( NS_ETH_EINTR ) & NS_ETH_EINTR_RX_MA;
        eth_iowrite32( status, NS_ETH_EINTR );

        /* check for errors */
        if( status & ( NS_ETH_EINTR_RXNOBUF |
                       NS_ETH_EINTR_RXBUFFUL |
                       NS_ETH_EINTR_RXOVL_DATA |
                       NS_ETH_EINTR_RXOVL_STAT ) ) {
                ERROR( "rx error status: 0x%x", status );
                        
                if( status & NS_ETH_EINTR_RXOVL_DATA ) {
                        /* [2], disable receiver */
                        eth_rmw32( NS_ETH_EGCR1,
                                & ~(NS_ETH_EGCR1_ERX | NS_ETH_EGCR1_ERXDMA ) );
                        eth_rmw32( NS_ETH_MAC1, & ~NS_ETH_MAC1_RXEN );
                        eth_rmw32( NS_ETH_EINTR, | NS_ETH_EINTR_RXOVL_DATA );

                        /* reinit receiver.
                           recovering may take an undefined time until HW has
                           reloaded DMA descriptors. Therefore we do it in a
                           workqueue and not in a tasklet. */
                        schedule_work( &priv->recover_rx_ovl_stall_ws );
                }
        }
                        
        /* process received frames */
        mask = NS_ETH_EINTR_RXDONEA;
        for( ring = 0; ring < NS_HW_RX_RINGS; ring++ ) {
                if( status & mask ) {
                        ns_eth_rx_process_ring( dev, ring );
                        eth_iowrite32( 1 << ring, NS_ETH_RXFREE );
                }

                mask >>= 1;
        }

        DEBUG_MARK( RX_INT_DONE );

        spin_unlock( &priv->lock_rx );

        return IRQ_HANDLED;
}

/**
 * ns_eth_rx_process_ring - iterates the ring descriptors and give rx frames to
 *                          kernel. Also allocates new buffer for next frame.
 * @dev:  device where frame is coming from
 * @ring: DMA ring that has unprocessed frames
 *
 * Must be called from IRQ, thus no locks
 */
static void ns_eth_rx_process_ring( struct net_device* dev, int ring )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        rx_buffer_desc_t* desc;
        int i;
        int end_ring = (ring+1)*NS_RX_DESC_RING;
       
        DEBUG_FN( DEBUG_RX | DEBUG_INT | DEBUG_MINOR );
        DEBUG_MARK( RX_INT_PROC_RING );

        desc = &priv->rx_desc[ ring * NS_RX_DESC_RING ];
        
        for( i = ring * NS_RX_DESC_RING; i < end_ring; i++, desc++ ) {
                if( desc->s.bits.full ) {
                        /* give kernel the frame */
                        struct sk_buff* skb = priv->skb_rx[ i ];

                        DEBUG_MARK( RX_INT_UNMAP_DMA );
                        /* so CPU can use it */
                        dma_unmap_single( NULL, desc->pdest, desc->len,
                                          DMA_FROM_DEVICE );

                        skb_put( skb, desc->len - 4 /* we need no CRC */ );
                        skb->protocol = eth_type_trans( skb, dev );
                        skb->ip_summed = CHECKSUM_HW;

                        /* align IP header to 16byte */
                        skb_reserve( skb, 2 );

                        DEBUG_MARK( RX_INT_GIVE_KERNEL );
                        /* let the kernel handle the frame now */
                        netif_rx( skb );
                        DEBUG_MARK( RX_INT_GIVE_KERNEL_DONE );

                        dev->last_rx = jiffies;

                        DEBUG_MARK( RX_INT_NEW_SKBUFF );
                        /* allocate a new buffer */
                        ns_eth_create_skbuff( dev, i );
                        DEBUG_MARK( RX_INT_NEW_SKBUFF_DONE );
                }
        }
        DEBUG_MARK( RX_INT_PROC_RING_DONE );
}

/**
 * ns_eth_recover_rxovl_stall - Handles a receiver overflow, see [2]
 * @dev: device to recover
 *
 * Should not be called in IRQ context as ns_eth_update_rx_desc takes some time
 */
static void ns_eth_recover_rx_ovl_stall( void* net_dev )
{
        struct net_device* dev = (struct net_device*) net_dev;
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        unsigned long      flags;

        DEBUG_FN( DEBUG_RX );
        
        spin_lock_irqsave( &priv->lock_rx, flags );
        /* reinitialize descriptor rings */
        ns_eth_update_rx_desc( dev, 1 );
        eth_rmw32( NS_ETH_MAC1, | NS_ETH_MAC1_RXEN |
                                NS_ETH_EGCR1_ERXDMA );
        spin_unlock_irqrestore( &priv->lock_rx, flags );
}

/**
 * ns_eth_create_skbuff - Allocates a skbuff and configures descriptor for it
 * @dev:   device to recover
 * @descr: descriptor index this skbuff is used on.
 * 
 * @return: 0 on failure, otherwise 1
 */
static int ns_eth_create_skbuff( struct net_device* dev, const int descr )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        struct sk_buff* skb;

        /* it's not in use any longer */
        priv->rx_desc[ descr ].s.bits.full = 0;

        /* allocate skb buffer (can be done in IRQ) */
        priv->skb_rx[ descr ] = skb = dev_alloc_skb( NS_HW_MAX_ETH_FRAME_LEN );
        if( NULL == skb ) {
                printk( KERN_ERR "%s: out of memory\n",
                        __func__ );
                /* disable descriptor */
                priv->rx_desc[ descr ].s.bits.enable = 0;
                return 0;
        }
        
        skb->dev = dev;
        
        /* configure descriptor */
        priv->rx_desc[ descr ].pdest = 
            (unsigned int) dma_map_single( NULL, skb->data, skb->len,
                                           DMA_FROM_DEVICE );

        priv->rx_desc[ descr ].len = NS_HW_MAX_ETH_FRAME_LEN;
        priv->rx_desc[ descr ].s.bits.wrap =
            ( descr == ( NS_RX_DESC_RING - 1 ) ) ? 1 : 0;
        priv->rx_desc[ descr ].s.bits.intr = 1;
        priv->rx_desc[ descr ].s.bits.enable = 1;

        return 1;
}

/**
 * ns_eth_get_stats - Copies controller counters to kernel stats.
 * @dev: device of statistics
 */
static struct net_device_stats* ns_eth_get_stats( struct net_device* dev )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;

        DEBUG_FN( DEBUG_MINOR );

        if( NULL == priv )
                return NULL;
        
        /* !TODO: not sure whether rx_errors and
         * tx_errors also contains sub errors */
        priv->stats.rx_bytes   += eth_ioread32( NS_ETH_STAT_RBYT );
        priv->stats.rx_packets += eth_ioread32( NS_ETH_STAT_RPKT );
        priv->stats.rx_errors  += eth_ioread32( NS_ETH_STAT_RCDE ) +
            eth_ioread32( NS_ETH_STAT_RCSE ) +
            eth_ioread32( NS_ETH_STAT_RUND ) +
            eth_ioread32( NS_ETH_STAT_RFRG ) +
            eth_ioread32( NS_ETH_STAT_RJBR );
        priv->stats.multicast        += eth_ioread32( NS_ETH_STAT_RMCA );
        priv->stats.rx_length_errors +=eth_ioread32( NS_ETH_STAT_RFLR );
        priv->stats.rx_over_errors   +=eth_ioread32( NS_ETH_STAT_ROVR );
        priv->stats.rx_crc_errors    +=eth_ioread32( NS_ETH_STAT_RFCS );
        priv->stats.rx_frame_errors  +=eth_ioread32( NS_ETH_STAT_RALN );
        
        priv->stats.tx_bytes   += eth_ioread32( NS_ETH_STAT_TBYT );
        priv->stats.tx_packets += eth_ioread32( NS_ETH_STAT_TPKT );
        priv->stats.tx_errors  += eth_ioread32( NS_ETH_STAT_TDFR ) +
            eth_ioread32( NS_ETH_STAT_TXCL ) +
            eth_ioread32( NS_ETH_STAT_TJBR ) +
            eth_ioread32( NS_ETH_STAT_TFCS ) +
            eth_ioread32( NS_ETH_STAT_TOVR ) +
            eth_ioread32( NS_ETH_STAT_TUND ) +
            eth_ioread32( NS_ETH_STAT_TFRG );
        priv->stats.collisions        += eth_ioread32( NS_ETH_STAT_TNCL );
        priv->stats.tx_aborted_errors += eth_ioread32( NS_ETH_STAT_TXCL );
        
        return &priv->stats;
}

/**
 * ns_eth_set_multicast_list - Configures the ethernet controller for broadcast/multicast access
 * @dev: device to configure
 */
static void ns_eth_set_multicast_list( struct net_device* dev )
{
        unsigned int unSAFR;

        DEBUG_FN( DEBUG_MINOR );

        unSAFR = eth_ioread32( NS_ETH_SAFR );

        /* disable all multicast modes */
        unSAFR &= ~( NS_ETH_SAFR_PRO |
                     NS_ETH_SAFR_PRM |
                     NS_ETH_SAFR_PRA );

        if( dev->flags & IFF_PROMISC ) {
                DEBUG( DEBUG_MINOR, "Promiscuous mode" );
                unSAFR |= NS_ETH_SAFR_PRO;
        } else if( dev->flags & IFF_ALLMULTI ) {
                DEBUG( DEBUG_MINOR, "Accept all multicast packets" );
                unSAFR |= NS_ETH_SAFR_BROAD |
                          NS_ETH_SAFR_PRM;
        } else if( !(dev->flags & IFF_MULTICAST) ) {
                DEBUG( DEBUG_MINOR,
                        "Accept only direct addressed and broadcast packets" );
                unSAFR |= NS_ETH_SAFR_BROAD;
        } else {
                DEBUG( DEBUG_MINOR,
                        "Accept multicast list" );
                
                /* store multicast list of kernel into chip */
                ns_eth_load_mca_hash_table( dev->mc_list );

                unSAFR |= NS_ETH_SAFR_BROAD |
                          NS_ETH_SAFR_PRA;
        }

        /* write back register */
        eth_iowrite32( unSAFR, NS_ETH_SAFR );
}

/***********************************************************************
 * MULTICAST ADDRESS HASH TABLE FUNCTIONS
 * Taken from netarm_eth.c and [1]
 ***********************************************************************/

/**
 * ns_eth_load_mca_hash_table -
 * @mc_list: multicast address list
 *
 * This function generates the multicast address hash table from the
 * multicast addresses contained in the provided linked list.  This
 * function loads this table into the Ethernet controller.
 */
static void ns_eth_load_mca_hash_table( struct dev_mc_list* mc_list )
{
        unsigned long       hash_table[ 2 ];
        struct dev_mc_list* mc_list_entry;
        int                 bit;

        DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );
        
        /* Generate the hash table. */

        CLEAR( hash_table );

        for( mc_list_entry = mc_list;
             mc_list_entry != NULL;
             mc_list_entry = mc_list_entry->next ) {
            DEBUG( DEBUG_MINOR,
                    "  multicast address:  %X %X %X %X %X %X",
                    mc_list_entry->dmi_addr[0], mc_list_entry->dmi_addr[1],
                    mc_list_entry->dmi_addr[2], mc_list_entry->dmi_addr[3],
                    mc_list_entry->dmi_addr[4], mc_list_entry->dmi_addr[5] );

            bit = ns_eth_calculate_hash_bit( mc_list_entry->dmi_addr );
            ns_eth_set_hash_bit( (unsigned char *) hash_table, bit );
        }
        DEBUG( DEBUG_MINOR, "  hash table:  %lX %lX",
                hash_table[ 0 ], hash_table[ 1 ] );

        /* Load the hash table into the Ethernet controller. */
        eth_iowrite32( hash_table[ 0 ], NS_ETH_HT1 );
        eth_iowrite32( hash_table[ 1 ], NS_ETH_HT2 );
}

/***********************************************************************
 * @Function: ns_set_hash_bit
 * @Return: nothing
 * @Descr: sets a bit in the hash table
 ***********************************************************************/

/** ns_eth_set_hash_bit - sets a bit in the hash table
 * @hash_table - array of chars where to set a bit
 * @bit - bitnumber in the array. It can be >>8, then the byte #(bit/8) is
 *        taken
 */
static void ns_eth_set_hash_bit( /*@inout@*/ unsigned char* hash_table,
                                 int bit )
{
        int byte_index;
        int bit_index;

        byte_index = bit >> 3;
        bit_index  = bit & 7;
        hash_table[ byte_index ] |= (1 << bit_index );
}

/** ns_eth_calculate_hash_bit - calculates the bits to set in the CRC hash table
 * @mac: 6 Byte mac address to calculate hash bit
 */

static int ns_eth_calculate_hash_bit( const unsigned char* mac )
{
        unsigned short  mac_in_short[ 3 ];
        unsigned short  bp;
        unsigned short  bx;
        unsigned long   crc;
        int             result;
        int             mac_word;
        int             bit_index;

        /* mac may not be 16 bit aligned, accesses could fail */
        memcpy( mac_in_short, mac, sizeof( mac_in_short ) );

        crc = 0xffffffffL;

        for( mac_word = 0; mac_word < 3; mac_word++ ) {
                bp = mac_in_short[ mac_word ]; /* get word of address */

                for( bit_index = 0; bit_index < 16; bit_index++ ) {
                        /* get high word of crc */
                        bx  = (unsigned short) (crc >> 16) & 0xffff;
                        bx  = ns_eth_rotate( bx, NS_LEFT, 1 );/*bit 31 to lsb*/
                        bx  ^= bp; /* combine with incoming */
                        crc <<= 1;
                        bx   &= 1;   /* get control bit */
                        if( bx )
                                crc ^= NS_POLYNOMIAL;
                        crc |= bx;
                        bp = ns_eth_rotate( bp, NS_RIGHT, 1 );
                }
        }

        /* CRC calculation done.  The 6-bit result resides in bits locations
           28:23. */

        result = (crc >> 23) & 0x3f;

        return result;
}

/** ns_eth_rotate - rotate an unsigned short a number of bits
 * @value:     value to rotate
 * @direction: direction to rotate
 * @bits:      number of bits to rotate
 * 
 * returns rotate value
 */

static unsigned short ns_eth_rotate( unsigned short value,
                                     int direction,
                                     int bits )
{
        int i;
        
        if( direction == NS_LEFT ) {
                for( i = 0; i < bits; i++ ) {
                        if( value & 0x8000 )
                                value = (value << 1) | 1; /* msb -> lsb*/
                        else
                                value <<= 1;
                }
        } else {                /* if ror */
                for( i = 0; i < bits; i++ ) {
                        if( value & 1 )                 /* lsb -> msb */
                                value = ( value >> 1 ) | 0x8000; 
                        else
                                value >>= 1;
                }
        }

        return value;
}

/** ns_eth_set_mac_address - copies the mac address
 * @dev: device to set mac address
 * @mac: mac address to set
 *
 * Always successeds
 */
static int ns_eth_set_mac_address( struct net_device* dev, void* mac )
{
        const struct sockaddr* addr = mac;

        memcpy( dev->dev_addr, addr, dev->addr_len );

        ns_eth_update_mac_address( dev );

        return 0;
}

/** ns_eth_update_mac_address - copies mac address of device to controller
 * @dev: device to set mac address for
 */
static void ns_eth_update_mac_address( struct net_device* dev )
{
        eth_iowrite32( dev->dev_addr[ 5 ] << 8 |
                dev->dev_addr[ 4 ], NS_ETH_SA1 );
        eth_iowrite32( dev->dev_addr[ 3 ] << 8 |
                dev->dev_addr[ 2 ], NS_ETH_SA2 );
        eth_iowrite32( dev->dev_addr[ 1 ] << 8 |
                dev->dev_addr[ 0 ], NS_ETH_SA3 );
}

/** ns_eth_irqs_disable - disables all ethernet interrupts
 */
static void ns_eth_irqs_disable( void )
{
        DEBUG_FN( DEBUG_INIT );

        disable_irq( IRQ_ETHRX );
        disable_irq( IRQ_ETHTX );
}

/** ns_irqs_enable - enables all ethernet interrupts
 */
static void ns_eth_irqs_enable( void )
{
        DEBUG_FN( DEBUG_INIT );

        enable_irq( IRQ_ETHRX );
        enable_irq( IRQ_ETHTX );
}

/**
 * ns_eth_enable - enable hardware and prepare DMA
 * @dev: device to enable
*/
static void ns_eth_enable( struct net_device* dev )
{
        DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

        /* no locks because used in ns_eth_open */
        /* enable hardware */

        /* enable receive and transmit FIFO, use 10/100 Mbps MII */
        eth_iowrite32( NS_ETH_EGCR1_ETXWM |
                   NS_ETH_EGCR1_ERX |
                   NS_ETH_EGCR1_ETX |
                   NS_ETH_EGCR1_RXALIGN |
                   NS_ETH_EGCR1_PHY_MODE_MII |
                   NS_ETH_EGCR1_ITXA,
                   NS_ETH_EGCR1 );
        wmb();

        /* EINTR can only be reseted with ERX and ETX on */
        eth_rmw32( NS_ETH_EINTR, & 0xffffffff );

        eth_iowrite32( NS_ETH_MAC1_RXEN, NS_ETH_MAC1 );
        
        /* the linux kernel may give packets < 60 bytes, for example arp */
        eth_iowrite32( NS_ETH_MAC2_CRCEN |
                NS_ETH_MAC2_PADEN,
                NS_ETH_MAC2 );

        /* program the first descriptors for each ring. */
        ns_eth_update_rx_desc( dev, 0 );

        eth_rmw32( NS_ETH_EGCR1, | NS_ETH_EGCR1_ERXDMA );

        /* ignore statistics interrupt */
        eth_iowrite32( NS_ETH_EINTR_RXOVL_DATA |
                   NS_ETH_EINTR_RXBUFC |
                   NS_ETH_EINTR_RXBUFC |
                   NS_ETH_EINTR_RXDONEA |
                   NS_ETH_EINTR_RXDONEB |
                   NS_ETH_EINTR_RXDONEC |
                   NS_ETH_EINTR_RXDONED |
                   NS_ETH_EINTR_RXNOBUF |
                   NS_ETH_EINTR_RXBUFFUL |
                   NS_ETH_EINTR_TXBUFNR |
                   NS_ETH_EINTR_TXERR |
                   NS_ETH_EINTR_TXIDLE,
                   NS_ETH_EINTREN );

        /* clear statistics */
        eth_iowrite32( NS_ETH_EGCR2_CLRCNT, NS_ETH_EGCR2 );
        /* enable statistics */
        eth_iowrite32( NS_ETH_EGCR2_AUTOZ | NS_ETH_EGCR2_STEN, NS_ETH_EGCR2 );
}

/**
 * ns_eth_update_rx_desc - copies the receive descriptors to controller
 * @dev: device to use
 *
 * must be called in a locked situation
 */
static void ns_eth_update_rx_desc( struct net_device* dev, char from_rx_stall )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        int timeout = 20;
        int i;

        DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

        for( i = 0; i < NS_HW_RX_RINGS; i++ )
                eth_iowrite32((unsigned int) (priv->dma_rx_desc + (i*NS_RX_DESC_RING) ),
                              NS_ETH_RXAPTR + ( i * 4 ) );

        if( from_rx_stall ) {
                eth_rmw32( NS_ETH_EGSR,  | NS_ETH_EGSR_RXINIT );
                eth_rmw32( NS_ETH_EGCR1, | NS_ETH_EGCR1_ERX );
                wmb();
        }

        eth_rmw32( NS_ETH_EGCR1, | NS_ETH_EGCR1_ERXINIT );
        wmb();

        while( --timeout && !(eth_ioread32( NS_ETH_EGSR ) & NS_ETH_EGSR_RXINIT)) {
                /* wait for finish */
                udelay( 1 );
                rmb();
        }
        eth_rmw32( NS_ETH_EGCR1, & ~NS_ETH_EGCR1_ERXINIT );

        if( !timeout )
                printk( KERN_WARNING DRIVER_NAME
                        ":ethernet descriptor update failed\n" );


        /* all 4 descriptors free */
        eth_iowrite32( 0xf, NS_ETH_RXFREE );
        wmb();
}

/** ns_eth_update_egcr - updates the EGCR registers to the link detected
 */
static void ns_eth_update_egcr( struct net_device* dev )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        unsigned int mac2;
        unsigned int ipgt;

        DEBUG_FN( DEBUG_LINK );
        
        mac2 = eth_ioread32( NS_ETH_MAC2 ) & ~NS_ETH_MAC2_FULLD;
        ipgt = eth_ioread32( NS_ETH_IPGT ) & ~NS_ETH_IPGT_MA;

        /* see [1] p. 344 */
        if( SPEED_100 == priv->phy_dev->speed )
                ipgt |= 0x15;
        else
                ipgt |= 0x12;

        if( DUPLEX_FULL == priv->phy_dev->duplex )
                mac2 |= NS_ETH_MAC2_FULLD;

        eth_iowrite32( mac2, NS_ETH_MAC2 );
}

/***********************************************************************
 * the MII low level stuff
 ***********************************************************************/

/*
 ** ns_mdio_probe - Probes for a PHY
 * @dev: device to probe on
 *
 * !return: 0 on failure otherwise 1
 * !see [6]
 */
static int ns_mdio_probe( struct net_device* dev )
{
        ns_eth_private_t* priv = (ns_eth_private_t*) dev->priv;
        struct phy_device* phy_dev = NULL;
        int phy_addr;
        
        /* search lowest phy */
        for( phy_addr = 0; phy_addr < PHY_MAX_ADDR; phy_addr++ ) {
                if( priv->mii_bus.phy_map[ phy_addr ] ) {
                        phy_dev = priv->mii_bus.phy_map[ phy_addr ];
                        break;
                }
        }

        if( NULL == phy_dev ) {
                ERROR( "no PHY found" );
                return 0;
        }
        
        BUG_ON( phy_dev->attached_dev );

        /* attach to phy_dev */

	phy_dev = phy_attach( dev, phy_dev->dev.bus_id, 0 );

	if( IS_ERR( phy_dev ) ) {
                ERROR( "Could not attach to PHY" );
                return 0;
        }

	phy_prepare_link( phy_dev, ns_mdio_adjust_link );
	phy_start_machine( phy_dev, NULL );

        priv->phy_dev        = phy_dev;

        printk( KERN_INFO DRIVER_NAME ": attached PHY driver [%s] "
                "(mii_bus:phy_addr=%s, irq=%d)\n",
                phy_dev->drv->name, phy_dev->dev.bus_id, phy_dev->irq );

        return 1;
}

/**
 * ns_mdio_adjust_link - The PHY has detected a link change. Update
 *                        controllers registers.
 * @dev: device to work on
 *
 */
static void ns_mdio_adjust_link( struct net_device* dev )
{
        DEBUG_FN( DEBUG_LINK );
        ns_eth_update_egcr( dev );
}

/**
 * ns_mdio_read - reads a byte from the PHY
 *
 */
static int ns_mdio_read( struct mii_bus* bus, int phy_id, int regnum )
{
        DEBUG_FN( DEBUG_MII_LOW );

        /* write MII register to be read */
        eth_iowrite32( phy_id << 8 | regnum, NS_ETH_MADR );
        eth_iowrite32( NS_ETH_MCMD_READ, NS_ETH_MCMD );

        if( !ns_mdio_poll_busy() )
            printk( KERN_WARNING DRIVER_NAME
                    ": MII still busy in read\n" );
        /* continue to read */

        eth_iowrite32( 0, NS_ETH_MCMD );

        return (unsigned short) (eth_ioread32( NS_ETH_MRDD ) );
}

/**
 * ns_mdio_read - writes a byte to the PHY
 *
 */
static int ns_mdio_write( struct mii_bus* bus, int phy_id, int regnum,
                              u16 val )
{
        DEBUG_FN( DEBUG_MII_LOW );

        /* write MII register to be written */
        eth_iowrite32( phy_id << 8 | regnum, NS_ETH_MADR );
        eth_iowrite32( val, NS_ETH_MWTD );

        if( !ns_mdio_poll_busy() )
                ERROR( " MII still busy in write" );

        return 0;
}

/**
 * ns_mdio_reset - resets the PHY
 *
 */
static int ns_mdio_reset( struct mii_bus* bus )
{
        DEBUG_FN( DEBUG_MII );

        /* MDIO access is _NOT_ locked by PHY layer*/
	spin_lock_bh( &bus->mdio_lock );

        eth_rmw32( NS_ETH_MAC1,  & ~NS_ETH_MAC1_SRST );
        wmb();

        eth_iowrite32( ns_mdio_get_clock_divisor( NS_ETH_DEFAULT_MAX_MDIO_CLK ),
                   NS_ETH_MCFG );

	spin_unlock_bh( &bus->mdio_lock );

        return 0;
}

/**
 * ns_mdio_get_clock_divisor
 * @max_mdio_clk: the maximum clock the PHY supports
 *
 * !return the clock divisor in NS_ETH_MCFG_CLKS
 */

static unsigned int ns_mdio_get_clock_divisor( unsigned int max_mdio_clk )
{
        struct {
                unsigned int clk_div;
                unsigned int clks; /* field for NS_ETH_MCFG_CLKS */
        } PHYClockDivisors[] = {
            {  4, NS_ETH_MCFG_CLKS_4 },
            {  6, NS_ETH_MCFG_CLKS_6 },
            {  8, NS_ETH_MCFG_CLKS_8 },
            { 10, NS_ETH_MCFG_CLKS_10 },
            { 20, NS_ETH_MCFG_CLKS_20 },
            { 30, NS_ETH_MCFG_CLKS_30 },
            { 40, NS_ETH_MCFG_CLKS_40 }
        };

        unsigned int clks = NS_ETH_MCFG_CLKS_40; /* defaults to greatest div */
        int i;

        DEBUG_FN( DEBUG_INIT );

        for( i = 0; i < ARRAY_SIZE( PHYClockDivisors ); i++ ) {
                int mdio_clock = ns9xxx_ahbclock() / PHYClockDivisors[i].clk_div;
                /* find first sysclock divisor that isn't higher than 2.5 MHz
                   clock */
                if( mdio_clock <= max_mdio_clk ) {
                        clks = PHYClockDivisors[ i ].clks;
                        DEBUG( DEBUG_INIT, "Real MDIO Clock: %i\n", mdio_clock );
                        break;
                }
        }

        DEBUG( DEBUG_INIT,
               "Taking MDIO Clock bit mask 0x%0x for max clock %i",
               clks,
               max_mdio_clk );

        /* return greatest divisor */
        return clks;
}

/**
 * ns_mdio_poll_busy - Polls the PHY whether a command is complete
 *
 * !return 0 on timeout
 */

static unsigned int ns_mdio_poll_busy( void )
{
        unsigned int timeout = 10000;

        DEBUG_FN( DEBUG_MII_LOW );

        while( ( ( eth_ioread32( NS_ETH_MIND ) & NS_ETH_MIND_BUSY )
                == NS_ETH_MIND_BUSY ) && timeout-- )
                cpu_relax();

        return timeout;
}

/**
 * ns_eth_init_module - initializes the module
 */

static int __init ns_eth_init_module( void )
{
        int i;
        int ret = -EBUSY;
        int gpio;
        
        printk( KERN_INFO "%s", ns_eth_version );

        /* U-Boot is configuring the ethernet GPIOs for us.
           Depending on the board, some may be configured differently. For
           documentation, request all GPIOs U-Boot has configured for us.
           They will be released later. */
        for( i = 0; i < ARRAY_SIZE( ns_eth_used_gpios ); i++ ) {
                gpio = ns_eth_used_gpios[ i ];
                if( ns9xxx_request_gpio( gpio, DRIVER_NAME ) ) {
                        ERROR( "Someone claims gpio %i, but it is configured for ethernet. Please fix it in other driver.", gpio );
                        goto error_gpio;
                }
                set_gpio_cfg( gpio, NS_GPIO_CFG_FUNC_0 );
        }
        
        /* allocate private field */
        ns_eth_device.priv = kzalloc( sizeof( ns_eth_private_t ), GFP_KERNEL );
        if( NULL == ns_eth_device.priv ) {
                ret = -ENOMEM;
                goto error_gpio;
        }

        if( ( ret = register_netdev( &ns_eth_device ) ) != 0 )
                goto error_register;
        
        return 0;

error_register:
        kfree( ns_eth_device.priv );
        ns_eth_device.priv = NULL;

error_gpio:
        /* release requested gpios */
        for( ; i >=0; i-- )
                ns9xxx_release_gpio( gpio );
        
        return ret;
}

/**
 * ns_eth_cleanup_module - removes the module
 */
static void __exit ns_eth_cleanup_module( void )
{
        int i;

        unregister_netdev( &ns_eth_device );

        for( i = 0; i < ARRAY_SIZE( ns_eth_used_gpios ); i++ )
                ns9xxx_release_gpio( ns_eth_used_gpios[ i ] );

        kfree( ns_eth_device.priv );
        ns_eth_device.priv = NULL;
}

module_init( ns_eth_init_module );
module_exit( ns_eth_cleanup_module );

#ifdef CONFIG_ARM_ETHER_NS9XXX_DEBUG
module_param( ns_eth_dbg_level, int, 000 );
MODULE_PARM_DESC( ns_eth_dbg_level, "debug output level" );
#endif /* CONFIG_ARM_ETHER_NS9XXX_DEBUG */

MODULE_AUTHOR( "Markus Pietrek <mpietrek@fsforth.de>" );
MODULE_DESCRIPTION( "Ethernet driver for NS9xx0" );
MODULE_LICENSE( "GPL" );

