/*
 * Copyright (C) 2005 Stephen Street / StreetFire Sound Labs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/spi.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/kfifo.h>

#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>

MODULE_AUTHOR("Stephen Street");
MODULE_DESCRIPTION("PXA Loopback SPI Protocol Driver");
MODULE_LICENSE("GPL");

#define BUF_SIZE (1024)

struct device_data {
	char name[BUS_ID_SIZE];
	struct miscdevice misc_device;
	struct semaphore driver_lock;
	struct list_head device_entry;
	atomic_t busy;
	spinlock_t read_lock;
	spinlock_t write_lock;
	struct kfifo *read_fifo;
	struct kfifo *write_fifo;
	wait_queue_head_t read_available;
	wait_queue_head_t write_available;
	struct spi_message msg;
	struct spi_transfer transfer;
	void *dma_buf;
};

static void post(void *context);

static LIST_HEAD(device_list);
static DECLARE_MUTEX(device_list_lock);

static inline struct device_data *find_dev_data(int minor)
{
	struct list_head *ptr;
	struct device_data *entry;
	
	down(&device_list_lock);
	
	list_for_each(ptr, &device_list) {
		entry = list_entry(ptr, struct device_data, device_entry);
		if (minor == entry->misc_device.minor) {
			up(&device_list_lock);
			return entry;
		}
	}
	
	up(&device_list_lock);
	return NULL;
}

static inline void add_device_data(struct device_data *dev_data)
{
	down(&device_list_lock);
	
	list_add(&dev_data->device_entry, &device_list);
	
	up(&device_list_lock);
}

static inline void remove_device_data(struct device_data *dev_data)
{
	down(&device_list_lock);
	
	list_del(&dev_data->device_entry);
	
	up(&device_list_lock);
}

static void kick(struct device_data *dev_data)
{
	unsigned int size;
	int status = 0;
	
	/* Make sure we are not busy */	
	if (!atomic_inc_and_test(&dev_data->busy)) {
		atomic_dec(&dev_data->busy);
		return;
	}
	
	/* Check for available data */
	size = kfifo_len(dev_data->write_fifo);
	if (!size) {
		atomic_dec(&dev_data->busy);
		return;
	}
	
	/* Read the data to send */
	if (kfifo_get(dev_data->write_fifo, dev_data->dma_buf, size) != size) {
		dev_err(dev_data->misc_device.dev, "write fifo failure\n");
		atomic_dec(&dev_data->busy);
		return;
	}
	
	/* Send async spi message */
	memset(&dev_data->msg, 0, sizeof(struct spi_message));
	memset(&dev_data->transfer, 0, sizeof(struct spi_transfer));
	dev_data->transfer.tx_buf = dev_data->dma_buf;
	dev_data->transfer.rx_buf = dev_data->dma_buf;
	dev_data->transfer.len = size;
	dev_data->msg.transfers = &dev_data->transfer;
	dev_data->msg.n_transfer = 1;
	dev_data->msg.complete = post;
	dev_data->msg.context = dev_data;
	status = spi_async(to_spi_device(dev_data->misc_device.dev), 
				&dev_data->msg);
	if (status < 0) {
		dev_err(dev_data->misc_device.dev, 
				"spi failure (%d)\n", 
				status);
		atomic_dec(&dev_data->busy);
	}

	wake_up_interruptible(&dev_data->write_available);
}

static void post(void *context)
{
	struct device_data *dev_data = (struct device_data *)context;
	unsigned int space;
	unsigned length;
	unsigned char local_buf[BUF_SIZE];
	
	/* Check spi status */
	if (dev_data->msg.status == 0) {

		/* Make room for new data */
		length = dev_data->msg.actual_length;
		space = BUF_SIZE - kfifo_len(dev_data->read_fifo);
		if (space < length)
			kfifo_get(dev_data->read_fifo, 
					local_buf, 
					length - space);

		/* Add the new data */
		if (kfifo_put(dev_data->read_fifo, dev_data->dma_buf, length) 
				!= length) {
			dev_err(dev_data->misc_device.dev, 
					"read fifo post failure\n");
		}
		
		/* Let readers know */
		wake_up_interruptible(&dev_data->read_available);
	}
	else
		dev_err(dev_data->misc_device.dev, 
				"spi failure (%d)\n", 
				dev_data->msg.status);
	
	atomic_dec(&dev_data->busy);
	kick(dev_data);
}

static ssize_t read(struct file *filp, 
			char __user *buf, 
			size_t count, 
			loff_t *f_pos)
{
	struct device_data *dev_data = (struct device_data *)filp->private_data;
	char local_buf[BUF_SIZE];
	size_t read_size;
	
	if (down_interruptible(&dev_data->driver_lock))
		return -ERESTARTSYS;
		
	while (kfifo_len(dev_data->read_fifo) == 0) {
		up(&dev_data->driver_lock);
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		if (wait_event_interruptible(dev_data->read_available, 
				(kfifo_len(dev_data->read_fifo) != 0)))
			return -ERESTARTSYS;
		if (down_interruptible(&dev_data->driver_lock))
			return -ERESTARTSYS;
	}
	
	read_size = min(count, (size_t)BUF_SIZE);
	read_size = min(read_size, kfifo_len(dev_data->read_fifo));
	
	if (kfifo_get(dev_data->read_fifo, local_buf, read_size) != read_size) {
		dev_err(dev_data->misc_device.dev, "read fifo failure");
		up(&dev_data->driver_lock);
		return -EIO;
	}
	
	if (copy_to_user(buf, local_buf, read_size)) {
		dev_err(dev_data->misc_device.dev, "user space failure");
		up(&dev_data->driver_lock);
		return -EFAULT;
	}
	
	up(&dev_data->driver_lock);
		
	return read_size;
}

static ssize_t write(struct file *filp, 
			const char __user *buf, 
			size_t count, 
			loff_t *f_pos)
{
	struct device_data *dev_data = (struct device_data *)filp->private_data;
	char local_buf[BUF_SIZE];
	size_t write_size;
	
	if (down_interruptible(&dev_data->driver_lock))
		return -ERESTARTSYS;
		
	while (kfifo_len(dev_data->write_fifo) == BUF_SIZE) {
		up(&dev_data->driver_lock);
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		if (wait_event_interruptible(dev_data->read_available, 
				(kfifo_len(dev_data->write_fifo) < BUF_SIZE)))
			return -ERESTARTSYS;
		if (down_interruptible(&dev_data->driver_lock))
			return -ERESTARTSYS;
	}
	
	write_size = min(count, (size_t)BUF_SIZE);
	write_size = min(write_size, 
				BUF_SIZE - kfifo_len(dev_data->write_fifo));
	
	if (copy_from_user(local_buf, buf, write_size)) {
		dev_err(dev_data->misc_device.dev, "user space failure");
		up(&dev_data->driver_lock);
		return -EFAULT;
	}
	
	if (kfifo_put(dev_data->write_fifo, local_buf, write_size) 
			!= write_size) {
		dev_err(dev_data->misc_device.dev, "write fifo failure");
		up(&dev_data->driver_lock);
		return -EIO;
	}
	
	kick(dev_data);

	up(&dev_data->driver_lock);
	
	return write_size;
}

static int open(struct inode *inode, struct file *filp)
{
	struct device_data *dev_data;
	
	dev_data = find_dev_data(iminor(inode));
	if (dev_data == NULL) {
		printk(KERN_ERR "no device data for minor %d\n", 
				iminor(inode));
		return ENODEV;
	}
	filp->private_data = dev_data;
					
	return 0;
}

static struct file_operations loopback_fops = {
	.owner = THIS_MODULE,
	.llseek = no_llseek,
	.write = write,
	.read = read,
	.open = open,
};

static int probe(struct device *dev)
{
	struct device_data *dev_data;
	struct spi_device *sdev;
	int status;
	
	/* Allocate driver data */
	dev_data = kcalloc(1, sizeof(struct device_data) + BUF_SIZE + 16, 
				GFP_KERNEL);
	if (!dev_data) {
		dev_err(dev, "problem allocating driver memory\n");
		status = -ENOMEM;
		goto out_error;
	}
	
	/* Allocated input/output fifos */
	dev_data->read_fifo = kfifo_alloc(BUF_SIZE, 
					GFP_KERNEL, 
					&dev_data->read_lock);
	if (IS_ERR(dev_data->read_fifo)) {
		status = -ENOMEM;
		goto out_memalloc;
	}
	
	dev_data->write_fifo = kfifo_alloc(BUF_SIZE, 
					GFP_KERNEL, 
					&dev_data->write_lock);
	if (IS_ERR(dev_data->write_fifo)) {
		status = -ENOMEM;
		goto out_readalloc;
	}

	/* Initialize the driver data */
	init_MUTEX(&dev_data->driver_lock);
	spin_lock_init(&dev_data->read_lock);
	spin_lock_init(&dev_data->write_lock);
	atomic_set(&dev_data->busy, -1);
	init_waitqueue_head(&dev_data->read_available);
	init_waitqueue_head(&dev_data->write_available);
	dev_data->dma_buf = dev_data + sizeof(struct device_data);
	dev_data->dma_buf = (void *)(((u32)(dev_data->dma_buf)&0xfffffff8)|8);
	INIT_LIST_HEAD(&dev_data->device_entry);
	dev_set_drvdata(dev, dev_data);
	
	/* Build misc device name from bus number and chip select */
	sdev = to_spi_device(dev);
	snprintf(dev_data->name, BUS_ID_SIZE, "slp%u%u", 
			sdev->master->bus_num,
			sdev->chip_select);
	
	/* Register as a misc device with dynamic minor number */
	dev_data->misc_device.name = dev_data->name;
	dev_data->misc_device.dev = dev;
	dev_data->misc_device.fops = &loopback_fops;
	dev_data->misc_device.minor = MISC_DYNAMIC_MINOR;
	status = misc_register(&dev_data->misc_device);
	if (status != 0) {
		dev_err(dev, "problem (%d) registering misc device\n", status);
		goto out_writealloc;
	}
	
	/* Add device data to internal device list */
	add_device_data(dev_data);
	
	return 0;

out_writealloc:
	kfifo_free(dev_data->write_fifo);

out_readalloc:
	kfifo_free(dev_data->read_fifo);

out_memalloc:
	kfree(dev_data);
	
out_error:
	return status;
}

static int remove(struct device *dev)
{
	struct device_data *dev_data = dev_get_drvdata(dev);
	
	misc_deregister(&dev_data->misc_device);
	remove_device_data(dev_data);
	kfree(dev_data);
	
	return 0;
}

struct device_driver loopback_spi = {
	.name = "loopback",
	.bus = &spi_bus_type,
	.owner = THIS_MODULE,
	.probe = probe,
	.remove = remove,
};

static int __init loopback_init(void)
{
printk("MOIN SPI-Loopback.........................................\n");
	return driver_register(&loopback_spi);
}
module_init(loopback_init);

static void __exit loopback_exit(void)
{
	driver_unregister(&loopback_spi);
}
module_exit(loopback_exit); 
