/*
 *  drivers/spi/spi-dev.c
 *
 *  Character device interface for SPI
 *
 *  Copyright (C) 2005 MontaVista Software, Inc <sources@mvista.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 */
#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/smp_lock.h>

#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/spi.h>

#define SPI_TRANSFER_MAX	65535L

static struct class *spidev_class;

static ssize_t spidev_read(struct file *file, char *buf, size_t count,
			   loff_t * offset);
static ssize_t spidev_write(struct file *file, const char *buf, size_t count,
			    loff_t * offset);

static int spidev_open(struct inode *inode, struct file *file);
static int spidev_release(struct inode *inode, struct file *file);

/**
 * spi_class_device_create - wrapper for class_device_create
 * @minor: sequental minor number of SPI slave device
 * @device: pointer to struct device embedded to spi_device
 */
struct class_device *spi_class_device_create(int minor, struct device *device)
{
	return class_device_create(spidev_class, NULL, MKDEV(SPI_MAJOR, minor), device, "spi%d", minor);
}

/**
 * spi_class_device_destroy - wrapper for class_device_destroy
 * @cdev: class device created by spi_class_device_create
 */
void spi_class_device_destroy(struct class_device *cdev)
{
	class_device_destroy(spidev_class, cdev->devt);
}

static struct file_operations spidev_fops = {
	.owner = THIS_MODULE,
	.llseek = no_llseek,
	.read = spidev_read,
	.write = spidev_write,
	.open = spidev_open,
	.release = spidev_release,
};

/**
 * spidev_read - read from the SPI device
 * @file: device file
 * @buf: buffer to read into
 * @count: number of bytes to read
 * @offset: offset to read from
 */
static ssize_t spidev_read(struct file *file, char __user *buf, size_t count,
			   loff_t * offset)
{
	int rc = 0;
	char *kbuf = kmalloc(count, GFP_DMA);
	struct spi_device *dev = (struct spi_device *)file->private_data;

	if (!kbuf)
		rc = -ENOMEM;
	else {
		rc = spi_read(dev, SPI_M_DMAUNSAFE, kbuf, count);
		if (rc < 0 || copy_to_user(buf, kbuf, count))
			rc = -EFAULT;
		kfree(kbuf);
	}
	return rc;
}

/**
 * spidev_write - write to the SPI device
 * @file: device file
 * @buf: buffer to write the data from
 * @count: number of bytes to write
 * @offset: start offset to write to
 */
static ssize_t spidev_write(struct file *file, const char __user *buf, size_t count,
			    loff_t * offset)
{
	int rc = 0;
	char *kbuf = kmalloc(count, GFP_DMA);
	struct spi_device *dev = (struct spi_device *)file->private_data;

	if (!kbuf)
		rc = -ENOMEM;
	else {
		if (!copy_from_user(kbuf, buf, count))
			rc = spi_write(dev, SPI_M_DMAUNSAFE, kbuf, count);
		else
			rc = -EFAULT;
		kfree(kbuf);
	}
	return rc;
}

struct spidev_openclose {
	unsigned int minor;
	struct file *file;
};

/**
 * spidev_do_open - open the SPI device
 * @the_device: device structure
 * @context: context of SPI device
 */
static int spidev_do_open(struct device *the_dev, void *context)
{
	struct spidev_openclose *o = (struct spidev_openclose *)context;
	struct spi_device *dev = TO_SPI_DEV(the_dev);

	pr_debug("device->minor = %d vs %d\n", dev->minor, o->minor);
	if (dev->minor == o->minor) {
		get_device(&dev->dev);
		o->file->private_data = dev;
		return 1;
	}

	return 0;
}

/**
 * spidev_open - open the SPI bus device
 * @inode: device inode
 * @file: device file
 */
int spidev_open(struct inode *inode, struct file *file)
{
	struct spidev_openclose o;
	int status;

	o.minor = iminor(inode);
	o.file = file;
	status = bus_for_each_dev(&spi_bus, NULL, &o, spidev_do_open);
	if (status == 0) {
		status = -ENODEV;
	}
	return status < 0 ? status : 0;
}

/**
 * spidev_release - release the SPI bus device
 * @inode: device inode
 * @file: device file
 */
static int spidev_release(struct inode *inode, struct file *file)
{
	struct spi_device *dev = file->private_data;

	if (dev)
		put_device(&dev->dev);
	file->private_data = NULL;

	return 0;
}

int __init spidev_init(void)
{
	int res;

	if ((res = register_chrdev(SPI_MAJOR, "spi", &spidev_fops)) != 0) {
		goto out;
	}

	spidev_class = class_create(THIS_MODULE, "spi");
	if (IS_ERR(spidev_class)) {
		printk(KERN_ERR "%s: error creating class\n", __FUNCTION__);
		res = -EINVAL;
		goto out_unreg;
	}

	return 0;

out_unreg:
	unregister_chrdev(SPI_MAJOR, "spi");
out:
	return res;
}

void __exit spidev_cleanup(void)
{
	class_destroy(spidev_class);
	unregister_chrdev(SPI_MAJOR, "spi");
}

MODULE_AUTHOR("dmitry pervushin <dpervushin@ru.mvista.com>");
MODULE_DESCRIPTION("SPI class device driver");
MODULE_LICENSE("GPL");
