/*
 * drivers/spi/spiccxp.c
 *
 * SPI support for the Digi ConnectCore XP Development Platform.
 *
 * @Author: Bernd Westermann
 * @References: [1] spipnx.c
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
/***********************************************************************
 *  @History:
 *     2005/11/30 : initial Version
 ***********************************************************************/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/spi.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/completion.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/dma.h>
#include <asm/hardware/clock.h>
#include <asm/arch/pxa-regs.h>

#include "spifs.h"

#define lock_device( dev )	/* down( &dev->sem ); */
#define unlock_device( dev )	/* up( &dev->sem );   */

static int spiccxp_interrupt(int irq, void *context, struct pt_regs *regs);
static int spiccxp_spi_init(struct device *bus_device);
static int spiccxp_xfer_nondma(struct spi_device *device, struct spi_msg *msg);
static int spiccxp_xfer_dma(struct spi_device *device, struct spi_msg *msg);
static int spiccxp_xfer(struct spi_msg *msg);
static int spiccxp_probe(struct device *bus_device);
static int spiccxp_remove(struct device *bus_device);
static int spiccxp_suspend(struct device *bus_device, pm_message_t state);
static int spiccxp_resume(struct device *bus_device);
static void spiccxp_control(struct spi_device *device, int code, u32 ctl);
static int spiccxp_request_hardware(struct spifs_driver_data *dd,
				   struct platform_device *dev);
static void spiccxp_free_hardware(struct spifs_driver_data *dd,
				 struct platform_device *dev);
static void spiccxp_set_clock(struct device *bus_device, u32 clock);
//static int spiccxp_scale(u32 level);

static struct spi_bus_driver spiccxp_driver = {
	.driver = {
		   .bus = &platform_bus_type,
		   .name = "spiccxp",
		   .probe = spiccxp_probe,
		   .remove = spiccxp_remove,
		   .suspend = spiccxp_suspend,
		   .resume = spiccxp_resume,
/*		   .scale = spiccxp_scale,*/
		   },
	.xfer = spiccxp_xfer,
	.set_clock = spiccxp_set_clock,
};

void spiccxp_set_mode(struct device *bus_device, int mode)
{
	struct spifs_driver_data *dd = dev_get_drvdata(bus_device);

	lock_device(bus_device);
	dd->dma_mode = mode;
	unlock_device(bus_device);
}

static void spiccxp_set_clock(struct device *bus_device, u32 clock)
{
	/* ignored */
}

static int spiccxp_probe(struct device *device)
{
	struct spifs_driver_data *dd;
	int err;

	dd = kmalloc(sizeof(struct spifs_driver_data), GFP_KERNEL);
	if (!dd) {
		err = -ENOMEM;
		goto probe_fail_1;
	}
	memset(dd, 0, sizeof(struct spifs_driver_data));
	/*
	 * these special values are needed because 0 is the valid
	 * assignment to IRQ# and DMA channel as well as DMA slave#
	 */
	dd->dma_channel = -1;
	dd->slave_nr = -1;
#if 0
	dd->irq = NO_IRQ;
#endif

	err = spiccxp_request_hardware(dd, to_platform_device(device));
	if (err)
		goto probe_fail_2;

	spi_bus_driver_init(&spiccxp_driver, device);

	dd->state = SPIFS_STATE_READY;

	dev_set_drvdata(device, dd);

	if (to_platform_device(device)->id == 0)
		spi_bus_populate(device, "EEPROM\0", NULL);

	return 0;

probe_fail_2:
	spiccxp_remove(device);
probe_fail_1:
	return err;
}

int spiccxp_remove(struct device *device)
{
	struct spifs_driver_data *dd = dev_get_drvdata(device);

	spiccxp_free_hardware(dd, to_platform_device(device));
	kfree(dd);
	return 0;
}

static void spiccxp_free_hardware(struct spifs_driver_data *dd,
				 struct platform_device *dev)
{
	lock_device(&dev->dev);

	pxa_set_cken(CKEN23_SSP1, 0); /* diff for ports */

	if (dd->io_flags & IORESOURCE_IRQ) {
		free_irq(dd->irq, dev->dev.bus_id);
		dd->io_flags &= ~IORESOURCE_IRQ;
	}
	if (dd->io_flags & IORESOURCE_MEM) {
		release_mem_region((unsigned long)dd->iostart,
				   dd->ioend - dd->iostart + 1);
		dd->io_flags &= ~IORESOURCE_MEM;
	}
	if (dd->io_flags & IORESOURCE_DMA) {
/*TODO PNDPND		spiccxp_release_dma(&dd->dma_channel); */
		dd->io_flags &= ~IORESOURCE_DMA;
	}

/*	spiccxp_clk_free(dd->clk); */

#ifdef DEBUG
	if (dd->io_flags)
		printk(KERN_ERR
		       "Oops, dd_ioflags for driver data %p is still 0x%x!\n",
		       dd, dd->io_flags);
#endif
	unlock_device(&dev->dev);
}

int spiccxp_request_hardware(struct spifs_driver_data *dd,
			    struct platform_device *dev)
{
	int err = 0;
	struct resource *rsrc;

	lock_device(&dev->dev);
	rsrc = platform_get_resource(dev, IORESOURCE_MEM, 0);
	if (rsrc) {
		dd->iostart = rsrc->start;
		dd->ioend = rsrc->end;
/*		dd->spi_regs = ioremap(dd->iostart, SZ_4K);*/
		if (!request_mem_region
		    (dd->iostart, dd->ioend - dd->iostart + 1, dev->dev.bus_id))
			err = -ENODEV;
		else
			dd->io_flags |= IORESOURCE_MEM;
	}
/*	dd->clk = spiccxp_clk_init(dev);*/
	if (IS_ERR(dd->clk)) {
		err = PTR_ERR(dd->clk);
		goto out;
	}

	pxa_set_cken(CKEN23_SSP1, 1); /* diff for ports */

	dd->irq = platform_get_irq(dev, 0);
#if 0
	if (dd->irq != NO_IRQ) {
#else
	if (dd->irq > 0) {
#endif
// 		dd->spi_regs->ier = 0;
// 		dd->spi_regs->stat = 0;

		if (!request_irq(dd->irq,
				 spiccxp_interrupt,
				 SA_INTERRUPT, dev->dev.bus_id, dd)) {
			err = -ENODEV;
		} else {
			dd->io_flags |= IORESOURCE_IRQ;
		}
	}

	rsrc = platform_get_resource(dev, IORESOURCE_DMA, 0);
	if (rsrc) {
		dd->slave_nr = rsrc->start;
		err = -1 /*TODO spiccxp_request_dma(&dd->dma_channel, dev->dev.bus_id)*/;
		if (!err)
			dd->io_flags |= IORESOURCE_DMA;
	}
out:
	unlock_device(&dev.dev);
	return err;
}

int __init spiccxp_init(void)
{
	printk("SPI bus driver for CCXP270\n");
	return spi_bus_driver_register(&spiccxp_driver);
}

void __exit spiccxp_cleanup(void)
{
	spi_bus_driver_unregister(&spiccxp_driver);
}

/*
 * spiccxp_interrupt
 *
 * This is IRQ handler for CCXP270 SPI bus. The handler signals one of
 * op_complete or threshold completion structures to indicate that
 * transfer is done or the FIFO buffer is empty/full
 * In any way, the interrupt should be acknowledged
 */
static irqreturn_t spiccxp_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
// 	struct spifs_driver_data *dd = dev_id;
//  	unsigned int i_stat = dd->spi_regs->stat;
//  	unsigned int ier = dd->spi_regs->ier;
// 
// 	pr_debug("status = %08X, ier = %08X\n", i_stat, ier);
// 	if ((i_stat & SPIFS_STAT_EOT) && (ier & SPIFS_IER_EOT)) {
// 		dd->spi_regs->ier &= ~SPIFS_IER_EOT;
// 		if (dd->progress)
// 			complete(&dd->op_complete);
// 	}
// 	if ((i_stat & SPIFS_STAT_THR) && (ier & SPIFS_IER_THR)) {
// 		dd->spi_regs->ier &= ~SPIFS_IER_THR;
// 		if (dd->progress)
// 			complete(&dd->threshold);
// 	}
// 	dd->spi_regs->stat |= SPIFS_STAT_CLRINT;
// 	dd->spi_regs->timer_status = 0L;

	return IRQ_HANDLED;
}

static int spiccxp_spi_init(struct device *bus_device)
{
//	struct spifs_driver_data *dd = dev_get_drvdata(bus_device);
//	int timeout;

//	dd->spi_regs->global = SPIFS_GLOBAL_RESET_SPI | SPIFS_GLOBAL_SPI_ON;
	mdelay(5);
//	dd->spi_regs->global = SPIFS_GLOBAL_SPI_ON;
// 	for (timeout = 10000; timeout >= 0; --timeout)
// 		if (dd->spi_regs->global &SPIFS_GLOBAL_SPI_ON)
// 			break;
// 	if (timeout <= 0)
// 		return -1;
// 
// 	dd->spi_regs->con = SPIFS_CON_MS;
// 	dd->spi_regs->con |= SPIFS_CON_BIDIR | (7 << 9);
// 	spiccxp_arch_init(dd->saved_clock / 1000000, dd->spi_regs);

	return 0;
}

void spiccxp_control(struct spi_device *device, int type, u32 ctl)
{
/*TODO	spiccxp_arch_control(device, type, ctl); */
}

/**
 * ssp_read_word - read a word from the SSP port
 *
 * Wait for a data word in the SSP receive FIFO, and return the
 * received data.  Data is LSB justified.
 *
 * Note: Currently, if data is not expected to be received, this
 * function will wait for ever.
 *
 * The caller is expected to perform the necessary locking.
 *
 * Returns:
 *   %-ETIMEDOUT	timeout occurred (for future)
 *   32-bit data	success
 */
int ssp_read_word(u32 port)
{
	while (!(SSSR_P(port) & SSSR_RNE))
		cpu_relax();

	return SSDR_P(port);
}

/**
 * ssp_write_word - write a word to the SSP port
 * @data: 32-bit, MSB justified data to write.
 *
 * Wait for a free entry in the SSP transmit FIFO, and write a data
 * word to the SSP port.
 *
 * The caller is expected to perform the necessary locking.
 *
 * Returns:
 *   %-ETIMEDOUT	timeout occurred (for future)
 *   0			success
 */
int ssp_write_word(u32 port, u32 data)
{
	while (!(SSSR_P(port) & SSSR_TNF))
		cpu_relax();

	SSDR_P(port) = data;

	return 0;
}

static int spiccxp_xfer_nondma(struct spi_device *dev, struct spi_msg *pmsg)
{
	int j, bptr;
	raw_spinlock_t lock;
	struct device *sdev = &dev->dev;
	struct spifs_driver_data *dd = dev_get_drvdata(sdev->parent);
	void (*chip_cs) (struct spi_device * device, int type, u32 ctl) =
	    dd->control ? dd->control : spiccxp_control;

	dev_dbg(sdev, "PIO processing message ...\n");

	chip_cs(dev, SELECT, 0);
// 	dd->spi_regs->con |= SPIFS_CON_THR;

	chip_cs(dev, MESSAGE_START, pmsg->flags & SPI_M_CS);
	if (pmsg->flags & SPI_M_RD) {
		/*  here we have to read data */
		chip_cs(dev, BEFORE_READ, 0);
		bptr = 0;
		init_completion(&dd->op_complete);
// 		dd->spi_regs->frm = 0x0000FFFF & (pmsg->len);
// 		dd->spi_regs->con &= ~SPIFS_CON_RXTX;
// 		dd->spi_regs->con |= SPIFS_CON_SHIFT_OFF;
// 		dd->spi_regs->ier = SPIFS_IER_EOT;
// 		dd->spi_regs->dat = 0x00;

		while (bptr < pmsg->len) {
			if ((pmsg->len) - bptr > FIFO_CHUNK_SIZE) {	/*  if there's data left for another  */
				init_completion(&dd->threshold);	/*  init wait queue for another chunk */
				spin_lock_irq(&lock);
// 				dd->spi_regs->ier |= SPIFS_IER_THR;	/*  chunk, then enable THR interrupt  */
				spin_unlock_irq(&lock);
				wait_for_completion(&dd->threshold);
			} else {
				wait_for_completion(&dd->op_complete);
			}
			for (j = bptr;
			     j <
			     (((pmsg->len) - bptr <
			       FIFO_CHUNK_SIZE) ? (pmsg->len) : (bptr +
								 FIFO_CHUNK_SIZE));
			     j++)
				spimsg_buffer_rd(pmsg)[j] = ssp_read_word(1); /* diff for ports */
			bptr +=
			    ((pmsg->len) - bptr <
			     FIFO_CHUNK_SIZE) ? (pmsg->len -
						 bptr) : FIFO_CHUNK_SIZE;
		}
// 		dd->spi_regs->con &= ~SPIFS_CON_SHIFT_OFF;
		dev_dbg(sdev, "[i]:SPI_M_RD: %x\n", *spimsg_buffer_rd(pmsg));
		chip_cs(dev, AFTER_READ, 0);
	} else if (pmsg->flags & SPI_M_WR) {
		/*  now we have to transmit data        */
		chip_cs(dev, BEFORE_WRITE, 0);
// 		dd->spi_regs->con |= SPIFS_CON_RXTX;
// 		dd->spi_regs->frm = 0x0000FFFF & (pmsg->len);
		bptr = 0;
		init_completion(&dd->op_complete);
// 		dd->spi_regs->ier = SPIFS_IER_EOT;

		while (bptr < pmsg->len) {
			for (j = bptr;
			     j <
			     (((pmsg->len) - bptr <
			       FIFO_CHUNK_SIZE) ? (pmsg->len) : (bptr +
								 FIFO_CHUNK_SIZE));
			     j++) {
				ssp_write_word(1, spimsg_buffer_wr(pmsg)[j]); /* diff for ports */
			}
			if ((pmsg->len) - bptr > FIFO_CHUNK_SIZE) {
				init_completion(&dd->threshold);
				spin_lock_irq(&lock);
// 				dd->spi_regs->ier |= SPIFS_IER_THR;
				spin_unlock_irq(&lock);
				wait_for_completion(&dd->threshold);
			}
			bptr +=
			    ((pmsg->len) - bptr <
			     FIFO_CHUNK_SIZE) ? (pmsg->len -
						 bptr) : FIFO_CHUNK_SIZE;
		}
		wait_for_completion(&dd->op_complete);
		pr_debug("[i]:SPI_M_WR: %x\n", *spimsg_buffer_wr(pmsg));
		chip_cs(dev, AFTER_WRITE, 0);
	}
	chip_cs(dev, MESSAGE_END, pmsg->flags & SPI_M_CSREL);

	chip_cs(dev, UNSELECT, 0);
// 	dd->spi_regs->ier &= ~SPIFS_IER_EOT;

	return 0;
}

//TODO DMA
int spiccxp_xfer_dma(struct spi_device *dev, struct spi_msg *pmsg)
{
	void *buf;
	struct device *sdev = &dev->dev;
	struct spifs_driver_data *dd = dev_get_drvdata(sdev->parent);
	void (*chip_cs) (struct spi_device * device, int type, u32 ctl) =
	    dd->control ? dd->control : spiccxp_control;

	dev_dbg(sdev, "DMA processing message...\n");

	chip_cs(dev, SELECT, 0);

	chip_cs(dev, MESSAGE_START, pmsg->flags & SPI_M_CS);

	if (pmsg->flags & SPI_M_RD) {
		buf = spimsg_buffer_rd(pmsg);
		chip_cs(dev, BEFORE_READ, 0);

		init_completion(&dd->op_complete);
/*		dd->spi_regs->con &= ~SPIFS_CON_RXTX;
		dd->spi_regs->con |= SPIFS_CON_SHIFT_OFF;
		dd->spi_regs->frm = 0x0000FFFF & (pmsg->len);
		dd->spi_regs->ier = SPIFS_IER_EOT;

		dd->spi_regs->dat = 0;
		while (!(dd->spi_regs->stat & SPIFS_STAT_THR)) {
			cpu_relax();
		}*/
/*		spiccxp_setup_dma(dd->iostart, dd->slave_nr, dd->dma_channel,
				 DMA_MODE_READ, pmsg->devbuf_rd, pmsg->len);*/
// 		spiccxp_start_dma(dd->dma_channel);
		wait_for_completion(&dd->op_complete);
// 		spiccxp_stop_dma(dd->dma_channel);
// 		dd->spi_regs->con &= ~SPIFS_CON_SHIFT_OFF;
		chip_cs(dev, AFTER_READ, 0);
	} else {
		buf = spimsg_buffer_wr(pmsg);
		chip_cs(dev, BEFORE_WRITE, 0);
// 		dd->spi_regs->con |= SPIFS_CON_RXTX;
// 		dd->spi_regs->frm = 0x0000FFFF & (pmsg->len);
// 		dd->spi_regs->con &= ~SPIFS_CON_SHIFT_OFF;

// 		spiccxp_setup_dma(dd->iostart, dd->slave_nr, dd->dma_channel,
// 				 DMA_MODE_WRITE, pmsg->devbuf_wr, pmsg->len);
// 		dd->spi_regs->ier = SPIFS_IER_EOT;
		init_completion(&dd->op_complete);
// 		spiccxp_start_dma(dd->dma_channel);
		wait_for_completion(&dd->op_complete);
// 		spiccxp_stop_dma(dd->dma_channel);
		chip_cs(dev, AFTER_WRITE, 0);
	}
	chip_cs(dev, MESSAGE_END, pmsg->flags & SPI_M_CSREL);

	chip_cs(dev, UNSELECT, 0);

	return 0;
}

static inline char *spimsg_flags_str(unsigned flags)
{
	static char s[4] = "";

	s[0] = 0;
	if (flags & SPI_M_RD)
		strcat(s, "R");
	if (flags & SPI_M_WR)
		strcat(s, "W");
	if (flags & ~(SPI_M_WR | SPI_M_RD))
		strcat(s, "*");
	return s;
}

int spiccxp_xfer(struct spi_msg *msg)
{
	struct spi_device *dev = msg->device;
	struct spifs_driver_data *dd = dev_get_drvdata(dev->dev.parent);
	int status;

	lock_device(dev->parent);
	if (!SPIFS_IS_READY(dd)) {
		status = -EIO;
		goto unlock_and_out;
	}

	if (msg->len >= 0xFFFF) {
		status = -EINVAL;
		goto unlock_and_out;
	}
	dev_dbg(&dev->dev, " message (%s), %d bytes\n",
		spimsg_flags_str(msg->flags), msg->len);

	if (spiccxp_spi_init(dev->dev.parent) < 0) {
		dev_dbg(&dev->dev, "spi_init failed, skipping the transfer\n");
		status = -EFAULT;
		goto unlock_and_out;
	}

	dd->progress = 1;
	status = dd->dma_mode ?
	    spiccxp_xfer_dma(dev, msg) : spiccxp_xfer_nondma(dev, msg);
	dd->progress = 0;
unlock_and_out:
	unlock_device(dev->parent);
	return status;
}

/*static int __spiccxp_do_scale(struct device *dev, void *ptr_level)
{
 	struct spifs_driver_data *dd = dev_get_drvdata(dev);
 	spiccxp_arch_do_scale(dd->saved_clock / 1000000, dd->spi_regs);
 	return 0;
}

static int spiccxp_scale(u32 level)
{
	return driver_for_each_dev(&spiccxp_driver.driver,
				   &level, __spiccxp_do_scale);
}*/

static int spiccxp_suspend(struct device *dev, pm_message_t state)
{
	int err = 0;
#ifdef CONFIG_PM
	struct spifs_driver_data *dd = dev_get_drvdata(dev);

	lock_device(dev);
	switch (level) {
	case SUSPEND_DISABLE:
		spiccxp_release_dma(&dd->dma_channel);	/* any sanity checks, eg using DMA,
							   will be done here */
		dd->state = SPIFS_STATE_SUSPENDED;
		break;
	case SUSPEND_SAVE_STATE:
		break;
	case SUSPEND_POWER_DOWN:
		spiccxp_clk_stop(dd->clk);
		break;
	}
	unlock_device(dev);
#endif
	return err;
}

static int spiccxp_resume(struct device *dev)
{
	int err = 0;
#ifdef CONFIG_PM
	struct spifs_driver_data *dd = dev_get_drvdata(dev);

	lock_device(dev);
	switch (level) {
	case RESUME_POWER_ON:
		spiccxp_clk_start(dd->clk);
		break;
	case RESUME_RESTORE_STATE:
		break;
	case RESUME_ENABLE:
		if (dd->slave_nr >= 0)
			spiccxp_request_dma(&dd->dma_channel, dev->bus_id);
		dd->state = SPIFS_STATE_READY;
		break;
	}
	unlock_device(dev);
#endif				/* CONFIG_PM */
	return err;
}

EXPORT_SYMBOL(spiccxp_set_mode);
MODULE_AUTHOR("Bernd Westermann <bernd_westermann@digi.com>");
MODULE_DESCRIPTION("SPI driver for the Digi ConnectCore XP Development Platform");
MODULE_LICENSE("GPL");
module_init(spiccxp_init);
module_exit(spiccxp_cleanup);
