/*
 * drivers/usb/gadget/ns9xxx_udc.c
 *
 * Copyright (C) 2006 Sascha Hauer <s.hauer@pengutronix.de>, Pengutronix
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/***********************************************************************
 *  @History:
 *     2007/04/03 : add external PHY MAX3454E (Bernd Westermann)
 *     2007/03/07 : add CC9P9360DEV support (Bernd Westermann)
 ***********************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>
#include <asm/byteorder.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/mach-types.h>
#include <asm/unaligned.h>
#include <asm/hardware.h>
#include <asm/arch/ns9xxx_udc.h>

#include <asm/arch/regs-bbus.h>

#include <linux/usb_ch9.h>
#include <linux/usb_gadget.h>

#include <asm/dma-mapping.h>

#include "ns9xxx_udc.h"

#define DRIVER_DESC "NS9XXX udc driver"
static char driver_name [] = "ns9xxx-udc";

struct bbus_dma_request {
	void *buf;
	dma_addr_t dma_buf;
	size_t bufsize;
	unsigned long dma_buflen;
	void *priv;
	unsigned int flags;
};

struct bbus_dma_channel {
	void __iomem *base;
	int no;

	struct bbus_dma_desc *desc; /* BBUS_DMA_MAX_DESC entries */

	struct bbus_dma_request *request[BBUS_DMA_MAX_DESC];
	dma_addr_t dma_desc;

	int inpos, outpos;

	int direction; /* DMA_FROM_DEVICE || DMA_TO_DEVICE */
	struct device *dev;
};

static const char       gadget_name [] = "ns9xxx_udc";
static const char 	*epnames[] = { "ep0", "ep1", "ep2", "ep3", "ep4", "ep5", "ep6", "ep7", "ep8", "ep9", "ep10", "ep11"};

enum ep0_state {
	EP0_IDLE,
	EP0_IN_DATA_PHASE,
	EP0_OUT_DATA_PHASE,
	EP0_OUT_ZLP,
};

struct ns9xxx_udc;

struct ns9xxx_ep {
	struct usb_ep				ep;
	struct ns9xxx_udc			*udc;
	const struct usb_endpoint_descriptor	*desc;
	struct list_head			queue;
	struct usb_gadget			*gadget;
	struct bbus_dma_channel			*dma_channel;
	u8					num;
	unsigned int				bytes;
	unsigned int				bEndpointAddress;
	unsigned				stopped : 1;
	unsigned				enabled : 1;
	unsigned int				ep_desc_val;
};

struct bbus_dma_desc;

struct ns9xxx_udc {
	void __iomem			*base;
	int irq;

	struct usb_gadget		gadget;
	struct usb_gadget_driver	*driver;
	struct device			*dev;

	struct ns9xxx_ep		ep[NS9XXX_ENDPOINTS];

	spinlock_t			lock;

	struct bbus_dma_desc		*dma_desc_table;
	dma_addr_t			dma_desc;

	void __iomem			*bbus_dma_base;

	int 				ignore_ep0_queue;

	struct bbus_dma_channel		*dma_channel_ep0out;
	dma_addr_t			dma_buf_ep0out;
	void				*cpu_buf_ep0out;

	struct ns9xxx_udc_mach_info	*mach;

	int				config;

	int				ep0state;

	struct timer_list		timer;
};

struct ns9xxx_request {
	struct usb_request		req;
	struct list_head		queue;
	void				*align_buf;
};

static struct ns9xxx_udc	*the_controller;

#if 0
static void dump_stsctrl(struct ns9xxx_udc *udc)
{
	unsigned int ctrl1 = readl(udc->base + UDFE_CTRL1);
	printk("ctrl0: 0x%08x\n",readl(udc->base + UDFE_CTRL0));
	printk("ctrl1: 0x%08x\n",ctrl1);
	printk("conf: %d int: %d alt: %d\n",ctrl1 & 0xf,
				(ctrl1 >> 4) & 0xf,
				(ctrl1 >> 8) & 0xf);
}
#endif

#ifdef CONFIG_USB_GADGET_DEBUG_FILES

static const char proc_node_name [] = "driver/udc";

static int
udc_proc_read(char *page, char **start, off_t off, int count,
		int *eof, void *_dev)
{
	char			*buf = page;
	char			*next = buf;
	unsigned		size = count;
	int			i, t;
	unsigned long		flags, tmp;
	struct ns9xxx_udc       *udc = _dev;
	char			*ep0state;

	if (off != 0)
		return 0;

	local_irq_save(flags);

	/* basic device status */
	t = scnprintf(next, size, DRIVER_DESC "\n"
		"%s\nGadget driver: %s\nHost %s\n\n",
		driver_name,
		udc->driver ? udc->driver->driver.name : "(none)",
		udc->mach->udc_is_connected() ? "full speed" : "disconnected");
	size -= t;
	next += t;


	switch(udc->ep0state) {
	case EP0_IDLE:
		ep0state = "idle";
		break;
	case EP0_IN_DATA_PHASE:
		ep0state = "in data phase";
		break;
	case EP0_OUT_DATA_PHASE:
		ep0state = "out data phase";
		break;
	case EP0_OUT_ZLP:
		ep0state = "out zlp";
		break;
	default:
		ep0state = "unknown";
		break;
	}

	t = scnprintf(next, size, "EP0 state: %s\n", ep0state);
	size -= t;
	next += t;

	tmp = readl(udc->base + UDFE_IE);
	t = scnprintf(next, size, "UDFE_IE: %s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
		(tmp & IRQ_SET_CFG) ? " setcfg" : "",
		(tmp & IRQ_SET_INTF) ? " setintf" : "",
		(tmp & IRQ_SSPND) ? " sspnd" : "",
		(tmp & IRQ_SOF) ? " sof" : "",
		(tmp & IRQ_URST) ? " rst" : "",
		(tmp & IRQ_FIFO) ? " fifo" : "",
		(tmp & IRQ_DMA(0)) ? " dma0" : "",
		(tmp & IRQ_DMA(1)) ? " dma1" : "",
		(tmp & IRQ_DMA(2)) ? " dma2" : "",
		(tmp & IRQ_DMA(3)) ? " dma3" : "",
		(tmp & IRQ_DMA(4)) ? " dma4" : "",
		(tmp & IRQ_DMA(5)) ? " dma5" : "",
		(tmp & IRQ_DMA(6)) ? " dma6" : "",
		(tmp & IRQ_DMA(7)) ? " dma7" : "",
		(tmp & IRQ_DMA(8)) ? " dma8" : "",
		(tmp & IRQ_DMA(9)) ? " dma9" : "",
		(tmp & IRQ_DMA(10)) ? " dma10" : "",
		(tmp & IRQ_DMA(11)) ? " dma11" : "",
		(tmp & IRQ_GBL_DMA) ? " gbldma" : "",
		(tmp & IRQ_GLB_EN) ? " gbl" : "");

	size -= t;
	next += t;

	tmp = readl(udc->base + UDFE_IS);
	t = scnprintf(next, size, "UDFE_IS: %s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n\n",
		(tmp & IRQ_SET_CFG) ? " setcfg" : "",
		(tmp & IRQ_SET_INTF) ? " setintf" : "",
		(tmp & IRQ_SSPND) ? " sspnd" : "",
		(tmp & IRQ_SOF) ? " sof" : "",
		(tmp & IRQ_URST) ? " rst" : "",
		(tmp & IRQ_FIFO) ? " fifo" : "",
		(tmp & IRQ_DMA(0)) ? " dma0" : "",
		(tmp & IRQ_DMA(1)) ? " dma1" : "",
		(tmp & IRQ_DMA(2)) ? " dma2" : "",
		(tmp & IRQ_DMA(3)) ? " dma3" : "",
		(tmp & IRQ_DMA(4)) ? " dma4" : "",
		(tmp & IRQ_DMA(5)) ? " dma5" : "",
		(tmp & IRQ_DMA(6)) ? " dma6" : "",
		(tmp & IRQ_DMA(7)) ? " dma7" : "",
		(tmp & IRQ_DMA(8)) ? " dma8" : "",
		(tmp & IRQ_DMA(9)) ? " dma9" : "",
		(tmp & IRQ_DMA(10)) ? " dma10" : "",
		(tmp & IRQ_DMA(11)) ? " dma11" : "",
		(tmp & IRQ_GBL_DMA) ? " gbldma" : "");

	size -= t;
	next += t;

	/* dump endpoint queues */
	for (i = 0; i < NS9XXX_ENDPOINTS; i++) {
		struct ns9xxx_ep	*ep = &udc->ep [i];
		struct ns9xxx_request	*req;
		struct bbus_dma_channel *channel = ep->dma_channel;

		const struct usb_endpoint_descriptor	*d;

		d = ep->desc;
		if (!d && i != 0)
			continue;

		t = scnprintf(next, size,"EP%02d%s:\n",
			i,
			ep->bEndpointAddress & USB_DIR_IN ? "IN " : "OUT");

		if (t <= 0 || t > size)
			goto done;
		size -= t;
		next += t;

		while(channel) {
			tmp = readl(channel->base + BBUS_DMA_SIER(channel->no));
			t = scnprintf(next, size, "\tBBUS_DMA_SIER %s: %s%s%s%s%s%s%s%s%s%s\n",
				(channel == udc->dma_channel_ep0out) ? "ep0out" : "",
				(tmp & BBUS_DMA_SIER_PCIE) ? " pcie" : "",
				(tmp & BBUS_DMA_SIER_CAIE) ? " caie" : "",
				(tmp & BBUS_DMA_SIER_NRIE) ? " nrie" : "",
				(tmp & BBUS_DMA_SIER_ECIE) ? " ecie" : "",
				(tmp & BBUS_DMA_SIER_NCIE) ? " ncie" : "",
				(tmp & BBUS_DMA_SIER_PCIP) ? " pcip" : "",
				(tmp & BBUS_DMA_SIER_CAIP) ? " caip" : "",
				(tmp & BBUS_DMA_SIER_NRIP) ? " nrip" : "",
				(tmp & BBUS_DMA_SIER_ECIP) ? " ecip" : "",
				(tmp & BBUS_DMA_SIER_NCIP) ? " ncip" : "");

			size -= t;
			next += t;
			if(i==0 && channel != udc->dma_channel_ep0out)
				channel = udc->dma_channel_ep0out;
			else
				channel = NULL;
		}

		if (list_empty(&ep->queue)) {
			t = scnprintf(next, size, "\t(nothing queued)\n");
			if (t <= 0 || t > size)
				goto done;
			size -= t;
			next += t;
			continue;
		}

		list_for_each_entry(req, &ep->queue, queue) {
			t = scnprintf(next, size,
				"\treq %p len %d/%d buf %p\n",
				&req->req, req->req.actual,
				req->req.length, req->req.buf);

			if (t <= 0 || t > size)
				goto done;
			size -= t;
			next += t;
		}
	}

done:
	local_irq_restore(flags);
	*eof = 1;
	return count - size;
}
#define create_proc_files() \
	create_proc_read_entry(proc_node_name, 0, NULL, udc_proc_read, udc)
#define remove_proc_files() \
	remove_proc_entry(proc_node_name, NULL)

#else	/* !CONFIG_USB_GADGET_DEBUG_FILES */

#define create_proc_files() do {} while (0)
#define remove_proc_files() do {} while (0)

#endif	/* CONFIG_USB_GADGET_DEBUG_FILES */

/* DMA stuff begin */
struct bbus_dma_channel bbus_dma_channels[BBUS_DMA_CHANNELS];

static int bbus_dma_direction(struct bbus_dma_channel *channel, int direction)
{
	int i;
	unsigned int control = BBUS_DMA_CR_MODE_BTE4, flags = 0;

	channel->direction = direction;

	writel(BBUS_DMA_CR_RESET, channel->base + BBUS_DMA_BUF_CR(channel->no));

	if(direction == DMA_TO_DEVICE) {
		control |= BBUS_DMA_CR_FLY_BY_READ;
		flags |= BBUS_DMA_LAST;
	}

	writel(control, channel->base + BBUS_DMA_BUF_CR(channel->no));

	if(direction == DMA_FROM_DEVICE)
		flags |= BBUS_DMA_FULL;

	for(i=0; i<BBUS_DMA_MAX_DESC; i++)
		channel->desc[i].flags = flags;

	channel->desc[BBUS_DMA_MAX_DESC - 1].flags = flags | BBUS_DMA_WRAP;

	return 0;
}

static int bbus_dma_init(struct device *dev, void __iomem *bbus_dma_base)
{
	int i;
	void __iomem *dma_base;
	dma_addr_t dma;

	dma_base = dma_alloc_coherent(dev, 4096, &dma, GFP_KERNEL);
	if( !dma_base ) {
		printk("dma_alloc_coherent failed\n");
		return -ENOMEM;
	}

	memset(dma_base, 0, 4096);

	for(i=0; i<BBUS_DMA_CHANNELS; i++) {
		struct bbus_dma_channel *channel = &bbus_dma_channels[i];

		channel->desc = (struct bbus_dma_desc *)(dma_base + i * 256);
		channel->dma_desc = dma + i * 256;

		dev_dbg(dev, "bbus dma%d: channel->desc: %p channel->dma_desc: 0x%08x\n",
			i,channel->desc,channel->dma_desc);

		channel->base = bbus_dma_base;
		channel->dev = dev;
		channel->no = i;

		writel(channel->dma_desc, channel->base + BBUS_DMA_BUF_DESC(i));
		writel(BBUS_DMA_SIER_NCIE | BBUS_DMA_SIER_ECIE |
		       BBUS_DMA_SIER_NRIE | BBUS_DMA_SIER_CAIE |
		       BBUS_DMA_SIER_PCIE,
			channel->base + BBUS_DMA_SIER(i));
	}
	return 0;
}

static void bbus_dma_exit(void)
{
	struct bbus_dma_channel *channel = &bbus_dma_channels[0];

	dma_free_coherent( channel->dev, 4096,
			   channel->desc, channel->dma_desc );
}

/* only for epnum > 0 */
static inline int ep_fifo_status(struct ns9xxx_udc *udc, int epnum)
{
	unsigned int adr, shift, val;

	adr = (epnum - 1) >> 2;
	shift = ((epnum - 1) % 4) << 3;
	val = (readl(udc->base + FIFO_ISR(1 + adr)) >> shift) & 0xff;
	writel(val << shift, udc->base + FIFO_ISR(1 + adr));

	return val;
}

/* only for epnum > 0 */
static inline void ep_fifo_ier(struct ns9xxx_udc *udc, int epnum, unsigned int mask)
{
	unsigned int adr, shift, val;

	adr = (epnum - 1) >> 2;
	shift = ((epnum - 1) % 4) << 3;

	val = readl(udc->base + FIFO_IER(1 + adr));
	val &= ~(0xff << shift);
	val |= mask << shift;
	writel(val, udc->base + FIFO_IER(1 + adr));

}

/* only for epnum > 0 */
static inline void ep_fifo_ack_int(struct ns9xxx_udc *udc, int epnum, unsigned int mask)
{
	unsigned int adr, shift;

	adr = (epnum - 1) >> 2;
	shift = ((epnum - 1) % 4) << 3;

	writel(mask<<shift, udc->base + FIFO_ISR(1 + adr));

}

static int ns9xxx_ep_enable (struct usb_ep *_ep,
		const struct usb_endpoint_descriptor *desc)
{
	struct ns9xxx_ep        *ep;
	struct ns9xxx_udc       *udc;
	unsigned int ep_desc_val = 0, fifo_sts_val = 0;
	struct bbus_dma_channel *channel;
	char 			*type = NULL;
	unsigned long		flags;

	ep = container_of (_ep, struct ns9xxx_ep, ep);
	udc = ep->udc;

	spin_lock_irqsave(&udc->lock, flags);

	if (!_ep || !desc || ep->desc || _ep->name == epnames[0]
			|| desc->bDescriptorType != USB_DT_ENDPOINT) {
		dev_dbg(udc->dev, "enable ep%d failed\n",ep->num);
		return -EINVAL;
	}

	ep->stopped = 0;
	ep->desc = desc;

	switch(desc->bmAttributes) {
	case USB_ENDPOINT_XFER_BULK:
		type = "bulk";
		ep_desc_val = DESC_EP_TYPE_BULK;
		fifo_sts_val = STS_CTRL_TYPE_BULK;
		break;
	case USB_ENDPOINT_XFER_INT:
		type = "interrupt";
		ep_desc_val = DESC_EP_TYPE_INTERRUPT;
		fifo_sts_val = STS_CTRL_TYPE_INTERRUPT;
		break;
	case USB_ENDPOINT_XFER_ISOC:
		type = "iso";
		ep_desc_val = DESC_EP_TYPE_ISO;
		fifo_sts_val = STS_CTRL_TYPE_ISO;
		break;
	}

	ep->bEndpointAddress = desc->bEndpointAddress;

	if (ep->bEndpointAddress & USB_DIR_IN) {
		fifo_sts_val |= STS_CTRL_IN;
		ep_desc_val |= DESC_EP_IN;
	} else {
		fifo_sts_val |= STS_CTRL_TIMEOUT_CNT(2);
	}

	dev_dbg(udc->dev, "enable %s (%s, %s)\n",_ep->name,
		type, ep->bEndpointAddress & USB_DIR_IN ? "in" : "out");

	ep_desc_val |= (udc->config << DESC_EP_CONFIG_SHIFT);

	channel = ep->dma_channel;
	bbus_dma_direction(channel, (ep->bEndpointAddress & USB_DIR_IN) ?
		DMA_TO_DEVICE : DMA_FROM_DEVICE);

	ep_desc_val |= ep->num;
	ep_desc_val |= DESC_EP_MAXSIZE(EP_FIFO_SIZE); /* FIXME: read from ep */

	writel(ep_desc_val, udc->base + USBD_EP_DESC(channel->no));
	ep->ep_desc_val = ep_desc_val;
	writel(fifo_sts_val, udc->base + FIFO_STS_CTRL(channel->no));
	writel( FIFO_PACKET_CTRL_MAX(EP_FIFO_SIZE), udc->base + FIFO_PACKET_CTRL(channel->no));

	if(ep->bEndpointAddress & USB_DIR_IN)
		ep_fifo_ier(udc, ep->num, EP_ERROR | EP_ACK);
	else
		ep_fifo_ier(udc, ep->num, EP_ERROR);

	ep->enabled = 1;

	spin_unlock_irqrestore(&udc->lock, flags);

	return 0;
}

static void
done(struct ns9xxx_ep *ep, struct ns9xxx_request *req, int status)
{
	unsigned int direction;
	unsigned int stopped = ep->stopped;
	list_del_init(&req->queue);

	if (req->req.status == -EINPROGRESS)
		req->req.status = status;
	else
		status = req->req.status;

	if(unlikely(!ep->num && ep->udc->ep0state == EP0_OUT_ZLP)) {
		direction = 0;
	} else
		direction = ep->bEndpointAddress & USB_DIR_IN;

	dma_unmap_single(ep->udc->gadget.dev.parent,
			req->req.dma, req->req.length,
			direction ? DMA_TO_DEVICE : DMA_FROM_DEVICE);

	req->req.dma = DMA_ADDR_INVALID;

	if(req->align_buf) {
		kfree(req->align_buf);
		req->align_buf = 0;
	}

	ep->bytes += req->req.actual;

	/* don't modify queue heads during completion callback */
	ep->stopped = 1;

	dev_dbg(ep->udc->dev, "complete %s req %p stat %d len %u/%u (totalbytes=%d)\n",
			ep->ep.name, &req->req, status,
			req->req.actual, req->req.length, ep->bytes);

	spin_unlock(&ep->udc->lock);
	req->req.complete(&ep->ep, &req->req);
	spin_lock(&ep->udc->lock);
	ep->stopped = stopped;

}

/* dequeue ALL requests */
static void nuke(struct ns9xxx_ep *ep, int status)
{
	struct ns9xxx_request	*req;
	struct bbus_dma_channel *channel = ep->dma_channel;
	unsigned long control;

	if (list_empty(&ep->queue))
		return;

	/* disable dma channel */
	control = readl(channel->base + BBUS_DMA_BUF_CR(channel->no)) & ~BBUS_DMA_CR_CE;
	writel( control, channel->base + BBUS_DMA_BUF_CR(channel->no));
	writel( control | BBUS_DMA_CR_RESET, channel->base + BBUS_DMA_BUF_CR(channel->no));
	writel( control, channel->base + BBUS_DMA_BUF_CR(channel->no));

	while (!list_empty(&ep->queue)) {
		req = list_entry(ep->queue.next, struct ns9xxx_request, queue);
		done(ep, req, status);
	}
}

static int ns9xxx_ep_disable (struct usb_ep *_ep)
{
	struct ns9xxx_ep	*ep;
	struct ns9xxx_udc	*udc;

	ep = container_of(_ep, struct ns9xxx_ep, ep);
	if (!_ep || !ep->desc)
		return -ENODEV;
	udc = ep->udc;

	nuke(ep, -ESHUTDOWN);

	ep->stopped = 1;
	ep->enabled = 0;
	ep->desc = 0;

	return 0;
}

static struct usb_request *
ns9xxx_ep_alloc_request (struct usb_ep *_ep, unsigned int gfp_flags)
{
	struct ns9xxx_request *req;

	req = kmalloc (sizeof *req, gfp_flags);
	if (!req)
		return 0;

	memset (req, 0, sizeof *req);
	INIT_LIST_HEAD (&req->queue);
	return &req->req;

}

static void
ns9xxx_ep_free_request (struct usb_ep *_ep, struct usb_request *_req)
{
	struct ns9xxx_request	*req;

	req = container_of (_req, struct ns9xxx_request, req);
	WARN_ON (!list_empty (&req->queue));
	kfree(req);
}

static void *
ns9xxx_ep_alloc_buffer ( struct usb_ep *_ep, unsigned bytes, dma_addr_t *dma,
	unsigned int mem_flags)
{
	char *retval;

	if (!the_controller->driver)
		return 0;
	retval = kmalloc (bytes, mem_flags);
	*dma = virt_to_phys (retval);

	return retval;
}

static void ns9xxx_ep_free_buffer ( struct usb_ep *_ep, void *buf,
	dma_addr_t dma,	unsigned bytes)
{
	if (bytes)
		kfree (buf);
}

static int kick_dma(struct ns9xxx_ep *ep, struct ns9xxx_request *req)
{
	struct ns9xxx_udc	*udc = ep->udc;
	struct bbus_dma_channel *channel;
	int len;
	unsigned int control, flags;
	int direction;

	if(unlikely(!ep->num && udc->ep0state == EP0_OUT_DATA_PHASE)) {
		channel = udc->dma_channel_ep0out;
		direction = 0;
	} else {
		channel = ep->dma_channel;
		direction = ep->bEndpointAddress & USB_DIR_IN;
	}

	len = req->req.length - req->req.actual;

	dev_dbg(udc->dev, "kick_dma: channel: %d req.length: %d actual: %d len: %d direction: 0x%08x\n",
		channel->no, req->req.length, req->req.actual, len, direction);

	channel->desc[0].sadr = req->req.dma + req->req.actual;
	channel->desc[0].buflen = len;

	flags = BBUS_DMA_INTERRUPT | BBUS_DMA_WRAP;
	if(direction)
		flags |= BBUS_DMA_FULL | BBUS_DMA_LAST;
	channel->desc[0].flags = flags;

	/* enable channel */
	control = readl(channel->base + BBUS_DMA_BUF_CR(channel->no)) | BBUS_DMA_CR_CE;
	writel( control, channel->base + BBUS_DMA_BUF_CR(channel->no));

	return 0;
}

static int
ns9xxx_ep_queue(struct usb_ep *_ep, struct usb_request *_req, unsigned int gfp_flags)
{
	struct ns9xxx_request	*req;
	struct ns9xxx_ep	*ep;
	struct ns9xxx_udc	*udc;
	unsigned long		flags;
	void			*buf;
	unsigned int		direction;

	req = container_of(_req, struct ns9xxx_request, req);
	if (unlikely (!_req || !_req->complete || !_req->buf
	|| !list_empty(&req->queue))) {
		if (!_req)
			pr_debug("ns9xxx_queue: 1 X X X\n");
		else
			pr_debug("ns9xxx_queue: 0 %01d %01d %01d\n",!_req->complete,!_req->buf, !list_empty(&req->queue));
		return -EINVAL;
	}

	ep = container_of(_ep, struct ns9xxx_ep, ep);
	if (unlikely (!_ep || (!ep->desc && ep->ep.name != epnames[0]))) {
		pr_debug("ep->desc: %p ep->ep.name: %s\n",ep->desc,ep->ep.name);
		return -EINVAL;
	}

	udc = ep->udc;

	if (unlikely (!udc->driver
			|| udc->gadget.speed == USB_SPEED_UNKNOWN)) {
		dev_dbg(udc->dev, "ns9xxx_queue: udc->driver: %p speed: %s\n",udc->driver, udc->gadget.speed == USB_SPEED_UNKNOWN ? "unknown" : "full");
		return -ESHUTDOWN;
	}

	spin_lock_irqsave(&udc->lock, flags);

	dev_dbg(udc->dev, "queue %s %p len %d\n",_ep->name,req, req->req.length);

	_req->status = -EINPROGRESS;
	_req->actual = 0;

	if(!ep->num && udc->ignore_ep0_queue) {
		_req->status = 0;
		_req->actual = _req->length;
		udc->ignore_ep0_queue = 0;
		done(ep, req, 0);
		return 0;
	}

	if(!ep->num && udc->ep0state == EP0_OUT_DATA_PHASE) {
		direction = 0;
	} else
		direction = ep->bEndpointAddress & USB_DIR_IN;

	if(((unsigned int)_req->buf & 0x3) && direction) {
		req->align_buf = kmalloc(_req->length, GFP_ATOMIC);
		memcpy(req->align_buf, _req->buf, _req->length);
		buf = req->align_buf;
	} else
		buf = _req->buf;

	_req->dma = dma_map_single(udc->dev, buf, _req->length,
		direction ? DMA_TO_DEVICE : DMA_FROM_DEVICE);

	/* kickstart this i/o queue? */
	if (list_empty(&ep->queue) && !ep->stopped) {
		kick_dma(ep, req);
	}

	/* irq handler advances the queue. */
	if (likely (req != 0))
		list_add_tail(&req->queue, &ep->queue);

	spin_unlock_irqrestore(&udc->lock, flags);
	return 0;
}

static void cancel_dma(struct ns9xxx_ep *ep)
{
	struct bbus_dma_channel *channel = ep->dma_channel;
	unsigned int control;

	control = readl(channel->base + BBUS_DMA_BUF_CR(channel->no)) | BBUS_DMA_CR_CA;
	writel( control, channel->base + BBUS_DMA_BUF_CR(channel->no));
}

static int ns9xxx_ep_dequeue (struct usb_ep *_ep, struct usb_request *_req)
{
	struct ns9xxx_ep	*ep;
	int			retval = -EINVAL;
	unsigned long		flags;
	struct ns9xxx_request	*req = 0;

	if (!the_controller->driver)
		return -ESHUTDOWN;

	if (!_ep || !_req)
		return retval;
	ep = container_of (_ep, struct ns9xxx_ep, ep);
	if (ep->ep.name == epnames[0])
		return -EINVAL;


    	dev_dbg(ep->udc, "dequeue %s (req=%p)\n",_ep->name, _req);

	spin_lock_irqsave (&ep->udc->lock, flags);
	list_for_each_entry (req, &ep->queue, queue) {
		if (&req->req == _req)
			break;
	}

	if (&req->req != _req) {
		spin_unlock_irqrestore (&ep->udc->lock, flags);
		return -EINVAL;
	}

	list_del_init (&req->queue);

	/* check if the dequeued request is already in progress */
	if(ep->queue.next == &req->queue && !ep->stopped)
		cancel_dma(ep);
	else
		done(ep, req, -ECONNRESET);

	spin_unlock_irqrestore (&ep->udc->lock, flags);

	dev_dbg(ep->udc, "dequeued req %p from %s, len %d buf %p\n",
			req, _ep->name, _req->length, _req->buf);

	return retval;
}

static int
ns9xxx_ep_set_halt (struct usb_ep *_ep, int value)
{
	/* FIXME: implement */
	return 0;
}

static struct usb_ep_ops ns9xxx_ep_ops = {
	.enable		= ns9xxx_ep_enable,
	.disable	= ns9xxx_ep_disable,

	.alloc_request	= ns9xxx_ep_alloc_request,
	.free_request	= ns9xxx_ep_free_request,

	.alloc_buffer	= ns9xxx_ep_alloc_buffer,
	.free_buffer	= ns9xxx_ep_free_buffer,

	.queue		= ns9xxx_ep_queue,
	.dequeue	= ns9xxx_ep_dequeue,

	.set_halt	= ns9xxx_ep_set_halt,
};

/* The controller handles the set_interface and set_configuration
 * requests in hardware without giving us the packet. The gadget
 * driver needs to know, so setup a fake packet and ignore the response
 * from the gadget driver.
 */
static int handle_set_interface(struct ns9xxx_udc *udc)
{
	struct usb_ctrlrequest	r;
	int ret;

	r.bRequest = USB_REQ_SET_INTERFACE;
	r.wIndex = (readl(udc->base + UDFE_CTRL1) >> 4) & 0xf; /* intf setting */
	r.wLength = 0;
	r.wValue = (readl(udc->base + UDFE_CTRL1) >> 8) & 0xf; /* alt setting */
	r.bRequestType = USB_RECIP_INTERFACE;

	dev_dbg(udc->dev, "SET_INTERFACE: intf: %d alt: %d\n",r.wIndex,r.wValue);

	udc->ignore_ep0_queue = 1;
	writel( IRQ_ALL & ~IRQ_DMA_EP0OUT, udc->base + UDFE_IE);
	spin_unlock(&udc->lock);

	ret = udc->driver->setup (&udc->gadget, &r);

	if (ret != 0) spin_lock(&udc->lock);
	writel( IRQ_ALL, udc->base + UDFE_IE);
	udc->ignore_ep0_queue = 0;

	writel( IRQ_SET_INTF, udc->base + UDFE_IS);

	return ret;
}

static int handle_set_configuration (struct ns9xxx_udc *udc)
{
	struct usb_ctrlrequest	r;
	int ret;

	r.bRequest = USB_REQ_SET_CONFIGURATION;
	r.wIndex = 0;
	r.wLength = 0;
	r.wValue = readl(udc->base + UDFE_CTRL1) & 0xf;
	r.bRequestType = 0;

	dev_dbg(udc->dev, "SET_CONFIGURATION: conf: %d\n",r.wValue);

	udc->config = r.wValue;

	/* disable ep0, the controller already responded to the message */
	udc->ignore_ep0_queue = 1;
	writel( IRQ_ALL & ~IRQ_DMA_EP0OUT, udc->base + UDFE_IE);
	spin_unlock(&udc->lock);

	ret = udc->driver->setup(&udc->gadget, &r);

	if (ret != 0) spin_lock(&udc->lock);
	writel( IRQ_ALL, udc->base + UDFE_IE);

	udc->ignore_ep0_queue = 0;

	/* say we are done */
	writel(PRG_CTRL_DONECSR | PRG_CTRL_CSRPRG, udc->base + UDFE_PRG_CTRL);
	writel( IRQ_SET_CFG, udc->base + UDFE_IS);

	return ret;
}

static int kick_dma1(struct ns9xxx_ep *ep, struct ns9xxx_request *req)
{
	struct ns9xxx_udc	*udc = ep->udc;
	struct bbus_dma_channel *channel;
	int len;
	unsigned int control, flags;
	int direction;

	if(unlikely(!ep->num && udc->ep0state == EP0_OUT_DATA_PHASE)) {
		channel = udc->dma_channel_ep0out;
		direction = 0;
	} else {
		channel = ep->dma_channel;
		direction = ep->bEndpointAddress & USB_DIR_IN;
	}

	len = req->req.length - req->req.actual;

	dev_dbg(udc->dev, "kick_dma: channel: %d req.length: %d actual: %d len: %d direction: 0x%08x\n",
		channel->no, req->req.length, req->req.actual, len, direction);

	channel->desc[0].sadr = req->req.dma + req->req.actual;
	channel->desc[0].buflen = len;

	flags = BBUS_DMA_INTERRUPT | BBUS_DMA_WRAP;
	if(direction)
		flags |= BBUS_DMA_FULL | BBUS_DMA_LAST;
	channel->desc[0].flags = flags;

	/* enable channel */
	control = readl(channel->base + BBUS_DMA_BUF_CR(channel->no)) | BBUS_DMA_CR_CE | BBUS_DMA_CR_BDR;
	writel( control, channel->base + BBUS_DMA_BUF_CR(channel->no));

	control = readl(udc->base + FIFO_STS_CTRL(channel->no));
	writel(control & ~STS_CTRL_CLR, udc->base + FIFO_STS_CTRL(channel->no));
	return 0;
}

static int handle_ep(struct ns9xxx_udc *udc, struct ns9xxx_ep *ep)
{
	struct bbus_dma_channel *channel;
	struct ns9xxx_request  *req = NULL;
	unsigned int dma_status;
	int status = 0;

	channel = ep->dma_channel;

	dma_status = readl(channel->base + BBUS_DMA_SIER(channel->no));
	writel(dma_status, channel->base + BBUS_DMA_SIER(channel->no));

	if (!list_empty(&ep->queue))
		req = container_of(ep->queue.next, struct ns9xxx_request, queue);
	else
		goto out;

	if(dma_status & BBUS_DMA_SIER_CAIP) {
		done(ep, req, -ECONNRESET);
		goto out;
	}

	if(!(dma_status & BBUS_DMA_SIER_NCIP))
		goto out;

	if(ep->bEndpointAddress & USB_DIR_IN) {
		if(ep_fifo_status(udc, ep->num) & EP_ERROR) {
			kick_dma1(ep, req);
			goto out;
		}

		req->req.actual = req->req.length;
		done(ep, req, status);
		if (!list_empty(&ep->queue) && !ep->stopped) {
			req = container_of(ep->queue.next, struct ns9xxx_request, queue);
			kick_dma(ep, req);
		}
	} else {
		req->req.actual += channel->desc[0].buflen;
		if(channel->desc[0].flags & STS_CTRL_OVFL) {
			/* recover from overflow
			 * Normally we would just setup the transfer again and
			 * wait for an interrupt. If we do this, we find get
			 * an overflow condition again and again.
			 * Instead we setup a dummy packet to flush the fifo, wait
			 * for the dummy packet and setup the transfer again.
			 */
			req->req.actual -= req->req.actual % ep->ep.maxpacket;

			/* hardware needs a rest */
			udelay(50);
			kick_dma(ep, req);

			while(!(readl(channel->base + BBUS_DMA_SIER(channel->no)) & BBUS_DMA_SIER_NCIP));

			dma_status = readl(channel->base + BBUS_DMA_SIER(channel->no));
			writel(dma_status, channel->base + BBUS_DMA_SIER(channel->no));

			kick_dma(ep,req);
		} else {
			done(ep, req, status);
			if (!list_empty(&ep->queue) && !ep->stopped) {
				req = container_of(ep->queue.next, struct ns9xxx_request, queue);
				kick_dma(ep, req);
			}
		}
	}
out:
	return 0;
}

static void setup_ep0_static_buf(struct ns9xxx_udc *udc) {
	struct                  bbus_dma_channel *channel;
	unsigned int control;

	channel = udc->dma_channel_ep0out;
	udc->dma_buf_ep0out = dma_map_single(udc->dev, udc->cpu_buf_ep0out, 1024, DMA_FROM_DEVICE);

	channel->desc[0].sadr = udc->dma_buf_ep0out;
	channel->desc[0].buflen = 8;
	channel->desc[0].flags = BBUS_DMA_INTERRUPT | BBUS_DMA_WRAP;

	control = readl(channel->base + BBUS_DMA_BUF_CR(channel->no)) | BBUS_DMA_CR_CE;
	writel( control, channel->base + BBUS_DMA_BUF_CR(channel->no));
}

static int handle_ep0(struct ns9xxx_udc *udc)
{
	struct usb_ctrlrequest	*r;
	struct                  bbus_dma_channel *channel_ep0_in, *channel_ep0_out;
	unsigned int flags;
	struct ns9xxx_ep *ep0 = &udc->ep[0];
	unsigned int dma_status, control;
	int ret, ovfl, transferred, timeout;
	struct ns9xxx_request  *req = NULL;

	channel_ep0_in = ep0->dma_channel;
	channel_ep0_out = udc->dma_channel_ep0out;

	dev_dbg(udc->dev, "status channel_ep0_in : 0x%08x\n", readl(channel_ep0_in->base + BBUS_DMA_SIER(channel_ep0_in->no)));
	dev_dbg(udc->dev, "status channel_ep0_out: 0x%08x\n", readl(channel_ep0_out->base + BBUS_DMA_SIER(channel_ep0_out->no)));

	switch(udc->ep0state) {
	case EP0_IDLE:
		dev_dbg(udc->dev, "%s: EP0_IDLE\n",__FUNCTION__);

		/* acknowledge irq status */
		dma_status = readl(channel_ep0_out->base + BBUS_DMA_SIER(channel_ep0_out->no));
		writel(dma_status, channel_ep0_out->base + BBUS_DMA_SIER(channel_ep0_out->no));

		flags = channel_ep0_out->desc[0].flags;

		/* zlp? just setup new buffer and ignore */
		if(!channel_ep0_out->desc[0].buflen) {
			setup_ep0_static_buf(udc);
			break;
		}

		dma_unmap_single(channel_ep0_out->dev, udc->dma_buf_ep0out, 1024, DMA_FROM_DEVICE);
		udc->dma_buf_ep0out = 0;

		if(!(flags & BBUS_DMA_SETUP_CMD)) {
			dev_dbg(udc->dev, "no setup command?\n");
			break;
		}

		r = (struct usb_ctrlrequest *)udc->cpu_buf_ep0out;

		le16_to_cpus (&r->wValue);
		le16_to_cpus (&r->wIndex);
		le16_to_cpus (&r->wLength);

		dev_dbg(udc->dev, "setup command: Type: 0x%02x Request: 0x%02x Value: 0x%04x Index: 0x%04x Length: 0x%04x\n",
			r->bRequestType, r->bRequest, r->wValue, r->wIndex, r->wLength);

		if(r->bRequestType & USB_DIR_IN)
			udc->ep0state = EP0_IN_DATA_PHASE;
		else
			udc->ep0state = EP0_OUT_DATA_PHASE;

		spin_unlock(&udc->lock);
		ret = udc->driver->setup (&udc->gadget, r);
		spin_lock(&udc->lock);

		if(ret<0) {
			unsigned int val;
			dev_dbg(udc->dev, "stalling ep0\n");
			udc->ep0state = EP0_IDLE;
			setup_ep0_static_buf(udc);
			val = readl(udc->base + FIFO_STS_CTRL(channel_ep0_in->no));
			writel( val | STS_CTRL_SEND_STALL, udc->base + FIFO_STS_CTRL(channel_ep0_in->no));
			while(!(readl(udc->base + FIFO_STS_CTRL(channel_ep0_in->no)) & STS_CTRL_STALL_SENT));
			writel( val, udc->base + FIFO_STS_CTRL(channel_ep0_in->no));
		}

		if(udc->ep0state == EP0_IN_DATA_PHASE)
			setup_ep0_static_buf(udc);

		break;

	case EP0_IN_DATA_PHASE:
		dev_dbg(udc->dev, "%s: EP0_IN_DATA_PHASE\n",__FUNCTION__);

		/* acknowledge irq status */
		dma_status = readl(channel_ep0_in->base + BBUS_DMA_SIER(channel_ep0_in->no));
		writel(dma_status, channel_ep0_in->base + BBUS_DMA_SIER(channel_ep0_in->no));

		if (!list_empty(&ep0->queue))
			req = container_of(ep0->queue.next, struct ns9xxx_request, queue);
		if(req) {
			req->req.actual = req->req.length;
			done(ep0, req, 0);
		}

		udc->ep0state = EP0_IDLE;
		break;

	case EP0_OUT_DATA_PHASE:
		dev_dbg(udc->dev, "EP0_OUT_DATA_PHASE\n");

		/* acknowledge irq status */
		dma_status = readl(channel_ep0_out->base + BBUS_DMA_SIER(channel_ep0_out->no));
		writel(dma_status, channel_ep0_out->base + BBUS_DMA_SIER(channel_ep0_out->no));

		ovfl = channel_ep0_out->desc[0].flags & STS_CTRL_OVFL;
		transferred = channel_ep0_out->desc[0].buflen;

		if (!list_empty(&ep0->queue))
			req = container_of(ep0->queue.next, struct ns9xxx_request, queue);
		if(!req) {
			break;
		}

		timeout = 1000;
		while(ovfl && timeout--) {
			/* overflow. need to discard last packet */
			control = readl(channel_ep0_out->base + BBUS_DMA_BUF_CR(channel_ep0_out->no));
			writel(control & ~BBUS_DMA_CR_CE, channel_ep0_out->base + BBUS_DMA_BUF_CR(channel_ep0_out->no));

			if (!list_empty(&ep0->queue))
				req = container_of(ep0->queue.next, struct ns9xxx_request, queue);

			kick_dma(&udc->ep[0], req);
			while(!(readl(channel_ep0_out->base + BBUS_DMA_SIER(channel_ep0_out->no)) & (1<<31)));

			dma_status = readl(channel_ep0_out->base + BBUS_DMA_SIER(channel_ep0_out->no));
			writel(dma_status, channel_ep0_out->base + BBUS_DMA_SIER(channel_ep0_out->no));
			transferred = channel_ep0_out->desc[0].buflen;
			ovfl = channel_ep0_out->desc[0].flags & STS_CTRL_OVFL;
		}

		req->req.actual += transferred;

		if(ovfl || req->req.actual < req->req.length) {
			kick_dma(&udc->ep[0], req);
			break;
		}

		/* send zlp */

		channel_ep0_in->desc[0].sadr = 0;
		channel_ep0_in->desc[0].buflen = 0;
		channel_ep0_in->desc[0].flags = BBUS_DMA_INTERRUPT | BBUS_DMA_FULL |
					BBUS_DMA_WRAP | BBUS_DMA_LAST;

		control = readl(channel_ep0_in->base + BBUS_DMA_BUF_CR(channel_ep0_in->no)) | BBUS_DMA_CR_CE;
		writel( control, channel_ep0_in->base + BBUS_DMA_BUF_CR(channel_ep0_in->no));
		udc->ep0state = EP0_OUT_ZLP;

		break;

	case EP0_OUT_ZLP:
		dev_dbg(udc->dev, "%s: EP0_OUT_ZLP\n",__FUNCTION__);

		/* acknowledge irq status */
		dma_status = readl(channel_ep0_in->base + BBUS_DMA_SIER(channel_ep0_in->no));
		writel(dma_status, channel_ep0_in->base + BBUS_DMA_SIER(channel_ep0_in->no));

		if (!list_empty(&ep0->queue))
			req = container_of(ep0->queue.next, struct ns9xxx_request, queue);
		if(req) {
			done(ep0, req, 0);
		}
		setup_ep0_static_buf(udc);
		udc->ep0state = EP0_IDLE;
		break;
	}
	return 0;
}

static irqreturn_t ns9xxx_udc_irq(int irq, void *dev, struct pt_regs *r)
{
	struct ns9xxx_udc      *udc = dev;
	unsigned int status, val;
	int i, handled;

	spin_lock(&udc->lock);
	do {
		handled = 0;
		status = readl(udc->base + UDFE_IS);
		if (status & IRQ_URST) {
			if (udc->gadget.speed == USB_SPEED_UNKNOWN) {
				udc->gadget.speed = USB_SPEED_FULL;
				mod_timer(&udc->timer, jiffies + (HZ/5));
			} else {
				dev_dbg(udc->dev, "USB reset detected\n");
				for (i = 0; i < NS9XXX_ENDPOINTS; i++) {
					struct ns9xxx_ep *ep = &udc->ep[i];

					nuke(ep, -ESHUTDOWN);
				}
			}
			writel( IRQ_URST, udc->base + UDFE_IS);
			val = readl(udc->base + UDFE_IE);
			writel( val & ~IRQ_URST, udc->base + UDFE_IE);
		}

		if (status & (IRQ_DMA_EP0OUT | IRQ_DMA_EP0IN)) {
			handle_ep0(udc);
			handled = 1;
		}

		if(status & IRQ_SET_CFG) {
			if(handle_set_configuration(udc))
				goto send_stall;
			handled = 1;
		}

		if(status & IRQ_SET_INTF) {
			if(handle_set_interface(udc))
				goto send_stall;
			handled = 1;
		}

		if(status & IRQ_SSPND) {
			writel(IRQ_SSPND, udc->base + UDFE_IS);
		}
#if 0
		if(status & IRQ_FIFO) {
			unsigned int fifo_status;
			fifo_status = readl(udc->base + FIFO_ISR(0));
			writel(fifo_status, udc->base + FIFO_ISR(0));

			for(i=1; i<11; i++) {
				if(!udc->ep[i].enabled)
					continue;
				fifo_status = ep_fifo_status(udc, i);
				if(fifo_status & EP_ERROR)
					printk("ep%d error\n",i);
				if(fifo_status & EP_ACK)
					printk("ep%d ack\n",i);
				ep_fifo_ack_int(udc, i, fifo_status);
			}

			handled = 1;
		}
#endif
		if (status & (IRQ_DMA_EP1_10)) {
			for(i=1; i<11; i++) {
				if(status & IRQ_DMA(i+1)) {
					handle_ep(udc, &udc->ep[i]);
				}
			}
			handled = 1;
		}
	} while(handled);

send_stall:

	spin_unlock(&udc->lock);
	return IRQ_HANDLED;
}

/*
 * device-scoped parts of the api to the usb controller hardware
 */
static int ns9xxx_get_frame(struct usb_gadget *_gadget)
{
	printk("%s\n",__FUNCTION__);
	return 0;
}

static int ns9xxx_wakeup(struct usb_gadget *_gadget)
{
	printk("%s\n",__FUNCTION__);
	return 0;
}

static const struct usb_gadget_ops ns9xxx_ops = {
	.get_frame          = ns9xxx_get_frame,
	.wakeup             = ns9xxx_wakeup,
	/* current versions must always be self-powered */
};

#define NS_BBUS_UTILITY_BASE_USBD_RESET (1<<12)

static int udc_enable (struct ns9xxx_udc *udc)
{
	int err = 0;
	udc->ep0state = EP0_IDLE;

#ifdef CONFIG_USB_GADGET_NS9360_EXT_MAX3454E
	/* GPIO 42..49 (17) used for USB-Device-EXTPHY */
	if (ns9xxx_request_gpio( GPIO_UDC_VP, "usb-device-ext-VP") != 0) {
		err = -1;
		goto skip_gpio;
	}
	ns9xxx_gpio_cfgpin( GPIO_UDC_VP, GPIO_CFG_FUNC_2 );
	if (ns9xxx_request_gpio( GPIO_UDC_VM, "usb-device-ext-VM") != 0) {
		err = -1;
		ns9xxx_release_gpio( GPIO_UDC_VP );
		goto skip_gpio;
	}
	ns9xxx_gpio_cfgpin( GPIO_UDC_VM, GPIO_CFG_FUNC_2 );
	if (ns9xxx_request_gpio( GPIO_UDC_DE, "usb-device-ext-DE") != 0) {
		err = -1;
		ns9xxx_release_gpio( GPIO_UDC_VP );
		ns9xxx_release_gpio( GPIO_UDC_VM );
		goto skip_gpio;
	}
	ns9xxx_gpio_cfgpin( GPIO_UDC_DE, GPIO_CFG_FUNC_2 | GPIO_CFG_INVERT );
	if (ns9xxx_request_gpio( GPIO_UDC_RCV, "usb-device-ext-RCV") != 0) {
		err = -1;
		ns9xxx_release_gpio( GPIO_UDC_VP );
		ns9xxx_release_gpio( GPIO_UDC_VM );
		ns9xxx_release_gpio( GPIO_UDC_DE );
		goto skip_gpio;
	}
	ns9xxx_gpio_cfgpin( GPIO_UDC_RCV, GPIO_CFG_FUNC_2 );
	if (ns9xxx_request_gpio( GPIO_UDC_ENUM, "usb-device-ext-ENUM") != 0) {
		err = -1;
		ns9xxx_release_gpio( GPIO_UDC_VP );
		ns9xxx_release_gpio( GPIO_UDC_VM );
		ns9xxx_release_gpio( GPIO_UDC_DE );
		ns9xxx_release_gpio( GPIO_UDC_RCV );
		goto skip_gpio;
	}
	ns9xxx_gpio_cfgpin( GPIO_UDC_ENUM, GPIO_CFG_OUTPUT );
	ns9xxx_gpio_setpin( GPIO_UDC_ENUM, 1 );
	if (ns9xxx_request_gpio( GPIO_UDC_SUSP, "usb-device-ctrl-ext-SUSP") != 0) {
		err = -1;
		ns9xxx_release_gpio( GPIO_UDC_VP );
		ns9xxx_release_gpio( GPIO_UDC_VM );
		ns9xxx_release_gpio( GPIO_UDC_DE );
		ns9xxx_release_gpio( GPIO_UDC_RCV );
		ns9xxx_release_gpio( GPIO_UDC_ENUM );
		goto skip_gpio;
	}
	ns9xxx_gpio_cfgpin( GPIO_UDC_SUSP, GPIO_CFG_FUNC_0 );
	if (ns9xxx_request_gpio( GPIO_UDC_SPEED, "usb-device-ctrl-ext-SPEED") != 0) {
		err = -1;
		ns9xxx_release_gpio( GPIO_UDC_VP );
		ns9xxx_release_gpio( GPIO_UDC_VM );
		ns9xxx_release_gpio( GPIO_UDC_DE );
		ns9xxx_release_gpio( GPIO_UDC_RCV );
		ns9xxx_release_gpio( GPIO_UDC_ENUM );
		ns9xxx_release_gpio( GPIO_UDC_SUSP );
		goto skip_gpio;
	}
	ns9xxx_gpio_cfgpin( GPIO_UDC_SPEED, GPIO_CFG_FUNC_0 );
#endif
skip_gpio:
	if (err != 0) {
		printk(KERN_WARNING "%s: Failed to reserve gpios\n", driver_name );
		return -EBUSY;
	}

	/* enable dynamic programming feature */
	writel( PRG_CTRL_CSRPRG, udc->base + UDFE_PRG_CTRL);

	writel( readl(BBUS_UTILITY_BASE_VA) | NS_BBUS_UTILITY_BASE_USBD_RESET, BBUS_UTILITY_BASE_VA);
#ifdef CONFIG_USB_GADGET_NS9360_INT_PHY
	/* use internal PHY */
	writel( NS_BBUS_USB_CFG_INT_PHY_USBD | NS_BBUS_USB_CFG_SPEED,  NS_BBUS_USB_CFG );
#else
#  ifdef CONFIG_USB_GADGET_NS9360_EXT_MAX3454E
	/* use external MAX3454E PHY */
	BBUS_USB_CFG = NS_BBUS_USB_CFG_SPEED;
#  endif
#endif
	writel( readl(BBUS_UTILITY_BASE_VA) & ~NS_BBUS_UTILITY_BASE_USBD_RESET, BBUS_UTILITY_BASE_VA);

	/* according to datasheet */
	writel( DESC_SETUP_VAL, udc->base + USBD_DESC_SETUP);

	/* set selfpowered */
	writel( CTRL1_28 | CTRL1_SPWR, udc->base + UDFE_CTRL1);
	writel( IRQ_ALL, udc->base + UDFE_IE);

	/* setup ep0
	 *
	 * Quote from the datasheet:
	 * Set the maximum packet size for the Control-OUT endpoint to 65,
	 * so all control packets are less than the maximum packet size of
	 * 64 bytes
	 */
	writel( DESC_EP_MAXSIZE(0x40), udc->base + USBD_EP_DESC(0));
	writel( FIFO_PACKET_CTRL_MAX(0x41), udc->base + FIFO_PACKET_CTRL(0));
	writel( 0, udc->base + FIFO_STS_CTRL(0));

	writel( FIFO_PACKET_CTRL_MAX(0x40), udc->base + FIFO_PACKET_CTRL(1));
	writel( STS_CTRL_IN, udc->base + FIFO_STS_CTRL(1));

	writel( (EP_ERROR | EP_ACK)<<8 | (EP_ERROR | EP_ACK),
		      udc->base + FIFO_IER(0));

	bbus_dma_init(udc->dev, udc->bbus_dma_base);

	bbus_dma_direction(udc->dma_channel_ep0out, DMA_FROM_DEVICE);
	bbus_dma_direction(udc->ep[0].dma_channel, DMA_TO_DEVICE);

	/* allocate a buffer for EP0OUT requests */
	udc->cpu_buf_ep0out = kmalloc(1024, GFP_KERNEL);

	setup_ep0_static_buf(udc);
	return 0;
}

static void nop_release (struct device *dev)
{
        printk("%s %s\n", __FUNCTION__, dev->bus_id);
}

static struct ns9xxx_udc memory = {
	.gadget = {
		.ops		= &ns9xxx_ops,
		.ep0		= &memory.ep[0].ep,
		.name		= gadget_name,
		.dev = {
			.bus_id		= "gadget",
			.release	= nop_release,
		},
	},
};

/*
 * 	udc_reinit - initialize software state
 */
static void udc_reinit(struct ns9xxx_udc *udc)
{
	int i;

	/* device/ep0 records init */
	INIT_LIST_HEAD (&udc->gadget.ep_list);
	INIT_LIST_HEAD (&udc->gadget.ep0->ep_list);
	udc->ep0state = EP0_IDLE;

	/* basic endpoint records init */
	for (i = 0; i < NS9XXX_ENDPOINTS; i++) {
		struct ns9xxx_ep *ep = &udc->ep[i];

		if (i != 0)
			list_add_tail (&ep->ep.ep_list, &udc->gadget.ep_list);

		ep->stopped = 0;
		ep->desc    = 0;
		ep->num     = i;
		ep->ep.ops  = &ns9xxx_ep_ops;
		ep->udc     = &memory;
		ep->ep.name = epnames[i];

		INIT_LIST_HEAD (&ep->queue);
	}

	udc->dma_channel_ep0out     = &bbus_dma_channels[0];
	udc->ep[0].dma_channel      = &bbus_dma_channels[1];
	udc->ep[0].bEndpointAddress = USB_DIR_IN;
	udc->ep[0].ep.maxpacket     = 64;

	for (i = 1; i < NS9XXX_ENDPOINTS; i++) {
		struct ns9xxx_ep *ep = &udc->ep[i];

		ep->dma_channel  = &bbus_dma_channels[i+1];
		ep->bytes        = 0;
		ep->ep.maxpacket = EP_FIFO_SIZE;
	}
}

/* when a driver is successfully registered, it will receive
 * control requests including set_configuration(), which enables
 * non-control requests.  then usb traffic follows until a
 * disconnect is reported.  then a host may connect again, or
 * the driver might get unbound.
 */
int usb_gadget_register_driver(struct usb_gadget_driver *driver)
{
	struct ns9xxx_udc *udc = the_controller;
	int ret;

    	printk("usb_gadget_register_driver() '%s'\n",
		driver->driver.name);

	if (!udc) {
		pr_debug("%s: !udc\n",__FUNCTION__);
		return -ENODEV;
	}
	if (udc->driver) {
		dev_dbg(udc->dev, "driver already present\n");
		return -EBUSY;
	}
	if (!driver->bind || !driver->unbind || !driver->setup
			|| driver->speed == USB_SPEED_UNKNOWN) {
		return -EINVAL;
	}

	/* first hook up the driver ... */
	udc->driver = driver;
	udc->gadget.dev.driver = &driver->driver;
	udc->gadget.dev.driver_data = &driver->driver;

	device_add(&udc->gadget.dev);
	ret = driver->bind(&udc->gadget);
	if (ret) {
		dev_dbg(udc->dev, "bind to driver %s --> error %d\n",
				driver->driver.name, ret);

		device_del(&udc->gadget.dev);
		udc->driver = NULL;
		udc->gadget.dev.driver = NULL;
		return ret;
	}

	udc_enable(udc);

	udc->mach->udc_command(NS9XXX_UDC_CMD_CONNECT);

	mod_timer(&udc->timer, jiffies + (HZ/5));
	return 0;
}

static void
stop_activity(struct ns9xxx_udc *udc, struct usb_gadget_driver *driver)
{
	int i;
	unsigned int val;

	/* don't disconnect drivers more than once */
	if (udc->gadget.speed == USB_SPEED_UNKNOWN)
		driver = 0;
	udc->gadget.speed = USB_SPEED_UNKNOWN;

	/* prevent new request submissions, kill any outstanding requests  */
	for (i = 0; i < NS9XXX_ENDPOINTS; i++) {
		struct ns9xxx_ep *ep = &udc->ep[i];

		nuke(ep, -ESHUTDOWN);
	}

	val = readl(udc->base + UDFE_IE);
	writel( val | IRQ_URST, udc->base + UDFE_IE);

	/* report disconnect; the driver is already quiesced */
	if (driver)
		driver->disconnect(&udc->gadget);

	/* re-init driver-visible data structures */
	udc_reinit(udc);
}

static void ns9xxxudc_connect_status(unsigned long _dev)
{
	struct ns9xxx_udc *udc = the_controller;

	if(!udc->mach->udc_is_connected()) {
		stop_activity(udc, udc->driver);
/*		if (udc->driver)
			udc->driver->disconnect(&udc->gadget);*/
		return;
	}

	mod_timer(&udc->timer, jiffies + (HZ/5));
}

int usb_gadget_unregister_driver(struct usb_gadget_driver *driver)
{
	struct ns9xxx_udc *udc = the_controller;
	unsigned long flags;

	if (!udc)
		return -ENODEV;
	if (!driver || driver != udc->driver)
		return -EINVAL;

	spin_lock_irqsave(&udc->lock, flags);
	udc->driver = NULL;
	udc->mach->udc_command(NS9XXX_UDC_CMD_DISCONNECT);
	stop_activity(udc, driver);
	del_timer_sync(&udc->timer);
	spin_unlock_irqrestore(&udc->lock, flags);

	driver->unbind(&udc->gadget);
	device_del(&udc->gadget.dev);

	bbus_dma_exit();

	if(udc->dma_buf_ep0out != DMA_ADDR_INVALID)
		dma_unmap_single(udc->dev, udc->dma_buf_ep0out, 1024, DMA_FROM_DEVICE);

	kfree(udc->cpu_buf_ep0out);

	/* Free gpios */
	ns9xxx_gpio_cfgpin( GPIO_UDC_VP, GPIO_CFG_OUTPUT );
	ns9xxx_release_gpio( GPIO_UDC_VP );
	ns9xxx_gpio_cfgpin( GPIO_UDC_VM, GPIO_CFG_OUTPUT );
	ns9xxx_release_gpio( GPIO_UDC_VM );
	ns9xxx_gpio_cfgpin( GPIO_UDC_DE, GPIO_CFG_OUTPUT );
	ns9xxx_release_gpio( GPIO_UDC_DE );
	ns9xxx_gpio_cfgpin( GPIO_UDC_RCV, GPIO_CFG_OUTPUT );
	ns9xxx_release_gpio( GPIO_UDC_RCV );
	ns9xxx_gpio_cfgpin( GPIO_UDC_ENUM, GPIO_CFG_OUTPUT );
	ns9xxx_release_gpio( GPIO_UDC_ENUM );
	ns9xxx_gpio_cfgpin( GPIO_UDC_SUSP, GPIO_CFG_OUTPUT );
	ns9xxx_release_gpio( GPIO_UDC_SUSP );
	ns9xxx_gpio_cfgpin( GPIO_UDC_SPEED, GPIO_CFG_OUTPUT );
	ns9xxx_release_gpio( GPIO_UDC_SPEED );

	return 0;
}

static int __devinit ns9xxx_udc_probe(struct platform_device *pdev)
{
	struct device	*dev = &pdev->dev;
	struct resource *res, *res_dma;
	struct ns9xxx_udc *udc = &memory;
	int ret;

	dev_info(dev, "gadget driver initializing\n");

	if (!dev->platform_data) {
		/* small (so we copy it) but critical! */
		dev_err(dev, "missing platform_data\n");
		ret = -ENODEV;
		goto out;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(dev, "could not get resource\n");
		ret = -ENODEV;
		goto out;
	}

	res_dma = platform_get_resource(pdev, IORESOURCE_MEM, 1);
	if (!res_dma) {
		dev_err(dev, "could not get resource\n");
		ret = -ENODEV;
		goto out;
	}

	if (!request_mem_region(res->start, res->end - res->start + 1, "ns9xxx-udc")) {
		dev_err(dev, "could not get memory region\n");
		ret = -EBUSY;
		goto out;
	}

	udc->base = ioremap(res->start, res->end - res->start + 1);
	if (!udc->base) {
		dev_err(dev, "failed to ioremap\n");
		ret = -ENOMEM;
		goto out_release;
	}

	if (!request_mem_region(res_dma->start, res_dma->end - res_dma->start + 1, "ns9xxx-udc")) {
		dev_err(dev, "could not get memory region\n");
		ret = -EBUSY;
		goto out_iounmap;
	}

	udc->bbus_dma_base = ioremap(res_dma->start, res_dma->end - res_dma->start + 1);
	if (!udc->bbus_dma_base) {
		dev_err(dev, "failed to ioremap\n");
		ret = -ENOMEM;
		goto out_release1;
	}

	udc->irq = platform_get_irq(pdev, 0);

	ret = request_irq(udc->irq, ns9xxx_udc_irq,
		SA_INTERRUPT, gadget_name, udc);
	if (ret != 0) {
		dev_err(dev, "can't get irq %i, err %d\n",
			udc->irq, ret);
		goto out_iounmap1;
	}

	spin_lock_init(&udc->lock);
	udc->dev = &pdev->dev;

	device_initialize(&udc->gadget.dev);
	udc->gadget.dev.parent = &pdev->dev;

	the_controller = udc;
	platform_set_drvdata(pdev, dev);

	udc->mach = pdev->dev.platform_data;
	udc->gadget.dev.dma_mask = dev->dma_mask;

	init_timer(&udc->timer);
	udc->timer.function = ns9xxxudc_connect_status;
	udc->timer.data = (unsigned long) udc;

	udc_reinit(udc);
	create_proc_files();

	return 0;

 out_iounmap1:
	iounmap(udc->bbus_dma_base);
 out_release1:
	release_mem_region(res_dma->start, res_dma->end - res_dma->start + 1);
 out_iounmap:
	iounmap(udc->base);
 out_release:
	release_mem_region(res->start, res->end - res->start + 1);
 out:
	return ret;

}

static int __devexit ns9xxx_udc_remove(struct platform_device *pdev)
{
	struct resource *res;
	struct ns9xxx_udc *udc = platform_get_drvdata(pdev);

	remove_proc_files();

	usb_gadget_unregister_driver(udc->driver);

	free_irq(udc->irq, udc);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	iounmap(udc->bbus_dma_base);
	release_mem_region(res->start, res->end - res->start + 1);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 1);
	iounmap(udc->base);
	release_mem_region(res->start, res->end - res->start + 1);

	platform_set_drvdata(pdev, 0);

	the_controller = 0;

	return 0;
}

EXPORT_SYMBOL (usb_gadget_unregister_driver);
EXPORT_SYMBOL (usb_gadget_register_driver);

static struct platform_driver ns9xxx_udc = {
	.probe		= ns9xxx_udc_probe,
	.remove		= __devexit_p(ns9xxx_udc_remove),
	.driver		= {
		.name	= (char *) driver_name,
		.owner	= THIS_MODULE,
	},
};

static int __devinit udc_init_module(void)
{
	return platform_driver_register(&ns9xxx_udc);
}
module_init(udc_init_module);

static void __devexit udc_exit_module(void)
{
	platform_driver_unregister(&ns9xxx_udc);
}
module_exit(udc_exit_module);

MODULE_AUTHOR("Sascha Hauer, Pengutronix");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
