/*
 * drivers/char/ns9xxx_hperf.h
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
/***********************************************************************
 *
 * !Revision: $Revision: 1.5 $
 * !Descr:    High Performance Timer Measure Functions. Timer 1 is configured
 *            on CPU clock.
 * !Usage:    o Mode 1 (just counting)
 *              #include <am/arch-ns9xxx/ns9xxx_hperf.h>
 *              ns_hperf_start();
 *              <do_something>
 *              printk( KERN_INFO "Elapsed %i ns\n", ns_hperf_to_nsec( ns_hperf_counter() ) )
 *            o Mode 2 (more complex)
 *              enum {
 *                  START_XMIT = 0,
 *                  START_XMIT_DONE,
 *                  TX_INT,
 *                  TX_INT_DONE
 *              };
 *              # define NAME(x) #x
 *              static const char* ns_hperf_type_names[] = {
 *                NAME( START_XMIT ),
 *                NAME( START_XMIT_DONE ),
 *              };
 *              #include <am/arch-ns9xxx/ns9xxx_hperf.h>
 *              ns_hperf_mark( START_XMIT );
 *              ns_hperf_mark( START_XMIT_DONE );
 *              ns_hperf_marks_dump();
 *              
 * !Author:   Markus Pietrek
 * !TODO:     o Set timer to dump marks on inactivity
 *            o reserve timer 1
 *
 ***********************************************************************/

#ifndef __ASM_ARCH_NS9XXX_HPERF_H
#define __ASM_ARCH_NS9XXX_HPERF_H

#include <asm/io.h>             /* iowrite32 */
#include <asm/arch-ns9xxx/system.h> /* ns9xxx_cpuclock */
#include <asm/arch-ns9xxx/ns9xxx_sys.h>

#ifndef NS_HPERF_SIZE
# define NS_HPERF_SIZE 128
#endif

/* local definitions */
#define NS_SYS_BASE_VA io_p2v( NS_SYS_BASE_PA )
#define sys_iowrite32( val, offs ) iowrite32( val, NS_SYS_BASE_VA + (offs) )
#define sys_ioread32( offs ) 	   ioread32( NS_SYS_BASE_VA + (offs) )
#define TIMER_USED		1
#define RELOAD_TIMER_VAL	0
#define TIMER_CFG \
        ( NS_SYS_TIMER_CTRL_TEN    | \
          NS_SYS_TIMER_CTRL_TLCS_1 | \
          NS_SYS_TIMER_CTRL_INTS   | \
          NS_SYS_TIMER_CTRL_TSZ )

#ifndef NS_HPERF_DUMP_ON_ELAPSED_NSEC
# define NS_HPERF_DUMP_ON_ELAPSED_NSEC	1000000000
#endif /* NS_HPERF_DUMP_ON_ELAPSED_NSEC */

#define NS_HPERF_MARKS_START (&ns_hperf_marks[ 0 ])

/**
 * struct ns_hperf_t  - Stores timing information for a specific mark type
 */
typedef struct {
        int type;
        int counter;
        int value;
} ns_hperf_t;

/* provide a declaration so just ns_hperf_start and ns_hperf_counter
   can be used */
static const char* ns_hperf_type_names[]; /* array with all types defined */
static ns_hperf_t  ns_hperf_marks[ NS_HPERF_SIZE ]; /* storing all marks  */
/* points to next marker used  */
static ns_hperf_t* ns_hperf_mark_next = NS_HPERF_MARKS_START;

/**
 * ns_hperf_start - starts or restarts the timer
 */

static inline void ns_hperf_start( void )
{
	sys_iowrite32( 0,                NS_SYS_TIMER_CTRL(  TIMER_USED ) );
	sys_iowrite32( RELOAD_TIMER_VAL, NS_SYS_TIMER_COUNT( TIMER_USED ) );
        /* go baby, go */
	sys_iowrite32( TIMER_CFG,        NS_SYS_TIMER_CTRL(  TIMER_USED ) );
}

/**
 * ns_hperf_counter - returns the elapsed CPU clocks since ns_hperf_start
 *
 * If the counter overflowed, a warning is printed
 */
static inline unsigned int ns_hperf_counter( void )
{
        unsigned int counter = sys_ioread32( NS_SYS_TIMER_READ( TIMER_USED ) );

        return counter;
}

/**
 * ns_hperf_to_nsec - converts the elapsed clocks from ns_hperf_get
 *                    to nanoseconds 
 *
 * If the counter overflowed, a warning is printed
 */
static inline unsigned int ns_hperf_to_nsec( unsigned int counter )
{
        return ( counter * ( 1000000 / ( ns9xxx_cpuclock() / 1000 ) ) );
}

/**
 * ns_hperf_info - prints some information about the counter
 */
static inline void ns_hperf_info( void )
{
        printk( KERN_DEBUG "ns_hperf: Providing timing resolution of %i ns for %i seconds\n",
                1000000 / (ns9xxx_cpuclock() / 1000),
                0xffffffff / ns9xxx_cpuclock() );
}

/**
 * ns_hperf_marks_dump - Prints collected information and restarts collecting
 */
static void ns_hperf_marks_dump( void )
{
        static int used = 0;
        const ns_hperf_t* mark;
        unsigned int last_usec = 0;

        if( !used ) {
                /* only once */
                ns_hperf_info();
                used = 1;
        }
        
        printk( KERN_DEBUG "Timing information\n" );

        /* dump it */
        for( mark = NS_HPERF_MARKS_START; mark < ns_hperf_mark_next; mark++) {
                unsigned int usec = ns_hperf_to_nsec( mark->counter ) / 1000;
                if( mark->type < ARRAY_SIZE( ns_hperf_type_names ) )
                        printk( KERN_DEBUG " %-24s: ",
                                ns_hperf_type_names[ mark->type ] );
                else
                        printk( KERN_ERR " **** UNKNOWN **** (%03i): ",
                                mark->type );

                if( !mark->counter )
                        printk( "(start)" );
                else
                        printk( "%08u us (+%06u)", usec, (usec-last_usec) );

                if( mark->value )
                        printk( " (=0x%08x)", mark->value );

                printk( "\n" );
                last_usec = usec;
        }

        /* reset statistics */
        ns_hperf_mark_next = NS_HPERF_MARKS_START;
}

/**
 * ns_hperf_mark - Marks the time. If nothing marked so far, starts the
 *                 counting. If too much time has passed since last
 *                 ns_hperf_mark, statistic is printed and reseted
 */
static void ns_hperf_mark( int type, int value )
{
        int counter = ns_hperf_counter();

        if( &ns_hperf_marks[ ARRAY_SIZE( ns_hperf_marks ) ] == ns_hperf_mark_next )
                ns_hperf_marks_dump();
        
        if( NS_HPERF_MARKS_START != ns_hperf_mark_next ) {
                unsigned int nsecs_elapsed = ns_hperf_to_nsec( counter - (ns_hperf_mark_next - 1 )->counter );
                if( nsecs_elapsed > NS_HPERF_DUMP_ON_ELAPSED_NSEC )
                        /* too much idle time */
                        ns_hperf_marks_dump();
        }
        
        if( NS_HPERF_MARKS_START == ns_hperf_mark_next  ) {
        	/* start */
        	counter = 0;
                /* everything is initalized, mark time from here */
                ns_hperf_start();
        }
        
        /* remember mark */
        ns_hperf_mark_next->counter = counter;
        ns_hperf_mark_next->type    = type;
        ns_hperf_mark_next->value   = value;

        ns_hperf_mark_next++;
}

#endif /* __ASM_ARCH_NS9XXX_HPERF_H */
