/*
 *  /targets/U-Boot.cvs/common/digi/cmd_bsp.c
 *
 *  Copyright (C) 2006 by Digi International Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version2  as published by
 *  the Free Software Foundation.
*/

/*
 *  !Revision:   $Revision: 1.47 $
 *  !Author:     Markus Pietrek
 *  !Descr:      
 *  !References: [1] http://www.linux-mtd.infradead.org/doc/nand.html
*/

/*
 * Digi CC specific functions
 */

#include <common.h>
#if (CONFIG_COMMANDS & CFG_CMD_BSP)
#include <command.h>
#if (CONFIG_COMMANDS & CFG_CMD_NAND)
# include <nand.h>
#endif

#include <jffs2/jffs2.h>
#include <net.h>                /* DHCP */

#include "cmd_bsp.h"
#include "env.h"
#include "fpga_checkbitstream.h"
#include "nvram.h"
#include "partition.h"
#include "helper.h"

#define SNFS	"snfs"
#define SMTD	"smtd"
#define NPATH	"npath"
#define RIMG    "rimg"
#define USRIMG  "usrimg"
#define CONSOLE "console"

/* to transform a value into a string, from environment.c */

#define CE( sCmd ) \
        do { \
                if( !(sCmd) )                   \
                        goto error; \
        } while( 0 )

#define RUN_CMD(sCmd,...)                            \
        ({                                        \
                char _szCmd[ 64 ];                \
                sprintf( _szCmd, sCmd, __VA_ARGS__ );   \
                (RunCmd( _szCmd ) );   \
        })

#if (CONFIG_COMMANDS & CFG_CMD_NAND)
# define RUN_CMD_PROTECT_ON( pPartEntry ) \
        ({ do { } while( 0 ); (1); })
# define RUN_CMD_PROTECT_OFF( pPartEntry )    \
        ({ do { } while( 0 ); (1); })
#else
# define RUN_CMD_PROTECT_ON( pPartEntry ) \
        RUN_CMD( "protect on 0x%qx +0x%qx", (pPartEntry)->ullStart, (pPartEntry)->ullSize )
# define RUN_CMD_PROTECT_OFF( pPartEntry ) \
        RUN_CMD( "protect off 0x%qx +0x%qx", (pPartEntry)->ullStart, (pPartEntry)->ullSize )
#endif  /* (CONFIG_COMMANDS & CFG_CMD_NAND) */

#ifdef CFG_CONS_INDEX_SUB_1
# define CONSOLE_INDEX ((CONFIG_CONS_INDEX)-1)
#else
# define CONSOLE_INDEX (CONFIG_CONS_INDEX)
#endif

DECLARE_GLOBAL_DATA_PTR;

typedef enum {
        IS_FLASH,
        IS_TFTP,
        IS_USB
} image_source_e;
    
/* ********** local typedefs ********** */

typedef struct {
        nv_os_type_e   eOSType; 
        const char*    szName;       /*! short OS name */
        const char*    szEnvVar; 
        nv_part_type_e ePartType;
        char           bForBoot;
        char           bRootFS;
} part_t;

typedef struct {
        const char*    szEnvVar; 
        const char*    szEnvDflt;
} env_default_t;
    
typedef struct {
        image_source_e eType; 
        const char*    szName;       /*! short image name */
} image_source_t;

/* ********** local functions ********** */

static int do_digi_dboot( cmd_tbl_t* cmdtp, int flag, int argc, char* argv[] );
static int do_digi_update( cmd_tbl_t* cmdtp, int flag, int argc, char* argv[] );
static int jffs2_mark_clean (long long offset, long long size);

static const part_t*      WhatPart( const char* szOS, char bForBoot);
static const image_source_t* WhatImageSource( const char* szSrc );
static int RunCmd( const char* szCmd );
static int GetIntFromEnvVar( /*@out@*/ int* piVal, const char* szVar,
                             char bSilent );
static int AppendPadding( const nv_param_part_t* pPart, void* pvImage,
                          /*@out@*/ int* piFileSize );
static size_t GetEraseSize( const nv_param_part_t* pPart );
static size_t GetPageSize( const nv_param_part_t* pPart );
static const nv_param_part_t* FindPartition( const char* szName );
static int GetDHCPEnabled( char* pcEnabled );
static int DoDHCP( void );

/* ********** local variables ********** */

static const env_default_t l_axEnvDynamic[] = {
	{ CONSOLE,       NULL                         	   },  /* auto-generated */
	{ "ebootaddr",   MK_STR( CFG_LOAD_ADDR )           },
        { "eimg",        "eboot-"CONFIG_MODULE_NAME_WCE    },    
        { "fimg",        "wifi.biu"                        },
	{ "ip",          "ip=${ipaddr}:${serverip}:${gatewayip}:${netmask}:" CONFIG_HOSTNAME ":eth0:off" },
	{ "loadaddr",    MK_STR( CFG_LOAD_ADDR )           },
        { "kimg",        CONFIG_LINUX_IMAGE_NAME           },
        { NPATH,         "/exports/nfsroot-"CONFIG_MODULE_NAME },
	{ "linuxloadaddr", MK_STR( CFG_LOAD_ADDR )           },
	{ "netosloadaddr", MK_STR( CFG_NETOS_LOAD_ADDR )   },
#if defined(CONFIG_NS9360) || defined(CONFIG_NS9750)
        /* NET+OS exists only for these platforms */
        { "nimg",        "image-"CONFIG_MODULE_NAME_NETOS".bin"},
#endif  /* defined(CONFIG_NS9360) || defined(CONFIG_NS9750) */
        { RIMG,          NULL                              }, /* auto-generated */
	{ SMTD,          ""                                },
	{ SNFS,          "root=nfs nfsroot=${serverip}:"   },
        { "std_bootarg", ""                                },
        { "uimg",        CONFIG_UBOOT_IMAGE_NAME           },
        { USRIMG,        NULL                              }, /* auto-generated */
	{ "wceloadaddr",   MK_STR( CFG_WCE_LOAD_ADDR )     },
        { "wimg",        "wce-"CONFIG_MODULE_NAME_WCE      },
};
    
static const part_t l_axPart[] = {
        { NVOS_LINUX, "linux",  "kimg",   NVPT_LINUX,      .bForBoot = 1 },
        { NVOS_EBOOT, "eboot",  "eimg",   NVPT_EBOOT,      .bForBoot = 1 },
        { NVOS_WINCE, "wce",    "wimg",   NVPT_WINCE,      .bForBoot = 1 },
        { NVOS_NETOS, "netos",  "nimg",   NVPT_NETOS,      .bForBoot = 1 },
        { NVOS_UBOOT, "uboot",  "uimg",   NVPT_UBOOT,                    },
        { NVOS_NONE,  "fpga",   "fimg",   NVPT_FPGA,                     },
        { NVOS_NONE,  "rootfs", RIMG,     NVPT_FILESYSTEM, .bRootFS = 1  },
        { NVOS_NONE,  "userfs", USRIMG,   NVPT_FILESYSTEM,               },
};

/* use MK because we later use l_axImgSrc[ IS_FLASH ] */
#define MK(x,y) [x] = { .eType = x, .szName = y }
static const image_source_t l_axImgSrc[] = {
        MK( IS_FLASH, "flash" ),
        MK( IS_TFTP,  "tftp"  ),
        MK( IS_USB,   "usb"   ),
};
#undef MK

/* ********** global functions ********** */

/* ********** local functions ********** */

static int do_digi_dboot( cmd_tbl_t* cmdtp, int flag, int argc, char* argv[] )
{
        int iLoadAddr    = -1;
        int iRes         = 0;
        char bAppendMtdRootToBootargs = 1;
        const part_t* pPart = NULL;
        const image_source_t* pImgSrc = NULL;
        const nv_param_part_t* pPartEntry = NULL;
	char szCmd[ 60 ]  = "";
	char szImg[ 60 ]  = "";
        char bIsNFSRoot   = 0;
        char bDHCPEnabled = 0;
        nv_os_type_e eOSType = NVOS_NONE;
        
        if( ( argc < 2 ) || ( argc > 3 ) )
                goto usage;

        /* check what to boot */

        pImgSrc = ( ( NVOS_EBOOT != pPart->eOSType ) ?
                    WhatImageSource( argv[ 2 ] ) :
                    &l_axImgSrc[ IS_FLASH ] );
        if( NULL == pImgSrc )
                goto usage;

        /* determine OS and/or partition */
        pPart = WhatPart( argv[ 1 ], 1 );
        if( NULL == pPart ) {
                if( IS_FLASH == pImgSrc->eType ) {
                        /* Find flash partition by name */
                        pPartEntry = FindPartition( argv[ 1 ] );
                        if( NULL == pPartEntry ) {
                                eprintf( "Partition %s not found\n",
                                         argv[ 1 ] );
                                goto error;
                        }
                        /* what OS is probably on that partition? */
                        switch( pPartEntry->eType ) {
                            case NVPT_LINUX: eOSType = NVOS_LINUX; break;
                            case NVPT_WINCE: eOSType = NVOS_WINCE; break;
                            case NVPT_EBOOT: eOSType = NVOS_EBOOT; break;
                            case NVPT_NETOS: eOSType = NVOS_NETOS; break;
                            default:
                                eprintf( "OS Type not detected for %s\n",
                                         argv[ 1 ] );
                                goto usage;
                        }
                } else {
                        /* everything else requires valid pPart for
                           os system configuration / environment variable */
                        eprintf( "Partition table contains no entries for: %s\n", argv[ 1 ] );
                        goto error;
                } /* if( IS_FLASH ) */
        } else {
                if( ( IS_FLASH == pImgSrc->eType ) &&
                    !NvParamPartFind( &pPartEntry, pPart->ePartType,
                                      NVFS_NONE, 0, 0 ) ) {
                        eprintf( "No partition found for %s\n",
                                 pPart->szName );
                        goto error;
                }
                
                eOSType = pPart->eOSType;
                /* retreive image name */
                if( NULL != pPart->szEnvVar ) {
                        const char* szTmp = GetEnvVar( pPart->szEnvVar, 0 );
                        CE( NULL != szTmp );
                        strcpy( szImg, szTmp );
                }
        } /* if( NULL == pPart ) */
        /* pPartEntry is valid on FLASH booting */

        CE( GetDHCPEnabled( &bDHCPEnabled ) );
        
        /* user input processed, determine addresses */
        switch( eOSType ) {
            case NVOS_LINUX:
                CE( GetIntFromEnvVar( &iLoadAddr, "linuxloadaddr", 0 ) );
                break;
            case NVOS_WINCE:
                CE( GetIntFromEnvVar( &iLoadAddr, "wceloadaddr", 0 ) );
                break;
            case NVOS_EBOOT:
                CE( GetIntFromEnvVar( &iLoadAddr, "ebootaddr", 0 ) );
                break;
            case NVOS_NETOS:
                CE( GetIntFromEnvVar( &iLoadAddr, "netosloadaddr", 0 ) );
                break;
            default:
                (void) GetIntFromEnvVar( &iLoadAddr, "loadaddr", 1 );
                eprintf( "Operating system not supported\n" );
        }

        if( -1 == iLoadAddr ) {
                eprintf( "variable loadaddr does not exists\n" );
                goto error;
        }

        if( bDHCPEnabled && ( IS_TFTP == pImgSrc->eType ) )
                /* makes no sense for USB or FLASH download to need DHCP */
                CE( DoDHCP() );

        /* run boot scripts and get images into RAM */
        switch( pImgSrc->eType ) {
            case IS_TFTP:
                bAppendMtdRootToBootargs = 0;  /* everything is loaded by network */
                sprintf( szCmd, "tftp 0x%x %s", iLoadAddr, szImg );
                if( RunCmd( szCmd ) )
                        iRes = 1;
                bIsNFSRoot = 1;
                break;
            case IS_USB:
                sprintf( szCmd,  "fatload usb 0 0x%x %s", iLoadAddr, szImg );

                if( RunCmd( "usb reset" ) &&
                    RunCmd( szCmd ) )
                        iRes = 1;
                break;
            case IS_FLASH:
            {
                    /* copy kernel partition */
                    iRes = PartRead( pPartEntry, (void*) iLoadAddr, 0, 0 );

                    /* !TODO. Should only be done for linux */
                    if( NvParamPartFind( &pPartEntry, NVPT_FILESYSTEM,
                                         NVFS_INITRD, 1, 0 ) ) {
                            /* we have a bootable initrd, copy it, too */
                            iRes = 0;
                            int iLoadAddrInitRD = 0;
                            if( GetIntFromEnvVar( &iLoadAddrInitRD,
                                                  "loadaddr_initrd", 0 ) )
                                    iRes = PartRead( pPartEntry, (void*) iLoadAddrInitRD, 0, 0 );
                    }

                    break;
            }
        } /* switch( pImgSrc->eType ) */

        CE( iRes );
        
        /* boot operating system */
        switch( eOSType ) {
            case NVOS_LINUX:
            {
                    char szBootargs[ 2048 ];
                    strcpy( szBootargs, "setenv bootargs " );
                     /* no snprintf, therefore strncat */
                    strncat( szBootargs, GetEnvVar( "std_bootarg", 0 ), sizeof( szBootargs ) );

                    if( !bDHCPEnabled || ( IS_TFTP == pImgSrc->eType ) ) {
                            /* 
                             * IS_TFTP for linux means also NFSROOT. Userspace
                             * can't do DHCP then. Therefore we give the DHCP
                             * settings here. Otherwise they are the values
                             * from NVRAM */
                            strncat( szBootargs, " ", sizeof( szBootargs ) );
                            strncat( szBootargs, GetEnvVar( "ip", 0 ), sizeof( szBootargs ) );
                    }
                    
                    strncat( szBootargs, " ", sizeof( szBootargs ) );
                    strncat( szBootargs, GetEnvVar( "console", 0 ), sizeof( szBootargs ) );

                    strncat( szBootargs, " ", sizeof( szBootargs ) );
                    if( bIsNFSRoot ) {
                            strncat( szBootargs, GetEnvVar( SNFS, 0 ), sizeof( szBootargs ) );
                            strncat( szBootargs, GetEnvVar( NPATH, 0 ), sizeof( szBootargs ) );
                    } else 
                            strncat( szBootargs, GetEnvVar( SMTD, 0 ), sizeof( szBootargs ) );
                    strncat( szBootargs, " ", sizeof( szBootargs ) );
                    
                    CE( PartStrAppendParts( szBootargs, sizeof( szBootargs ) ) );
                    if( bAppendMtdRootToBootargs ) {
                            strncat( szBootargs, " ", sizeof( szBootargs ) );
                            CE( PartStrAppendRoot( szBootargs, sizeof( szBootargs ) ) );
                    }

                    CE( RunCmd( szBootargs ) );
                    sprintf( szCmd, "bootm" );
                    break;
            }
            case NVOS_WINCE:
            case NVOS_EBOOT:
                sprintf( szCmd, "go 0x%x", iLoadAddr );
                break;
            case NVOS_NETOS:
                sprintf( szCmd, "bootm 0x%x", iLoadAddr );
                break;
            default:
                eprintf( "Operating system not supported\n" );
        } /* switch( eOSType ) */

        if( NVOS_EBOOT == eOSType )
		printf( "%s will be booted now from %s\n",
                        argv[ 1 ], pImgSrc->szName );
	else
		printf( "%s will be booted now\n", argv[ 1 ] );

        /* check from what source to boot */
        iRes = RunCmd( szCmd );

        return !iRes;

usage:
        printf( "Usage:\n%s\n", cmdtp->usage );
        return 1;

error:
        return 1;
}

static int do_digi_update( cmd_tbl_t* cmdtp, int flag, int argc, char* argv[] )
{
        int iLoadAddr  = -1;
        const image_source_t* pImgSrc = NULL;
        const part_t* pPart           = NULL;
        const char*   szTmp           = NULL;
        const nv_param_part_t* pPartEntry = NULL;
        int iPartition = 0;
        int iRes       = 0;
        int iFileSize  = 0;
        crc32_t uiCRC;
	char szCmd[ 60 ] = "";
	char szImg[ 60 ] = "";
        char bDHCPEnabled = 0;
        
        if( ( argc != 3 ) && ( argc != 4 ) )
                goto usage;

        /* check what to update */
        pPart = WhatPart( argv[ 1 ], 0 );
        /* internal abbrevation name userfs/rootfs/u-boot/not found,
           try it with argv[1] as name */
        if( NULL == pPart ) {
                pPartEntry = FindPartition( argv[ 1 ] );
                if( NULL == pPartEntry ) {
                        eprintf( "Partition %s not found\n",
                                argv[ 1 ] );
                        goto error;
                }
        } else {
                /* using predefined partition information */
                /* find first rootfs partition */
                if( !NvParamPartFind( &pPartEntry, pPart->ePartType,
                                      NVFS_NONE, pPart->bRootFS, iPartition ) ) {
                        eprintf( "No partition found for %s\n", pPart->szName );
                        goto error;
                }
        }
        
        pImgSrc = WhatImageSource( argv[ 2 ] );
        if( NULL == pImgSrc )
                goto usage;

        /* determine file to update */
        if( argc == 4 )
                szTmp = argv[ 3 ];
        else if( NULL != pPart ) {
                /* not present, but we have a partition definition */
                szTmp = GetEnvVar( pPart->szEnvVar, 0 );
                CE( NULL != szTmp );
        } else {
                eprintf( "Require filename\n" );
                goto error;
        }
        
        strncat( szImg, szTmp, sizeof( szImg ) );

        CE( GetDHCPEnabled( &bDHCPEnabled ) );
        
        /* we check result later, it may be set otherwise */
        (void) GetIntFromEnvVar( &iLoadAddr, "loadaddr", 1 );

        /* we require it being set from download tool */
        setenv( "filesize", "" );

        if( bDHCPEnabled )
                CE( DoDHCP() );

        /* get images into RAM */
        switch( pImgSrc->eType ) {
            case IS_TFTP:
                sprintf( szCmd, "tftp 0x%x %s", iLoadAddr, szImg );
                if( RunCmd( szCmd ) )
                        iRes = 1;
                break;
            case IS_USB:
                sprintf( szCmd, "fatload usb 0 0x%x %s", iLoadAddr, szImg );
                if( RunCmd( "usb reset" ) &&
                    RunCmd( szCmd ) )
                        iRes = 1;
                break;
            case IS_FLASH:
                /* makes really no sense in the update func */
                goto usage;
        }

        CE( iRes );

        /* should be set by download tool */
        CE( GetIntFromEnvVar( &iFileSize, "filesize", 0 ) );

        uiCRC = crc32( 0, (uchar*) iLoadAddr, iFileSize );
        printf( "Calculated checksum = 0x%x\n", uiCRC );
        
        /* run some checks based on the image */

        switch( pPartEntry->eType ) {
            case NVPT_UBOOT:
                CE( WaitForYesPressed( "Do you really want to overwrite U-Boot flash partition" ) );
                break;
            case NVPT_FPGA:
            {
#ifdef CFG_FPGA_SIZE
                    if( CFG_FPGA_SIZE != ( iFileSize - 4 ) ) {
                            /* +4 because of checksum added */
                            eprintf( "Expecting FPGA to have size 0x%x\n",
                                     CFG_FPGA_SIZE );
                            goto error;
                    }
                    
                    printf( "Updating FPGA firmware ...\n" );
                    if( LOAD_FPGA_FAIL == fpga_checkbitstream( (uchar*) iLoadAddr, iFileSize ) ) {
                            eprintf( "Updating FPGA firmware failed\n" );
                            goto error;
                    }

                    /* fpga_checkbitstream doesn't print \0 */
                    printf( "\n" );
                    CE( WaitForYesPressed( "Do you really want to overwrite FPGA firmware" ) );
#else  /* CONFIG_FPGA_SIZE */
                    printf( "No FPGA available\n" );
#endif  /* CONFIG_FPGA_SIZE */
                    break;
            }
            default:
                iRes = 1;
                /* do nothing */
        }
        
        /* update images */

        /* fit we into it */
        /* !TODO. calculate bad blocks into it. PartWrite is exact. */
        if( PartSize( pPartEntry ) < iFileSize ) {
                /* 0 means to end of flash */
                eprintf( "Partition too small\n" );
                goto error;
        }
        
        /* erase complete partition */
        CE( RUN_CMD_PROTECT_ON( pPartEntry ) );
        CE( PartErase( pPartEntry ) );
        CE( AppendPadding( pPartEntry, (void*) iLoadAddr, &iFileSize ) );

        /* write partition */
        CE( PartWrite( pPartEntry, (void*) iLoadAddr, iFileSize ) );
        /* verify it */
        CE( PartVerify( pPartEntry, (void*) iLoadAddr, iFileSize ) );

        if( ( NVPT_FILESYSTEM == pPartEntry->eType ) &&
            ( NVFS_JFFS2      == pPartEntry->flags.fs.eType ) ) {
                printf( "Writing cleanmarkers\n" );
                CE( jffs2_mark_clean( pPartEntry->ullStart,
                                      pPartEntry->ullSize ) );
        }
        
        /* verify is done internally by write */
        CE( RUN_CMD_PROTECT_OFF( pPartEntry ) );

        switch( pPartEntry->eType ) {
            case NVPT_FPGA:
            case NVPT_UBOOT:
                /* be user friendly. fpga is not loaded */
                printf( "Reboot system so update takes effect\n" );
                break;
            default:
                break;
        }

        printf( "Update successful\n" );

        return 0;
usage:
        printf( "Usage:\n%s\n", cmdtp->usage );

error:
        return 1;
}

static int jffs2_mark_clean (long long offset, long long size)
{
	struct mtd_info *nand;
	int i, magic_ofs, magic_len;
	long long end;
	unsigned char magic[] = {0x85, 0x19, 0x03, 0x20, 0x08, 0x00, 0x00, 0x00};

#if !(CONFIG_COMMANDS & CFG_CMD_NAND)
# error writing of CLEAN MARKER not yet implemented
#endif
	for (i = 0; i < CFG_MAX_NAND_DEVICE; i++) {
		if (nand_info[i].name)
			break;
	}
	nand = &nand_info[i];

        /* !see [1] */
	switch(nand->oobsize) {
		case 8:
			magic_ofs = 0x06;
			magic_len = 0x02;
                        break;
		case 16:
			magic_ofs = 0x08;
			magic_len = 0x08;
			break;
		case 64:
			magic_ofs = 0x10;
			magic_len = 0x08;
			break;
		default:
			printf("Cannot set markers on this oobsize!\n");
                        goto error;
	}

	/* calculate end */
	for(end = offset + size; offset < end; offset += nand->erasesize) {
                size_t iTmp;
		/* skip if block is bad */
		if(nand->block_isbad(nand, offset))
			continue;

		/* modify oob */

		if( nand->write_oob(nand, offset + magic_ofs, magic_len, &iTmp, magic) )
                        goto error;
	} /* for( end=offset) */

	return 1;

error:
        return 0;
}

static int do_envreset( cmd_tbl_t* cmdtp, int flag, int argc, char* argv[] )
{
	switch( argc ) {
            case 1:
		break;
                
            default:
		printf( "Usage:\n%s\n", cmdtp->usage );
		return -1;
	}

	printf( "Environment will be set to Default now!\n" );

        NvEnvUseDefault();

        return saveenv();
}

static int do_printenv_dynamic( cmd_tbl_t* cmdtp, int flag, int argc, char* argv[] )
{
        int i;

        for( i = 0; i < ARRAY_SIZE( l_axEnvDynamic ); i++ ) {
                const char* szVar = l_axEnvDynamic[ i ].szEnvVar;
                printf( "%s=%s\n", szVar,  GetEnvVar( szVar, 0 ) );
        }
        
        return 0;
}

static const part_t* WhatPart( const char* szOS, char bForBoot )
{
        int i = 0;
        const part_t* pPart = NULL;
        
        while( i < ARRAY_SIZE( l_axPart ) ) {
                if( !strcmp( szOS, l_axPart[ i ].szName ) &&
                    ( !bForBoot || l_axPart[ i ].bForBoot ) ) {
                        const nv_param_part_t* pPartEntry;

                        pPart = &l_axPart[ i ];

                        /* abbreviated name found, does it exists? */
                        /* find first rootfs partition */
                        if( !NvParamPartFind( &pPartEntry, pPart->ePartType,
                                             NVFS_NONE, pPart->bRootFS, 0 ) )
                                /* no */
                                pPart = NULL;
                        break;
                }
                i++;
        } /* !while */

        return pPart;
}

static const image_source_t* WhatImageSource( const char* szSrc )
{
        int i = 0;
        
        while( i < ARRAY_SIZE( l_axImgSrc ) ) {
                if( !strcmp( szSrc, l_axImgSrc[ i ].szName ) )
                        return &l_axImgSrc[ i ];

                i++;
        } /* !while */

        return NULL;
}

/*! \brief Runs command and prints error on failure */
/*! \return 0 on failure otherwise 1
 */
static int RunCmd( const char* szCmd )
{
        int iRes = ( run_command( szCmd, 0 ) >= 0 );
        if( !iRes )
                eprintf( "command %s failed\n", szCmd );

        return iRes;
}

/*! \brief Returns dynamic or normal environment variable */
/*! \param bSilent true if not existant should not be reported
 */
const char* GetEnvVar( const char* szVar, char bSilent )
{
        const char* szTmp = NULL;
        static char szTmpBuf[ 128 ];

        szTmp = getenv( (char*) szVar );
        if( ( NULL == szTmp ) ) {
                /* variable not yet defined, try to read it in l_axPart */
                if( !strcmp( szVar, RIMG ) || !strcmp( szVar, USRIMG ) ) {
                        char bRIMG = !strcmp( szVar, RIMG );
                        const nv_param_part_t* pPartEntry;
                        size_t iEraseSize = GetEraseSize( NULL );
                        
                        /* find first rootfs partition */
                        if( !NvParamPartFind( &pPartEntry, NVPT_FILESYSTEM,
                                              NVFS_NONE, bRIMG, 0 ) )
                                return "";
                        
                        iEraseSize = GetEraseSize( pPartEntry );
                        
                        if( bRIMG )
                                sprintf( szTmpBuf, "%s-%i.jffs2",
                                         ( (CONFIG_IMAGE_JFFS2_BASENAME == NULL) ? "rootfs-"CONFIG_MODULE_NAME :
                                           CONFIG_IMAGE_JFFS2_BASENAME ), iEraseSize / 1024 );
                        else
                                sprintf( szTmpBuf, "userfs-%s-%i.jffs2",
                                         CONFIG_MODULE_NAME, iEraseSize / 1024 );
                        szTmp = szTmpBuf;
                } else if( !strcmp( szVar, CONSOLE ) ) {
                        sprintf( szTmpBuf, "console="CFG_CONSOLE"%i,%s",
                                 CONSOLE_INDEX,
                                 getenv( "baudrate" ) );
                        szTmp = szTmpBuf;
                } else {
                        int i = 0;
                        
                        while( i < ARRAY_SIZE( l_axEnvDynamic ) ) {
                                if( !strcmp( l_axEnvDynamic[ i ].szEnvVar, szVar ) ) {
                                        szTmp = l_axEnvDynamic[ i ].szEnvDflt;
                                        break;
                                }
                                i++;
                        } /* while */
                }
                
                if( ( NULL == szTmp ) && !bSilent )
                        eprintf( "variable %s does not exists\n", szVar );
        }
        
        
        return szTmp;
}

/*! \brief Converts an environment variable to an integer */
/*! \return 0 on failure otherwise 1
 */
static int GetIntFromEnvVar( int* piVal, const char* szVar, char bSilent )
{
        int iRes = 0;
        const char* szVal = GetEnvVar( szVar, bSilent );

        if( NULL != szVal ) {
                *piVal = simple_strtoul( szVal, NULL, 16 );
                iRes = 1;
        }
        
        return iRes;
}

/*! \brief padds the image, also adds JFFS2 Padding if it's a jffs2 partition. */
/*! \return 0 on failure otherwise wise 1
 */
static int AppendPadding( const nv_param_part_t* pPart, void* pvImage,
                          int* piFileSize )
{
        int iPageSize = GetPageSize( pPart );
        int iBytesFreeInBlock;

        if( !iPageSize )
                return 0;

        iBytesFreeInBlock = iPageSize - (*piFileSize % iPageSize);

        if( ( NVPT_FILESYSTEM == pPart->eType ) &&
            ( NVFS_JFFS2      == pPart->flags.fs.eType ) &&
            ( iBytesFreeInBlock >= sizeof( struct jffs2_unknown_node ) ) ) {
                /* JFFS CRC32 starts from 0xfffffff, our crc32 from 0x0 */
                uint32_t uiStart = 0xffffffff;
                /* write padding to avoid Empty block messages.
                   see linux/fs/jffs2/wbuf.c:flush_wbuf */
                struct jffs2_unknown_node* pNode = (struct jffs2_unknown_node*) ( pvImage + *piFileSize );

                printf( "Padding last sector\n" );

                /* 0 for JFFS2 PADDING Type, see wbuf.c */
                memset( pvImage + *piFileSize, 0x0, iBytesFreeInBlock );

                *piFileSize       += iBytesFreeInBlock;
                iBytesFreeInBlock -= sizeof( *pNode );
                
                CLEAR( *pNode );
                
                /* no cpu_to_je16, it's always host endianess in U-Boot */
                pNode->magic    = JFFS2_MAGIC_BITMASK;
                pNode->nodetype = JFFS2_NODETYPE_PADDING;
                
                pNode->totlen   = iBytesFreeInBlock;
                /* don't CRC32 the hdr_crc itself */
                pNode->hdr_crc  = crc32( uiStart, (uchar*) pNode, sizeof( *pNode ) - 4 ) ^ uiStart;
        } else {/* if( JFFS2 ) */
                /* clear remaining stuff */
                memset( pvImage + *piFileSize, 0xff, iBytesFreeInBlock );
        }
        
        return 1;
}

static size_t GetEraseSize( const nv_param_part_t* pPart ) 
{
        size_t iEraseSize;

#if (CONFIG_COMMANDS & CFG_CMD_NAND)
        iEraseSize = nand_info[ 0 ].erasesize;
#else
# error GetEraseSize not supported yet
#endif

        return iEraseSize;
}

static size_t GetPageSize( const nv_param_part_t* pPart )
{
        size_t iPageSize;

#if (CONFIG_COMMANDS & CFG_CMD_NAND)
        iPageSize = nand_info[ 0 ].oobblock;
#else
        /* NOR doesn't know anthing about pages */
        iPageSize = GetEraseSize( pPart );
#endif

        return iPageSize;
}

static const nv_param_part_t* FindPartition( const char* szName )
{
        nv_critical_t*               pCrit      = NULL;
        const nv_param_part_table_t* pPartTable = NULL;
        unsigned int u;

        if( !CW( NvCriticalGet( &pCrit ) ) )
                return NULL;

        pPartTable = &pCrit->s.p.xPartTable;

        for( u = 0; u < pPartTable->uiEntries; u++ )
                if( !strcmp( szName, pPartTable->axEntries[ u ].szName ) )
                        return &pPartTable->axEntries[ u ];

        return NULL;
}

static int GetDHCPEnabled( char* pcEnabled )
{
        nv_critical_t* pCrit = NULL;

        *pcEnabled = 0;
        CE( NvCriticalGet( &pCrit ) );

        *pcEnabled = pCrit->s.p.xIP.axDevice[ 0 ].flags.bDHCP;

        return 1;

error:
        return 0;
}

/*! \brief performs a DHCP/BOOTP request without loading any file */
/*! \return 0 on failure otherwise 1
 */
static int DoDHCP( void )
{
        int  iRes = 0;
        char szTmp[ 80 ];
        
        /* get current value */
        strncpy( szTmp, getenv( "autoload" ), sizeof( szTmp ) );
        /* autoload is used in BOOTP to do TFTP itself.
           We do tftp ourself because we don't need bootfile */
        setenv( "autoload", "n" );
        if( NetLoop( DHCP ) >= 0 ) {
                /* taken from netboot_common. But we don't need autostart */
                netboot_update_env();
                iRes = 1;
        }
        /* restore old autoload */
        setenv( "autoload", szTmp );

        return iRes;
}


U_BOOT_CMD(
	dboot,	3,	0,	do_digi_dboot,
	"dboot   - Digi ConnectCore modules boot commands\n",
	"OS source\n"
	"  - boots 'OS' of 'source'\n"
	"    values for `OS': linux, wce, eboot, netos or any partition name\n"
	"    values for `source': flash, tftp, usb\n"
);

U_BOOT_CMD(
	update,	4,	0,	do_digi_update,
	"update  - Digi ConnectCore modules update commands\n",
	"partition source [file]\n"
	"  - updates 'partition' via 'source'\n"
	"    values for 'partition': uboot, linux, rootfs, userfs, eboot, wce, netos\n"
        "                            or any partition name\n"
	"    values for 'source': tftp, usb\n"
	"    values for 'file'  : the file to be used for updating\n"
);

U_BOOT_CMD(
	envreset,	1,	0,	do_envreset,
	"envreset- Sets environment variables to default setting\n",
	"\n"
	"  - overwrites all current variables\n"
);

U_BOOT_CMD(
	printenv_dynamic,	1,	0,	do_printenv_dynamic,
	"printenv_dynamic- Prints all dynamic variables\n",
	"\n"
	"  - Prints all dynamic variables\n"
);

#endif	/* (CONFIG_COMMANDS & CFG_CMD_BSP) */
